package nl.wur.bif.pantools.analysis.region_analysis;

import nl.wur.bif.pantools.Pantools;
import nl.wur.bif.pantools.utils.BeanUtils;
import nl.wur.bif.pantools.utils.FileUtils;
import nl.wur.bif.pantools.utils.Globals;
import nl.wur.bif.pantools.utils.GraphUtils;
import nl.wur.bif.pantools.utils.cli.mixins.SelectGenomes;
import nl.wur.bif.pantools.utils.cli.mixins.SelectHmGroups;
import nl.wur.bif.pantools.utils.cli.mixins.ThreadNumber;
import nl.wur.bif.pantools.utils.cli.validation.Constraints.OutputDirectory;
import picocli.CommandLine.*;

import java.nio.file.Path;
import java.util.concurrent.Callable;

import static nl.wur.bif.pantools.utils.Globals.THREADS;
import static nl.wur.bif.pantools.utils.Globals.setGenomeSelectionOptions;

/**
 * Extracts genes from a pangenome based on a list of homology groups.
 * @author Dirk-Jan van Workum, Wageningen University, the Netherlands.
 */
@Command(name = "extract_homology_groups", sortOptions = false)
@OutputDirectory(directory = "outputPath")
public class ExtractHomologyGroupCLI implements Callable<Integer> {
    @Spec Model.CommandSpec spec;
    @Mixin private ThreadNumber threadNumber;
    @ArgGroup private SelectGenomes selectGenomes;
    @ArgGroup private SelectHmGroups selectHmGroups = new SelectHmGroups();

    @ParentCommand
    private Pantools pantools;

    @Option(names = {"-o", "--output"},
            defaultValue = "extracted_region")
    private Path outputPath;

    @Option(names = {"--flanking"},
            defaultValue = "0",
            description = "The number of flanking bases to include in the output. Default: 0"
    )
    private int flanking;

    @Option(names = {"--write-annotations"},
            defaultValue = "true",
            negatable = true,
            fallbackValue = "true",
            description = "Write the annotations in GFF3 format. Default: true"
    )
    private boolean writeAnnotations;

    @Option(names = {"--write-gfa"},
            defaultValue = "false",
            negatable = true,
            fallbackValue = "true",
            description = "Write the GFA file; can be slow! Default: false"
    )
    private boolean writeGfa;

    @Option(names = {"--write-homology"},
            defaultValue = "true",
            negatable = true,
            fallbackValue = "true",
            description = "Write the homology groups. Default: true"
    )
    private boolean writeHomology;

    @Option(names = {"--write-unitigs"},
            defaultValue = "false",
            negatable = true,
            fallbackValue = "true",
            description = "Write the unitig (node) presence/absence; can be slow! Default: false"
    )
    private boolean writeUnitigs;

    @Option(names = {"--write-repeats"},
            defaultValue = "false",
            negatable = true,
            fallbackValue = "true",
            description = "Write the repeat annotations. Default: false"
    )
    private boolean writeRepeats;

    @Option(names = {"--reads-file"},
            description = "Tab-delimited file with library names and corresponding read files. Default: none"
    )
    private Path readsFile;

    @Option(names = {"--filter-high-frequency"},
            defaultValue = "true",
            negatable = true,
            fallbackValue = "true",
            description = "Filter out high-frequency k-mers for read reconstruction. Default: true"
    )
    private boolean filterHighFrequency;

    //TODO: add --write-blast option to visualize BLASTP and BLASTN results

    @Override
    public Integer call() throws Exception {
        pantools.createLogger(spec);
        BeanUtils.argValidation(spec, this, selectGenomes, selectHmGroups);

        pantools.setPangenomeGraph("pangenome");
        crossValidate();
        setGlobalParameters();
        if (writeRepeats) {
            GraphUtils.validateRepeats();
        }

        // always use fast option, except when writing GFA or unitigs, or using read files
        boolean fast = !writeGfa && !writeUnitigs && readsFile == null;
        if (!fast) {
            Pantools.logger.warn("Using a slow method for extraction.");
        }

        GraphUtils.validateAnnotations(); // check that annotations are present
        GraphUtils.validateHomologyGroups(); // check that homology groups are present

        final RegionExtractor regionExtractor = new RegionExtractor(outputPath);
        regionExtractor.extractHmGroups(selectHmGroups.getHomologyGroups(), flanking, writeAnnotations, writeGfa, writeHomology, writeUnitigs, writeRepeats, FileUtils.parseReadsFile(readsFile), filterHighFrequency, fast);

        return 0;
    }

    private void setGlobalParameters() {
        setGenomeSelectionOptions(selectGenomes);
        THREADS = threadNumber.getnThreads();
    }

    private void crossValidate() {
        // Assert that K is smaller than 64 when using reads (because of storing 2bit version of k-mers in long)
        if (readsFile != null && Globals.K_SIZE > 63) {
            throw new ParameterException(spec.commandLine(), "K should be smaller than 63 when using reads");
        }
    }

    public Pantools getPantools() {
        return pantools;
    }

    public Path getOutputPath() {
        return outputPath;
    }
}
