package nl.wur.bif.pantools.analysis.region_analysis;

import jakarta.validation.constraints.Max;
import jakarta.validation.constraints.Min;
import jakarta.validation.constraints.NotNull;
import jakarta.validation.constraints.Pattern;
import nl.wur.bif.pantools.Pantools;
import nl.wur.bif.pantools.utils.BeanUtils;
import nl.wur.bif.pantools.utils.FileUtils;
import nl.wur.bif.pantools.utils.Globals;
import nl.wur.bif.pantools.utils.FileUtils;
import nl.wur.bif.pantools.utils.GraphUtils;
import nl.wur.bif.pantools.utils.cli.mixins.SelectGenomes;
import nl.wur.bif.pantools.utils.cli.mixins.ThreadNumber;
import nl.wur.bif.pantools.utils.cli.validation.Constraints.OutputDirectory;
import nl.wur.bif.pantools.utils.graph.GenomeRange;
import picocli.CommandLine.*;

import java.nio.file.Path;
import java.util.concurrent.Callable;

import static nl.wur.bif.pantools.utils.Globals.THREADS;
import static nl.wur.bif.pantools.utils.Globals.setGenomeSelectionOptions;

/**
 * Extract a region in one genome from the pangenome graph based on coordinates.
 * @author Dirk-Jan van Workum, Wageningen University, the Netherlands.
 */
@Command(name = "extract_region", sortOptions = false)
@OutputDirectory(directory = "outputPath")
public class ExtractRegionCLI implements Callable<Integer> {
    @Spec Model.CommandSpec spec;
    @Mixin private ThreadNumber threadNumber;
    @ArgGroup private SelectGenomes selectGenomes;

    @ParentCommand
    private Pantools pantools;

    @Option(names = {"-o", "--output"},
            defaultValue = "extracted_region")
    private Path outputPath;

    @Option(names = {"--use-kmers"},
            defaultValue = "true",
            negatable = true,
            fallbackValue = "true",
            description = "Use k-mers for extraction. Default: true"
    )
    private boolean useKmers;

    @Option(names = {"--use-homology"},
            defaultValue = "false",
            negatable = true,
            fallbackValue = "true",
            description = "Use homology groups for extraction. Default: false"
    )
    private boolean useHomology;

    @Option(names = {"--use-blast"},
            defaultValue = "false",
            negatable = true,
            fallbackValue = "true",
            description = "Use BLAST results for extraction. Default: false"
    )
    private boolean useBlast;

    @Option(names = {"--allow-single-hits"},
            defaultValue = "false",
            negatable = true,
            fallbackValue = "true",
            description = "Allow single transcripts/k-mers to be used for extraction. Default: false (to reduce spurious hits)"
    )
    private boolean allowSinglePositions;

    @Option(names = {"--search-self"},
            defaultValue = "false",
            negatable = true,
            fallbackValue = "true",
            description = "Perform search in the input genome. Default: false"
    )
    private boolean searchSelf;

    @Option(names = {"--kmer-density"},
            defaultValue = "1000",
            description = "K-mer density for k-mer sampling. Default: one k-mer per 1000 bases."
    )
    @Min(value = 1)
    private int kmerDensity;

    @Option(names = {"--max-evalue"},
            defaultValue = "1e-5",
            description = "The maximum E-value for BLAST hits. Default: 1e-5"
    )
    private double maxEvalue;

    @Option(names = {"--max-distance"},
            defaultValue = "10000",
            description = "The maximum allowed distance between two transcripts/k-mers/BLAST hits; this determines the maximum gap that can be spanned within one block. Default: 10000"
    )
    @Min(value = 1)
    private int maxDistance;

    @Option(names = {"--max-frequency"},
            defaultValue = "5",
            description = "The maximum frequency of a k-mer/transcript/BLAST hits in a target genome. K-mers/transcripts above this threshold are not taken into account; but for BLAST, this value determines the number of most significant hits above the e-value threshold. Default: 5"
    )
    @Min(value = 1)
    private int maxFrequency;

    @Option(names = {"--fract-positions"},
            defaultValue = "0.01",
            description = "The fraction of total k-mers/transcripts/BLAST hits that need to be found to create a hit; reducing will give more (divergent) hits, increasing will only give conservative, contiguous hits. Default: 0.01"
    )
    @Min(value = 0)
    @Max(value = 1)
    private double fractPositions;

    @Option(names = {"--flanking"},
            defaultValue = "0",
            description = "The number of flanking bases to include in the output. Default: 0"
    )
    private int flanking;

    @Option(names = {"--write-annotations"},
            defaultValue = "true",
            negatable = true,
            fallbackValue = "true",
            description = "Write the annotations in GFF3 format. Default: true"
    )
    private boolean writeAnnotations;

    @Option(names = {"--write-gfa"},
            defaultValue = "false",
            negatable = true,
            fallbackValue = "true",
            description = "Write the GFA file; can be slow! Default: false"
    )
    private boolean writeGfa;

    @Option(names = {"--write-homology"},
            defaultValue = "true",
            negatable = true,
            fallbackValue = "true",
            description = "Write the homology groups. Default: true"
    )
    private boolean writeHomology;

    @Option(names = {"--write-unitigs"},
            defaultValue = "false",
            negatable = true,
            fallbackValue = "true",
            description = "Write the unitig (node) presence/absence; can be slow! Default: false"
    )
    private boolean writeUnitigs;

    @Option(names = {"--write-repeats"},
            defaultValue = "false",
            negatable = true,
            fallbackValue = "true",
            description = "Write the repeat annotations. Default: false"
    )
    private boolean writeRepeats;

    @Option(names = {"--reads-file"},
            description = "Tab-delimited file with library names and corresponding read files. Default: none"
    )
    private Path readsFile;

    @Option(names = {"--filter-high-frequency"},
            defaultValue = "true",
            negatable = true,
            fallbackValue = "true",
            description = "Filter out high-frequency k-mers for read reconstruction. Default: true"
    )
    private boolean filterHighFrequency;

    //TODO: add --write-blast option to visualize BLASTP and BLASTN results

    @NotNull(message = "Please provide a valid genome range: genomeNr#sequenceName:start-end")
    @Option(names = "--genome-range")
    @Pattern(regexp = "^[0-9]+#[^#:\\s]+:[0-9]+-[0-9]+$", message = "{pattern.genome-range}")
    private String genomeRangeStr;

    @Override
    public Integer call() throws Exception {
        pantools.createLogger(spec);
        BeanUtils.argValidation(spec, this, selectGenomes);

        pantools.setPangenomeGraph("pangenome");
        crossValidate();
        setGlobalParameters();

        if (writeAnnotations || useBlast) {
            GraphUtils.validateAnnotations(); // check that annotations are present
        }
        if (useHomology || writeHomology) {
            GraphUtils.validateHomologyGroups(); // check that homology groups are present
        }
        if (writeRepeats) {
            GraphUtils.validateRepeats(); // check that repeat annotations are present
        }

        // always use fast option, except when writing GFA or unitigs, or using read files
        boolean fast = !writeGfa && !writeUnitigs && readsFile == null;
        if (!fast) {
            Pantools.logger.warn("Using a slow method for extraction.");
        }

        final RegionExtractor regionExtractor = new RegionExtractor(outputPath);
        final GenomeRange inputGenomeRange = GenomeRange.parseGenomeRange(genomeRangeStr);
        regionExtractor.setGenomeRange(inputGenomeRange);
        regionExtractor.extractRegion(useKmers, useHomology, useBlast, allowSinglePositions,
                searchSelf, kmerDensity, maxEvalue, maxDistance, fractPositions, maxFrequency, flanking,
                writeAnnotations, writeGfa, writeHomology, writeUnitigs, writeRepeats, FileUtils.parseReadsFile(readsFile), filterHighFrequency, fast);

        return 0;
    }

    private void setGlobalParameters() {
        THREADS = threadNumber.getnThreads();
        setGenomeSelectionOptions(selectGenomes);
        THREADS = threadNumber.getnThreads();
    }

    private void crossValidate() {
        // Assert that --num-kmer-samples is only used for kmer method
        if (!useKmers) {
            if (kmerDensity != 1000) {
                throw new ParameterException(spec.commandLine(), "--num-kmer-samples can only be used with --use-kmers=true");
            }
        }

        // Assert that --max-evalue is only used for BLAST method
        if (!useBlast) {
            if (maxEvalue != 1e-5) {
                throw new ParameterException(spec.commandLine(), "--max-evalue can only be used with --use-blast=true");
            }
        }

        // Assert that K is smaller than 64 when using reads (because of storing 2bit version of k-mers in long)
        if (readsFile != null && Globals.K_SIZE > 63) {
            throw new ParameterException(spec.commandLine(), "K should be smaller than 63 when using reads");
        }
    }

    public Pantools getPantools() {
        return pantools;
    }
    public Path getOutputPath() {
        return outputPath;
    }
}
