package nl.wur.bif.pantools.analysis.region_analysis;

import com.google.common.collect.ImmutableList;
import htsjdk.samtools.fastq.FastqRecord;
import htsjdk.samtools.fastq.FastqWriterFactory;
import htsjdk.samtools.fastq.FastqWriter;
import htsjdk.samtools.fastq.AsyncFastqWriter;
import me.tongfei.progressbar.ProgressBar;
import me.tongfei.progressbar.ProgressBarBuilder;
import me.tongfei.progressbar.ProgressBarStyle;
import nl.wur.bif.pantools.Pantools;
import nl.wur.bif.pantools.utils.Globals;
import nl.wur.bif.pantools.utils.Utils;
import nl.wur.bif.pantools.utils.graph.graph_objects.GraphRange;
import nl.wur.bif.pantools.utils.reader.FastqReader;

import java.io.*;
import java.nio.file.Files;
import java.nio.file.Path;
import java.util.ArrayList;
import java.util.Collections;
import java.util.List;
import java.util.concurrent.*;
import java.util.concurrent.atomic.AtomicInteger;
import java.util.stream.Collectors;
import java.util.stream.LongStream;

public class ReadReconstructor {
    private static final int HIGH_KMER_FREQUENCY_THRESHOLD = Globals.GENOME_DB.num_genomes * 100; // 100 times the number of genomes; definition from GenomeLayer.java
    private final double MINIMUM_MATCHING_KMERS_FRAC = 0.6;
    private static final int BATCH_SIZE = 10_000;
    private static final int QUEUE_CAPACITY = 10_000_000;
    private static final int PROGRESS_BAR_UPDATE_INTERVAL = 100;

    private final ImmutableList<Long> allKmers; // all k-mers stored in 2bit format and sorted for binary search
    private final ImmutableList<Long> highFrequencyKmers; // high frequency k-mers stored in 2bit format and sorted for binary search
    private BlockingQueue<FastqRecord[]> readQueue;
    private volatile boolean finishedReadingFromQueue = false;
    private final AtomicInteger activeFilterThreads;
    private final AtomicInteger totalFilterCount;
    private BlockingQueue<FastqRecord[]> writeQueue;
    private boolean finishedWritingToQueue = false;
    private ProgressBarManager pbm;

    private String libraryName;

    private static final AtomicInteger preFilterCount = new AtomicInteger(0);
    private static final AtomicInteger postFilterCount = new AtomicInteger(0);

    public ReadReconstructor(GraphRange graphRange, boolean filterHighFrequency) {
        // obtain all k-mers from the graph range
        Pantools.logger.info("Obtaining k-mers from graph range.");
        this.allKmers = ImmutableList.copyOf(LongStream.of(graphRange.getKmersAs2bit()).boxed().collect(Collectors.toList()));
        if (filterHighFrequency) {
            this.highFrequencyKmers = ImmutableList.copyOf(LongStream.of(graphRange.getKmersAs2bit(HIGH_KMER_FREQUENCY_THRESHOLD)).boxed().collect(Collectors.toList()));
        } else {
            this.highFrequencyKmers = ImmutableList.of();
        }
        Pantools.logger.info("Obtained {} k-mers from graph range (of which {} are high-frequency).", allKmers.size(), highFrequencyKmers.size());

        // check that high-frequency k-mers are a strict subset of all k-mers
        if (!allKmers.containsAll(highFrequencyKmers)) {
            Pantools.logger.error("High-frequency k-mers are not a subset of all k-mers.");
            Pantools.logger.debug("All k-mers: {}", allKmers);
            Pantools.logger.debug("High-frequency k-mers: {}", highFrequencyKmers);
            Pantools.logger.debug("High-frequency k-mers not in all k-mers: {}", highFrequencyKmers.stream().filter(k -> !allKmers.contains(k)).collect(Collectors.toList()));
            throw new IllegalArgumentException("High-frequency k-mers are not a subset of all k-mers");
        }

        this.activeFilterThreads = new AtomicInteger(0);
        this.totalFilterCount = new AtomicInteger(0);
    }

    private void reset() {
        this.readQueue = null;
        this.finishedReadingFromQueue = false;
        this.activeFilterThreads.set(0);
        this.totalFilterCount.set(0);
        this.writeQueue = null;
        this.finishedWritingToQueue = false;
        this.pbm = null;

        preFilterCount.set(0);
        postFilterCount.set(0);
    }

    public void reconstruct(String libraryName, Path[] readFiles, Path outputPath) {
        Pantools.logger.info("Reconstructing region for {} from {} read files.", libraryName, readFiles.length);
        reset();
        this.libraryName = libraryName;

        // create output directory if it does not exist
        createOutputDirectory(outputPath);

        // set read length
        if (readFiles.length != 1 && readFiles.length != 2) {
            Pantools.logger.error("Invalid number of read files for library {}: {}", libraryName, readFiles.length);
            throw new IllegalArgumentException("Invalid number of read files.");
        }
        boolean pairedReads = readFiles.length == 2;

        // process the reads using a thread pool
        try {
            pbm = new ProgressBarManager();

            // Use separate thread pools for different tasks
            ExecutorService readerExecutor = Executors.newSingleThreadExecutor();
            final int filterThreadCount = Math.max(1, Globals.THREADS - 2);
            ExecutorService filterExecutor = Executors.newFixedThreadPool(
                    filterThreadCount,
                    r -> {
                        Thread t = new Thread(r);
                        t.setPriority(Thread.MAX_PRIORITY); // Prioritize filtering threads
                        return t;
                    }
            );
            ExecutorService writerExecutor = Executors.newSingleThreadExecutor();
            Pantools.logger.debug("Created thread pools for reader (1), filter ({}), and writer (1) tasks.", filterThreadCount);

            // Configure larger queue sizes
            readQueue = new LinkedBlockingQueue<>(QUEUE_CAPACITY);
            writeQueue = new LinkedBlockingQueue<>(QUEUE_CAPACITY);

            // Start reader thread
            readerExecutor.submit(new fastqFileReader(readFiles, pairedReads));

            // Start filter threads
            Pantools.logger.trace("READSTATS\tlibrary\tread_name\tnumber_k-mers\tthreshold_matches\tmatches\thigh_frequency_matches\trelevant_matches");
            activeFilterThreads.set(filterThreadCount);
            for (int i = 0; i < filterThreadCount; i++) {
                filterExecutor.submit(new fastqReadFilter(pairedReads));
            }

            // Start writer thread
            writerExecutor.submit(new fastqFileWriter(outputPath, libraryName, pairedReads));

            // Shutdown executors
            readerExecutor.shutdown();
            filterExecutor.shutdown();
            writerExecutor.shutdown();

            // Wait for completion
            readerExecutor.awaitTermination(10, TimeUnit.DAYS);
            filterExecutor.awaitTermination(10, TimeUnit.DAYS);
            writerExecutor.awaitTermination(10, TimeUnit.DAYS);

            Pantools.logger.debug("Filtered {} reads down to {} reads", preFilterCount.get(), postFilterCount.get());
        } catch (InterruptedException e) {
            Pantools.logger.error("Error while waiting for tasks to complete: {}", e.getMessage());
            throw new RuntimeException(e);
        }

        // reconstruct the region through assembly
        //TODO: filtering the reads as above works only so-so for lettuce

        // write the reconstructed region to the output path

    }

    private void createOutputDirectory(Path outputPath) {
        if (!Files.exists(outputPath)) {
            try {
                Files.createDirectories(outputPath);
            } catch (IOException ioe) {
                Pantools.logger.error("Error while creating output directory: {}", ioe.getMessage());
                throw new RuntimeException(ioe);
            }
        }
    }

    private static class ProgressBarManager {
        private final ProgressBar filteringBar;
        private volatile long totalReadsProcessed = 0;

        public ProgressBarManager() {
            filteringBar = new ProgressBarBuilder()
                    .setTaskName("Filtering fastq")
                    .setInitialMax(1)
                    .setStyle(ProgressBarStyle.ASCII)
                    .setUpdateIntervalMillis(PROGRESS_BAR_UPDATE_INTERVAL)
                    .build();
        }

        public ProgressBar getFilteringBar() {
            return filteringBar;
        }

        public void updateReadCount(long count) {
            totalReadsProcessed = count;
            filteringBar.maxHint(totalReadsProcessed);
        }
    }

    private class fastqFileReader implements Runnable {
        private final Path[] readFiles;
        private final boolean pairedReads;
        private long readCount;

        public fastqFileReader(Path[] readFiles, boolean pairedReads) {
            this.readFiles = readFiles;
            this.pairedReads = pairedReads;
            this.readCount = 0;
        }

        @Override
        public void run() {
            readFastqFiles(readFiles, pairedReads);
        }

        private void readFastqFiles(Path[] readFiles, boolean pairedReads) {
            try {
                if (pairedReads) {
                    readPairedEndFiles(readFiles);
                } else {
                    readSingleEndFile(readFiles[0]);
                }
            } catch (IOException e) {
                Pantools.logger.error("Error while reading fastq files: {}", e.getMessage());
                throw new RuntimeException(e);
            }
        }

        private void readPairedEndFiles(Path[] readFiles) throws IOException {
            try (FastqReader reader1 = new FastqReader(readFiles[0]);
                 FastqReader reader2 = new FastqReader(readFiles[1])) {
                while (reader1.hasNext() && reader2.hasNext()) {
                    FastqRecord read1 = reader1.next();
                    FastqRecord read2 = reader2.next();
                    try {
                        readQueue.put(new FastqRecord[]{read1, read2});
                    } catch (InterruptedException ie) {
                        Pantools.logger.error("Error while adding reads to queue: {}", ie.getMessage());
                        throw new RuntimeException(ie);
                    }
                    readCount++;
                }
                pbm.updateReadCount(readCount);
            } finally {
                finishedReadingFromQueue = true;
            }
        }

        private void readSingleEndFile(Path readFile) throws IOException {
            try (FastqReader reader = new FastqReader(readFile)) {
                while (reader.hasNext()) {
                    FastqRecord read1 = reader.next();
                    try {
                        readQueue.put(new FastqRecord[]{read1});
                    } catch (InterruptedException ie) {
                        Pantools.logger.error("Error while adding reads to queue: {}", ie.getMessage());
                        throw new RuntimeException(ie);
                    }
                    readCount++;
                }
                pbm.updateReadCount(readCount);
            } finally {
                finishedReadingFromQueue = true;
            }
        }
    }

    private class fastqReadFilter implements Runnable {
        private final boolean pairedReads;
        private final List<FastqRecord[]> batch;

        public fastqReadFilter(boolean pairedReads) {
            this.pairedReads = pairedReads;
            this.batch = new ArrayList<>(BATCH_SIZE);
        }

        @Override
        public void run() {
            filterReadsForKmers(pairedReads);
        }

        private void filterReadsForKmers(boolean pairedReads) {
            try {
                ProgressBar pb2 = pbm.getFilteringBar();
                if (pairedReads) {
                    filterPairedEndReads(pb2);
                } else {
                    filterSingleEndReads(pb2);
                }
            } catch (InterruptedException e) {
                Pantools.logger.error("Error while filtering reads: {}", e.getMessage());
                throw new RuntimeException(e);
            }
        }

        private void filterPairedEndReads(ProgressBar pb) throws InterruptedException {
            try {
                while (!finishedReadingFromQueue || !readQueue.isEmpty()) {
                    // Batch processing
                    batch.clear();
                    readQueue.drainTo(batch, BATCH_SIZE);

                    if (batch.isEmpty()) {
                        FastqRecord[] reads = readQueue.poll(100, TimeUnit.MILLISECONDS);
                        if (reads != null) {
                            batch.add(reads);
                        }
                    }

                    if (!batch.isEmpty()) {
                        processBatch(batch, pb);
                    }
                }
            } finally {
                // Last filter thread marks writing as complete
                if (activeFilterThreads.decrementAndGet() == 0) {
                    finishedWritingToQueue = true;
                }
            }
        }

        private void processBatch(List<FastqRecord[]> batch, ProgressBar pb) throws InterruptedException {
            for (FastqRecord[] reads : batch) {
                preFilterCount.incrementAndGet();
                if (reads != null && (containsEnoughMatchingKmers(reads[0]) ||
                        containsEnoughMatchingKmers(reads[1]))) {
                    writeQueue.put(reads);
                    postFilterCount.incrementAndGet();
                }
                pb.step();
            }
        }

        private void filterSingleEndReads(ProgressBar pb) throws InterruptedException {
            try {
                while (!finishedReadingFromQueue || !readQueue.isEmpty()) {
                    FastqRecord[] reads = readQueue.poll(100, TimeUnit.MILLISECONDS);
                    if (reads != null && containsEnoughMatchingKmers(reads[0])) {
                        writeQueue.put(reads);
                    }
                    pb.step();
                }
            } finally {
                // Last filter thread marks writing as complete
                if (activeFilterThreads.decrementAndGet() == 0) {
                    finishedWritingToQueue = true;
                }
            }
        }

        /**
         * Main method to check if a read contains enough matching kmers.
         * Strategy here is to count how many k-mers in the read are part of the
         * allKmers list (all k-mers in a region) and how many are part of the
         * highFrequencyKmers list (high frequency k-mers in the same region).
         * If the number of informative (relevant) k-mers is above a certain
         * threshold, the read is considered to be informative.
         * @param read input read
         * @return true if the read contains enough matching kmers, false otherwise
         */
        private boolean containsEnoughMatchingKmers(FastqRecord read) {
            // Get all kmers at once to reduce method calls
            final long[] readKmers = Utils.kmerizeAs2bit(read.getReadString(), Globals.K_SIZE);

            // Count matching kmers
            int matchCount = 0;
            int highFrequencyCount = 0;

            for (long kmer : readKmers) {
                if (Collections.binarySearch(allKmers, kmer) >= 0) {
                    matchCount++;
                }

                // Check if we found a high frequency kmer for logging
                if (Collections.binarySearch(highFrequencyKmers, kmer) >= 0) {
                    highFrequencyCount++;
                }
            }

            // Return false if no matching kmers are found
            if (matchCount == 0) {
                return false;
            }

            // Log matching kmers
            Pantools.logger.trace("READSTATS\t{}\t{}\t{}\t{}\t{}\t{}\t{}", libraryName,
                    read.getReadName(), readKmers.length, MINIMUM_MATCHING_KMERS_FRAC * readKmers.length, matchCount, highFrequencyCount, matchCount - highFrequencyCount);

            // Return true if enough matching kmers are found
            return (matchCount - highFrequencyCount) >= MINIMUM_MATCHING_KMERS_FRAC * readKmers.length;
        }
    }

    private class fastqFileWriter implements Runnable {
        private final Path outputPath;
        private final String libraryName;
        private final boolean pairedReads;

        public fastqFileWriter(Path outputPath, String libraryName, boolean pairedReads) {
            this.outputPath = outputPath;
            this.libraryName = libraryName;
            this.pairedReads = pairedReads;
        }

        @Override
        public void run() {
            writeReadsToTempFile(outputPath, libraryName, pairedReads);
        }

        private void writeReadsToTempFile(Path outputPath, String libraryName, boolean pairedReads) {
            FastqWriterFactory writerFactory = new FastqWriterFactory();
            try {
                if (pairedReads) {
                    writePairedEndReads(outputPath, libraryName, writerFactory);
                } else {
                    writeSingleEndReads(outputPath, libraryName, writerFactory);
                }
            } catch (IOException | InterruptedException e) {
                Pantools.logger.error("Error while writing reads to temp file: {}", e.getMessage());
                throw new RuntimeException(e);
            }
        }

        private void writePairedEndReads(Path outputPath, String libraryName, FastqWriterFactory writerFactory) throws IOException, InterruptedException {
            int readCount = 0;

            Path temporaryReadsFile1 = outputPath.resolve(String.format("%s.matching_1.fq.gz", libraryName));
            Path temporaryReadsFile2 = outputPath.resolve(String.format("%s.matching_2.fq.gz", libraryName));
            try (FastqWriter writer1 = new AsyncFastqWriter(writerFactory.newWriter(temporaryReadsFile1.toFile()), QUEUE_CAPACITY);
                 FastqWriter writer2 = new AsyncFastqWriter(writerFactory.newWriter(temporaryReadsFile2.toFile()), QUEUE_CAPACITY)) {
                while (!finishedWritingToQueue || !writeQueue.isEmpty()) {
                    FastqRecord[] reads = writeQueue.poll(100, TimeUnit.MILLISECONDS);
                    if (reads != null) {
                        writer1.write(reads[0]);
                        writer2.write(reads[1]);
                        readCount++;
                    }
                }
            }

            Pantools.logger.debug("Wrote {} reads to temporary files.", readCount);
        }

        private void writeSingleEndReads(Path outputPath, String libraryName, FastqWriterFactory writerFactory) throws IOException, InterruptedException {
            int readCount = 0;

            Path temporaryReadsFile = outputPath.resolve(String.format("%s.matching.fq.gz", libraryName));
            try (FastqWriter writer = new AsyncFastqWriter(writerFactory.newWriter(temporaryReadsFile.toFile()), QUEUE_CAPACITY)) {
                while (!finishedWritingToQueue || !writeQueue.isEmpty()) {
                    FastqRecord[] reads = writeQueue.poll(100, TimeUnit.MILLISECONDS);
                    if (reads != null) {
                        writer.write(reads[0]);
                        readCount++;
                    }
                }
            }

            Pantools.logger.debug("Wrote {} reads to temporary file.", readCount);
        }
    }
}
