package nl.wur.bif.pantools.analysis.region_analysis;

import nl.wur.bif.pantools.Pantools;
import nl.wur.bif.pantools.utils.GraphUtils;
import nl.wur.bif.pantools.utils.graph.GenomeRange;
import nl.wur.bif.pantools.utils.graph.IndexGraphExtractor;
import nl.wur.bif.pantools.utils.graph.graph_objects.Graph;
import nl.wur.bif.pantools.utils.graph.graph_objects.GraphRange;
import nl.wur.bif.pantools.utils.writer.*;

import java.io.IOException;
import java.nio.file.Path;
import java.util.Arrays;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

/**
 * Extracts a region in one genome from the pangenome graph.
 * @author Dirk-Jan van Workum, Wageningen University, the Netherlands.
 */
public class RegionExtractor {
    private final Path outputPath;
    private GenomeRange genomeRange;

    /**
     * Constructor.
     *
     * @param outputPath The path to the output file.
     */
    public RegionExtractor(Path outputPath) {
        this.outputPath = outputPath;
    }

    /**
     * Sets a genome range.
     * Should be formatted as: genome#sequence:start-end.
     *
     * @param genomeRange The genome range.
     */
    public void setGenomeRange(GenomeRange genomeRange) {
        Pantools.logger.debug("Setting genome range '{}'.", genomeRange);
        this.genomeRange = genomeRange;
    }

    /**
     * Sets a gene name to be used for the range.
     *
     * @param gene    The gene name.
     * @param context The number of flanking genes to include.
     */
    public void setGene(String gene, int context) {
        Pantools.logger.debug("Setting gene '{}' with context of {} genes.", gene, context);
        this.genomeRange = GraphUtils.getGenomeRangeForGene(gene, context);
    }

    /**
     * Extracts a region from the pangenome graph by searching a query GenomeRange in the pangenome.
     * @param useKmers Whether to use k-mers for extraction.
     * @param useHomology Whether to use homology groups for extraction.
     * @param useBlast Whether to use BLAST results for extraction.
     * @param allowSinglePositions Whether to allow single positions as a hit.
     * @param searchSelf Whether to search in the input genome.
     * @param kmerDensity The k-mer density for k-mer sampling (distance between k-mers).
     * @param maxEvalue The maximum E-value for a BLAST hit.
     * @param maxDistance The maximum distance between two k-mers/transcripts to consider them as part of the same block.
     * @param fractPositions The fraction of positions to use for determining a hit.
     * @param maxFrequency The maximum frequency of a transcript or k-mer in a target genome.
     * @param flanking The number of flanking bases to include in the output.
     * @param writeAnnotations Whether to write the annotations in GFF3 format.
     * @param writeGfa Whether to write the GFA file.
     * @param writeHomology Whether to write the homology groups.
     * @param writeUnitigs Whether to write the unitig (node) presence/absence.
     * @param writeRepeats Whether to write the repeats.
     * @param readFiles            The read files to use for reconstruction.
     * @param filterHighFrequency  Whether to filter out high frequency for read reconstruction.
     * @param fast                 Whether to use the fast version of the algorithm (without obtaining all nodes).
     * @throws IOException If an IO error occurs during conversion.
     */
    public void extractRegion(final boolean useKmers, final boolean useHomology, final boolean useBlast, final boolean allowSinglePositions,
                              final boolean searchSelf, final int kmerDensity, final double maxEvalue, final int maxDistance, final double fractPositions, final int maxFrequency, final int flanking,
                              final boolean writeAnnotations, final boolean writeGfa, final boolean writeHomology, final boolean writeUnitigs, final boolean writeRepeats, final HashMap<String, Path[]> readFiles,
                              final boolean filterHighFrequency, final boolean fast)
            throws IOException {
        Pantools.logger.debug("Extracting region from pangenome for {}", this);
        Pantools.logger.debug("useKmers: {}; useHomology: {}; useBlast: {} allowSinglePositions: {}; searchSelf: {}; kmerDensity: {}; maxDistance: {}; fractPositions: {}; maxFrequency: {}; flanking: {}; writeAnnotations: {}; writeGfa: {}; writeHomology: {}; writeUnitigs: {}; writeRepeats: {}; readFiles: {}; fast: {}",
                useKmers, useHomology, useBlast, allowSinglePositions, searchSelf, kmerDensity, maxDistance, fractPositions, maxFrequency, flanking, writeAnnotations, writeGfa, writeHomology, writeUnitigs, writeRepeats, readFiles, fast);

        // make sure that genomeRange has been set
        if (genomeRange == null) {
            throw new IllegalStateException("No genome range has been set.");
        }

        // get the exact range
        final GraphRange graphRange = GraphRange.getRangeFromGenomeRange(genomeRange, useKmers, useHomology, useBlast, allowSinglePositions, searchSelf, kmerDensity, maxEvalue, maxDistance, fractPositions, maxFrequency, flanking, fast);
        Pantools.logger.debug("Range: {}", graphRange);

        // write the sequences + annotations
        writeRegion(writeAnnotations, writeGfa, writeHomology, writeUnitigs, writeRepeats, useBlast, graphRange);

        // reconstruct reads
        reconstructRegion(graphRange, readFiles, filterHighFrequency);
    }

    /**
     * Extracts multiple genes from the pangenome graph by searching for all homologues of the gene.
     *
     * @param genes            The gene names.
     * @param context          The number of flanking genes to search for.
     * @param flanking         The number of flanking bases to include in the output.
     * @param writeAnnotations Whether to write the annotations in GFF3 format.
     * @param writeGfa         Whether to write the GFA file.
     * @param writeHomology    Whether to write the homology groups.
     * @param writeUnitigs     Whether to write the unitig (node) presence/absence.
     * @param writeRepeats     Whether to write the repeats.
     * @param readFiles        The read files to use for reconstruction.
     * @param filterHighFrequency  Whether to filter out high frequency for read reconstruction.
     * @param fast             Whether to use the fast version of the algorithm (without obtaining all nodes).
     * @throws IOException If an IO error occurs during conversion.
     */
    public void extractGenes(String[] genes, int context, int flanking, boolean writeAnnotations, boolean writeGfa, boolean writeHomology, boolean writeUnitigs, boolean writeRepeats, final HashMap<String, Path[]> readFiles, boolean filterHighFrequency, boolean fast) throws IOException {
        if (genes == null || genes.length == 0) {
            Pantools.logger.error("No genes provided.");
            throw new IllegalArgumentException("No genes provided.");
        }

        Pantools.logger.debug("Extracting genes from pangenome for {} to {}", Arrays.toString(genes), this);
        Pantools.logger.debug("flanking: {}; writeAnnotations: {}; writeGfa: {}; writeHomology: {}; writeUnitigs: {}; writeRepeats: {}",
                flanking, writeAnnotations, writeGfa, writeHomology, writeUnitigs, writeRepeats);

        // get the exact range by just following homology groups
        final GraphRange graphRange = GraphRange.getRangeForGenes(genes, context, flanking, fast);
        Pantools.logger.debug("Range: {}", graphRange);

        // write the sequences + annotations
        writeRegion(writeAnnotations, writeGfa, writeHomology, writeUnitigs, writeRepeats, false, graphRange);

        // reconstruct reads
        reconstructRegion(graphRange, readFiles, filterHighFrequency);
    }

    /**
     * Extracts multiple homology groups from the pangenome graph.
     *
     * @param hmGroups         The homology group node identifiers.
     * @param flanking         The number of flanking bases to include in the output.
     * @param writeAnnotations Whether to write the annotations in GFF3 format.
     * @param writeGfa         Whether to write the GFA file.
     * @param writeHomology    Whether to write the homology groups.
     * @param writeUnitigs     Whether to write the unitig (node) presence/absence.
     * @param writeRepeats     Whether to write the repeats.
     * @param readFiles        The read files to use for reconstruction.
     * @param filterHighFrequency  Whether to filter out high frequency for read reconstruction.
     * @param fast             Whether to use the fast version of the algorithm (without obtaining all nodes).
     * @throws IOException If an IO error occurs during conversion.
     */
    public void extractHmGroups(List<Long> hmGroups, int flanking, boolean writeAnnotations, boolean writeGfa, boolean writeHomology, boolean writeUnitigs, boolean writeRepeats, final HashMap<String, Path[]> readFiles, boolean filterHighFrequency, boolean fast) throws IOException {
        if (hmGroups == null || hmGroups.isEmpty()) {
            Pantools.logger.error("No homology groups provided.");
            throw new IllegalArgumentException("No homology groups provided.");
        }

        Pantools.logger.debug("Extracting homology groups from pangenome for {} to {}", hmGroups, this);
        Pantools.logger.debug("flanking: {}; writeAnnotations: {}; writeGfa: {}; writeHomology: {}; writeUnitigs: {}; writeRepeats: {}",
                flanking, writeAnnotations, writeGfa, writeHomology, writeUnitigs, writeRepeats);

        // get the exact range by just following homology groups
        final GraphRange graphRange = GraphRange.getRangeForHmGroups(hmGroups, flanking, fast);
        Pantools.logger.debug("Range: {}", graphRange);

        // write the sequences + annotations
        writeRegion(writeAnnotations, writeGfa, writeHomology, writeUnitigs, writeRepeats, false, graphRange);

        // reconstruct reads
        reconstructRegion(graphRange, readFiles, filterHighFrequency);
    }

    /**
     * Extracts genes from the pangenome graph by searching for functional annotations.
     *
     * @param functions        The functional annotations.
     * @param flanking         The number of flanking bases to include in the output.
     * @param writeAnnotations Whether to write the annotations in GFF3 format.
     * @param writeGfa         Whether to write the GFA file.
     * @param writeHomology    Whether to write the homology groups.
     * @param writeUnitigs     Whether to write the unitig (node) presence/absence.
     * @param writeRepeats     Whether to write the repeats.
     * @param readFiles        The read files to use for reconstruction.
     * @param filterHighFrequency  Whether to filter out high frequency for read reconstruction.
     * @param fast             Whether to use the fast version of the algorithm (without obtaining all nodes).
     * @throws IOException If an IO error occurs during conversion.
     */
    public void extractFunctions(String[] functions, int flanking, boolean writeAnnotations, boolean writeGfa, boolean writeHomology, boolean writeUnitigs, boolean writeRepeats, final HashMap<String, Path[]> readFiles, boolean filterHighFrequency, boolean fast) throws IOException {
        if (functions == null || functions.length == 0) {
            Pantools.logger.error("No functions provided.");
            throw new IllegalArgumentException("No functions provided.");
        }

        Pantools.logger.debug("Extracting homology groups from pangenome for {} to {}", Arrays.toString(functions), this);
        Pantools.logger.debug("flanking: {}; writeAnnotations: {}; writeGfa: {}; writeHomology: {}; writeUnitigs: {}; writeRepeats: {}",
                flanking, writeAnnotations, writeGfa, writeHomology, writeUnitigs, writeRepeats);

        // get the exact range by just following homology groups
        final GraphRange graphRange = GraphRange.getRangeForFunctions(functions, flanking, fast);
        Pantools.logger.debug("Range: {}", graphRange);

        // write the sequences + annotations
        writeRegion(writeAnnotations, writeGfa, writeHomology, writeUnitigs, writeRepeats, false, graphRange);

        // reconstruct reads
        reconstructRegion(graphRange, readFiles, filterHighFrequency);
    }

    /**
     * Writes the region to the output path.
     *
     * @param writeAnnotations Whether to write the annotations in GFF3 format.
     * @param writeGfa Whether to write the GFA file.
     * @param writeHomology Whether to write the homology groups.
     * @param writeUnitigs Whether to write the unitig (node) presence/absence.
     * @param writeRepeats Whether to write the repeats.
     * @param writeBlast Whether to write the latest BLAST results.
     * @param graphRange The graph range to write.
     * @throws IOException If an IO error occurs during conversion.
     */
    private void writeRegion(boolean writeAnnotations, boolean writeGfa, boolean writeHomology, boolean writeUnitigs, boolean writeRepeats, boolean writeBlast, GraphRange graphRange) throws IOException {
        // write the sequences + annotations
        final FastaWriter fastaWriter = new FastaWriter();
        fastaWriter.write(graphRange, outputPath);
        if (writeAnnotations && GraphUtils.containsAnnotations()) {
            final AnnotationWriter annotationGffWriter = new AnnotationGffWriter();
            annotationGffWriter.write(graphRange, outputPath, writeRepeats, writeBlast);
        } else if (writeAnnotations) {
            Pantools.logger.warn("No annotations found in the pangenome.");
        } else {
            Pantools.logger.debug("Skipping annotation writing.");
        }

        // write a table with phasing information if available
        if (GraphUtils.containsPhasing()) {
            final PhasingWriter phasingWriter = new PhasingWriter();
            phasingWriter.write(graphRange, outputPath.resolve("phasing.tsv"));
        } else {
            Pantools.logger.debug("No phasing information found in the pangenome.");
        }

        // write the graph + annotations
        if (writeGfa) {
            final IndexGraphExtractor indexGraphExtractor = new IndexGraphExtractor();
            final Graph graph = indexGraphExtractor.extractNucleotideGraph(graphRange);
            final GfaWriter gfaWriter = new Gfa1Writer();
            gfaWriter.write(graph, outputPath.resolve("dbg.gfa"));
            if (GraphUtils.containsAnnotations()) {
                final AnnotationWriter annotationCsvWriter = new AnnotationCsvWriter();
                annotationCsvWriter.write(graphRange, outputPath.resolve("dbg_annotations.csv"), writeRepeats, false);
            }
        } else {
            Pantools.logger.debug("Skipping GFA writing.");
        }

        // write homology groups and presence/absence
        if (writeHomology && GraphUtils.containsHomologyGroups()) {
            // write collinearity
            final CollinearityWriter collinearityWriter = new CollinearityWriter();
            collinearityWriter.write(graphRange, outputPath, writeRepeats, writeBlast);

            // write gene presence/absence (occurrence of homology groups in graphRange)
            final PresenceAbsenceWriter presenceAbsenceWriter = new HomologyPresenceAbsenceWriter();
            presenceAbsenceWriter.write(graphRange, outputPath.resolve("hm_presence_absence.tsv"));
        } else if (writeHomology) {
            Pantools.logger.warn("No homology groups found in the pangenome.");
        } else {
            Pantools.logger.debug("Skipping homology writing.");
        }

        // write unitig presence/absence (occurrence of unitigs in graphRange)
        if (writeUnitigs) {
            final PresenceAbsenceWriter unitigPresenceAbsenceWriter = new UnitigPresenceAbsenceWriter();
            unitigPresenceAbsenceWriter.write(graphRange, outputPath.resolve("unitig_presence_absence.tsv"));
        } else {
            Pantools.logger.debug("Skipping unitig writing.");
        }
    }

    /**
     * Reconstructs reads from the pangenome graph if read files are provided.
     * @param graphRange The graph range to reconstruct.
     * @param readFiles  The read files to use for reconstruction.
     * @param filterHighFrequency Whether to filter out high frequency for read reconstruction.
     */
    private void reconstructRegion(GraphRange graphRange, HashMap<String, Path[]> readFiles, boolean filterHighFrequency) {
        // if there are read files, reconstruct their sequence for the region
        if (readFiles !=null && !readFiles.isEmpty()) {
            final ReadReconstructor readReconstructor = new ReadReconstructor(graphRange, filterHighFrequency);
            for (Map.Entry<String, Path[]> entry : readFiles.entrySet()) {
                final String libraryName = entry.getKey();
                final Path[] files = entry.getValue();
                final Path outputDirectory = outputPath.resolve(libraryName);
                readReconstructor.reconstruct(libraryName, files, outputDirectory);
            }
        }
    }

    /**
     * toString method.
     * @return A string representation of this object.
     */
    @Override
    public String toString() {
        return String.format("RegionExtractor{outputPath=%s, genomeRange=%s}", outputPath, genomeRange);
    }
}
