package nl.wur.bif.pantools.utils;

import me.tongfei.progressbar.ProgressBar;
import nl.wur.bif.pantools.Pantools;
import nl.wur.bif.pantools.analysis.classification.Classification;
import nl.wur.bif.pantools.construction.index.IndexDatabase;
import nl.wur.bif.pantools.construction.index.IndexPointer;
import nl.wur.bif.pantools.construction.index.kmer;
import nl.wur.bif.pantools.construction.sequence.SequenceDatabase;
import nl.wur.bif.pantools.utils.graph.GenomeRange;
import nl.wur.bif.pantools.utils.graph.graph_objects.GraphRange;
import org.codehaus.plexus.util.StringUtils;
import org.neo4j.graphdb.*;
import org.neo4j.graphdb.factory.GraphDatabaseFactory;
import org.neo4j.graphdb.factory.GraphDatabaseSettings;
import org.neo4j.unsafe.impl.batchimport.input.DataException;

import java.io.BufferedWriter;
import java.io.File;
import java.io.IOException;
import java.nio.file.Files;
import java.nio.file.Path;
import java.util.*;
import java.nio.file.Paths;
import java.util.stream.Collectors;
import java.util.stream.Stream;

import static nl.wur.bif.pantools.construction.build_pangenome.GenomeLayer.locate;
import static nl.wur.bif.pantools.utils.Globals.*;
import static nl.wur.bif.pantools.utils.Utils.check_current_grouping_version;
import static nl.wur.bif.pantools.utils.Utils.connect_pangenome;

/**
 * Create, load and validate neo4j graph databases
 *
 * @author Robin van Esch
 */
public final class GraphUtils {

    private GraphUtils() {}

    /**
     * Create a new graph database service for the Neo4j database.
     * A shutdown hook is configured to make sure the database is always shut down correctly.
     *
     * @param databaseDirectory path to the pangenome database directory
     */
    public static void createGraphDatabaseService(Path databaseDirectory) {
        final Path graphDatabasePath = databaseDirectory.resolve("databases").resolve("graph.db");
        GRAPH_DB = new GraphDatabaseFactory()
                .newEmbeddedDatabaseBuilder(graphDatabasePath.toFile())
                .setConfig(GraphDatabaseSettings.keep_logical_logs, "4 files")
                .newGraphDatabase();

        Runtime.getRuntime().addShutdownHook(new Thread(GRAPH_DB::shutdown));
    }

    /**
     * Verify if the database contains a pangenome or a panproteome and set type-specific parameters.
     */
    public static void setDatabaseParameters() {
        try (Transaction ignored = GRAPH_DB.beginTx();
             ResourceIterator<Node> pangenomeNodes = GRAPH_DB.findNodes(PANGENOME_LABEL)) {
            final Node pangenomeNode = pangenomeNodes.next();
            total_genomes = (int) pangenomeNode.getProperty("num_genomes");
            if (pangenomeNode.hasProperty("k_mer_size")) {
                K_SIZE = (int) pangenomeNode.getProperty("k_mer_size");
                connect_pangenome();
            } else {
                PROTEOME = true;
            }
            if (pangenomeNodes.hasNext()) {
                throw new DataException("Database contains more than one pangenome and can only contain one");
            }
        } catch (NotFoundException nfe) {
            throw new NotFoundException("Unable to find the pangenome node in the graph database");
        }
    }

    /**
     * Creates and connects to genome, index and graph databases of the pangenome.
     *
     * @param databaseDirectory path to the pangenome database directory
     * @throws IOException when directories cannot be created in the database directory
     */
    public static void createPangenomeDatabase(Path databaseDirectory) throws IOException {
        createGraphDatabaseService(databaseDirectory);
        GENOME_DB = new SequenceDatabase(WORKING_DIRECTORY + GENOME_DATABASE_PATH, PATH_TO_THE_GENOMES_FILE);
        INDEX_DB = new IndexDatabase(WORKING_DIRECTORY + INDEX_DATABASE_PATH, PATH_TO_THE_GENOMES_FILE, GENOME_DB, K_SIZE);
        Files.createDirectories(databaseDirectory.resolve("databases").resolve("genome.db").resolve("genomes"));
        Files.createDirectory(databaseDirectory.resolve("log")); //TODO: remove
    }

    /**
     * Creates and connects to graph databases of the panproteome.
     *
     * @param databaseDirectory path to the pangenome database directory
     * @throws IOException when directories cannot be created in the database directory
     */
    public static void createPanproteomeDatabase(Path databaseDirectory) throws IOException {
        createGraphDatabaseService(databaseDirectory);
        Files.createDirectory(databaseDirectory.resolve("proteins"));
        Files.createDirectory(databaseDirectory.resolve("log")); //TODO: remove
    }

    /**
     * Validate that the database is a pangenome or a panproteome if one is required.
     *
     * @param type database type; panproteome or pangenome
     */
    public static void validateGraphType(String type) {
        final String databaseType;
        try (Transaction ignored = GRAPH_DB.beginTx();
             ResourceIterator<Node> pangenomeNodes = GRAPH_DB.findNodes(PANGENOME_LABEL)) {
            final Node pangenomeNode = pangenomeNodes.next();
            if (pangenomeNode.hasProperty("k_mer_size")) {
                databaseType = "pangenome";
            } else {
                databaseType = "panproteome";
            }
        }
        if (!databaseType.equals(type)) {
            throw new DataException(
                    String.format("%s database detected; this function is for %ss only",
                            StringUtils.capitalise(databaseType), type));
        }
    }

    /**
     * Check if the pangenome has genome nodes
     */
    public static void validateGenomes() {
        try (Transaction ignored = GRAPH_DB.beginTx();
             ResourceIterator<Node> genomeNodes = GRAPH_DB.findNodes(GENOME_LABEL)) {
            if (!genomeNodes.hasNext()) {
                Pantools.logger.error("The pangenome does not contain genomes.");
                throw new RuntimeException("Invalid graph structure");
            }
        } catch (NotFoundException nfe) {
            Pantools.logger.error("Could not find the genome nodes in the graph database.");
            throw new RuntimeException("Invalid graph structure", nfe);
        }
    }

    /**
     * Check if the pangenome has phasing information on (some of) the sequence nodes
     */
    public static boolean containsPhasing() {
        try (Transaction ignored = GRAPH_DB.beginTx();
             ResourceIterator<Node> sequenceNodes = GRAPH_DB.findNodes(SEQUENCE_LABEL)) {
            while (sequenceNodes.hasNext()) {
                Node sequenceNode = sequenceNodes.next();
                if (sequenceNode.hasProperty("phasing_ID")) {
                    return true;
                }
            }
        } catch (NotFoundException nfe) {
            Pantools.logger.error("Could not find the phasing nodes in the graph database.");
            throw new RuntimeException("Invalid graph structure", nfe);
        }
        return false;
    }

    /**
     * check if pangenome contains annotation and if not throw an exception
     */
    public static void validateAnnotations() {
        try (Transaction ignored = GRAPH_DB.beginTx();
             ResourceIterator<Node> annotationNodes = GRAPH_DB.findNodes(ANNOTATION_LABEL)) {
            if (!annotationNodes.hasNext()) {
                Pantools.logger.error("The pangenome does not contain annotations. Please run add_annotations first.");
                throw new RuntimeException("Invalid graph structure");
            }
        } catch (NotFoundException nfe) {
            Pantools.logger.error("Could not find the annotation nodes in the graph database.");
            throw new RuntimeException("Invalid graph structure", nfe);
        }
    }

    /**
     * check if pangenome contains annotations and return as boolean
     */
    public static boolean containsAnnotations() {
        try (Transaction ignored = GRAPH_DB.beginTx();
             ResourceIterator<Node> annotationNodes = GRAPH_DB.findNodes(ANNOTATION_LABEL)) {
            if (annotationNodes.hasNext()) {
                return true;
            }
        } catch (NotFoundException ignored) {
        }
        return false;
    }

    /**
     * check if pangenome/panproteome contains homology groups and if not throw an exception
     */
    public static void validateHomologyGroups() {
        try (Transaction ignored = GRAPH_DB.beginTx();
             ResourceIterator<Node> homologyGroupNodes = GRAPH_DB.findNodes(HOMOLOGY_GROUP_LABEL)) {
            if (!homologyGroupNodes.hasNext()) {
                Pantools.logger.error("The pangenome/panproteome does not contain homology groups. Please run group first.");
                throw new RuntimeException("Invalid graph structure");
            }
        } catch (NotFoundException nfe) {
            Pantools.logger.error("Could not find the homology group nodes in the graph database.");
            throw new DataException("Invalid graph structure");
        }
    }

    /**
     * check if pangenome contains homology groups and return as boolean
     */
    public static boolean containsHomologyGroups() {
        try (Transaction ignored = GRAPH_DB.beginTx();
             ResourceIterator<Node> homologyGroupNodes = GRAPH_DB.findNodes(HOMOLOGY_GROUP_LABEL)) {
            if (homologyGroupNodes.hasNext()) {
                return true;
            }
        } catch (NotFoundException ignored) {
        }
        return false;
    }

    /**
     * check if pangenome contains repeat nodes and if not throw an exception
     */
    public static void validateRepeats() {
        try (Transaction ignored = GRAPH_DB.beginTx();
             ResourceIterator<Node> repeatNodes = GRAPH_DB.findNodes(REPEAT_LABEL)) {
            if (!repeatNodes.hasNext()) {
                Pantools.logger.error("The pangenome does not contain repeats. Please run add_repeats first.");
                throw new RuntimeException("Invalid graph structure");
            }
        } catch (NotFoundException nfe) {
            Pantools.logger.error("Could not find the repeat nodes in the graph database.");
            throw new RuntimeException("Invalid graph structure", nfe);
        }
    }

    /**
     * check if pangenome contains repeat nodes and return as boolean
     */
    public static boolean containsRepeats() {
        try (Transaction ignored = GRAPH_DB.beginTx();
             ResourceIterator<Node> repeatNodes = GRAPH_DB.findNodes(REPEAT_LABEL)) {
            if (repeatNodes.hasNext()) {
                return true;
            }
        } catch (NotFoundException ignored) {
        }
        return false;
    }

    /**
     * check if pangenome contains accession information and if not throw an exception
     */
    public static void validateAccessions() {
        try (Transaction ignored = GRAPH_DB.beginTx();
             ResourceIterator<Node> accessionNodes = GRAPH_DB.findNodes(ACCESSION_LABEL)) {
            if (!accessionNodes.hasNext()) {
                throw new RuntimeException("Invalid graph structure");
            }
        } catch (NotFoundException nfe) {
            Pantools.logger.error("Could not find the accession nodes in the graph database.");
            throw new RuntimeException("Invalid graph structure", nfe);
        }
    }

    /**
     * check if pangenome contains PAV/VCF accession information and if not throw an exception
     *
     * @param type type of accession information (PAV/VCF)
     */
    public static void validateAccessions(String type) {
        assert (type.matches("PAV|VCF"));
        try (Transaction ignored = GRAPH_DB.beginTx();
             ResourceIterator<Node> accessionNodes = GRAPH_DB.findNodes(ACCESSION_LABEL, type, true)) {
            if (!accessionNodes.hasNext()) {
                Pantools.logger.error("The pangenome does not contain {} information.", type);
                throw new RuntimeException("Invalid graph structure");
            }
        } catch (NotFoundException nfe) {
            Pantools.logger.error("Could not find {} nodes in the graph database.", type);
            throw new RuntimeException("Invalid graph structure", nfe);
        }
    }

    /**
     * Remove a node and ints relationships from the graph database.
     *
     * @param node node to remove
     */
    public static void removeNode(Node node) { // TODO: make general neo4j functions
        for (Relationship relationship : node.getRelationships()) {
            relationship.delete();
        }
        node.delete();
    }

    /**
     * Checks if this pangenome contains accessions (variants in VCF format) that were added with add_variants.
     *
     * @return true if the pangenome contains accessions, false otherwise
     */
    public static boolean containsVariantInformation() {
        final boolean containsVcf;
        try (Transaction ignored = GRAPH_DB.beginTx();
             ResourceIterator<Node> accessionNodes = GRAPH_DB.findNodes(ACCESSION_LABEL, "VCF", true)) {
            containsVcf = accessionNodes.hasNext();
        }
        return containsVcf;
    }

    /**
     * Checks if the database contains alignment output for the given alignment mode.
     * @param alignmentMethod alignment method to check (per-group, multiple-groups, regions, functions)
     * @param alignmentModeShort alignment mode to check (nuc, prot, var)
     * @param trimming whether the alignment has to be trimmed
     * @return true if the database contains alignment output, false otherwise
     */
    public static boolean containsAlignmentOutput(String alignmentMethod, String alignmentModeShort, boolean trimming) {
        if (!alignmentMethod.matches("per-group|multiple-groups|regions|functions")) {
            throw new IllegalArgumentException("Invalid alignment method");
        }
        if (!alignmentModeShort.matches("nuc|prot|var")) {
            throw new IllegalArgumentException("Invalid alignment mode");
        }

        final String fastaFileName = alignmentModeShort + (trimming ? "_trimmed" : "") + ".fasta";
        Path alignmentDirectory;
        boolean containsAlignmentOutput = false;
        long numberOfAlignments = 0;

        // check grouping version
        try (Transaction ignored = GRAPH_DB.beginTx()) {
            check_current_grouping_version();
        }

        // set the alignment directory based on the alignment method
        switch (alignmentMethod) {
            case "per-group":
            case "multiple-groups":
                alignmentDirectory = Paths.get(WORKING_DIRECTORY)
                        .resolve("alignments")
                        .resolve(alignmentModeShort.equals("var") ?
                                "msa_" + alignmentMethod.replace("-", "_") + "_var" :
                                "msa_" + alignmentMethod.replace("-", "_"))
                        .resolve("grouping_v" + grouping_version);
                break;
            case "regions":
                alignmentDirectory = Paths.get(WORKING_DIRECTORY)
                        .resolve("alignments")
                        .resolve("msa_regions");
                break;
            case "functions":
                alignmentDirectory = Paths.get(WORKING_DIRECTORY)
                        .resolve("alignments")
                        .resolve("msa_functions");
                break;
            default:
                throw new NoSuchElementException("Invalid alignment method");
        }

        // check if the alignment directory exists and count the number of alignments
        if (Files.exists(alignmentDirectory)) {
            Pantools.logger.debug("Found alignment directory {}", alignmentDirectory);
            containsAlignmentOutput = true;
            try (Stream<Path> paths = Files.list(alignmentDirectory)) {
                numberOfAlignments = paths
                        .filter(Files::isDirectory)
                        .filter(path -> {
                            final File file = path.resolve("output").resolve(fastaFileName).toFile();
                            return file.exists() && file.length() > 0;
                        })
                        .count();
            } catch (IOException e) {
                Pantools.logger.error("Unable to list directories in {}", alignmentDirectory);
                throw new RuntimeException("Unable to list directories");
            }
        } else {
            Pantools.logger.debug("Alignment directory {} does not exist", alignmentDirectory);
        }

        // if no alignments were found, the database does not contain correct alignment output
        if (numberOfAlignments == 0) {
            containsAlignmentOutput = false;
        }

        // log the number of alignments that were found
        Pantools.logger.debug("Found {} alignments for alignment method {} and alignment mode '{}'",
                numberOfAlignments, alignmentMethod, alignmentModeShort);

        // return whether the database contains alignment output
        return containsAlignmentOutput;
    }

    /**
     * Obtain all incoming edges of a node that have a property named after the content of origin.
     * @param node the nucleotide node to get the incoming edges from
     * @param genomeNr the genome number to search for
     * @return a HashSet containing all incoming edges that have a property named after the content of origin
     * NB: assumes within transaction!
     */
    public static HashSet<Relationship> getIncomingEdges(Node node, int genomeNr) {
        final HashSet<Relationship> incomingEdges = new HashSet<>();
        final String startOriginString = String.format("G%dS", genomeNr);

        for (Relationship incomingEdge : node.getRelationships(Direction.INCOMING)) {
            for (String propertyKey : incomingEdge.getPropertyKeys()) {
                if (propertyKey.startsWith(startOriginString)) {
                    incomingEdges.add(incomingEdge);
                }
            }
        }

        Pantools.logger.trace("Found incoming edges: {} (length: {})",
                Arrays.toString(incomingEdges.toArray()),
                incomingEdges.size());
        return incomingEdges;
    }

    /**
     * Get the number of each sequence for a given genome based on its ID.
     * @param genome the genome number to get the sequence numbers for
     * @return a hashmap with the sequence id as key and the number as value
     */
    public static HashMap<String, Integer> getSequenceNumbers(int genome) {
        HashMap<String, Integer> sequenceNumbers = new HashMap<>();

        int numSequences;
        try {
            numSequences = GENOME_DB.num_sequences[genome];
        } catch (ArrayIndexOutOfBoundsException aioobe) {
            Pantools.logger.error("Could not find genome {}", genome);
            throw new RuntimeException("Invalid genome number");
        }

        for (int i = 1; i <= numSequences; i++) {
            try {
                String sequence = GENOME_DB.sequence_titles[genome][i];
                String[] fields = sequence.split("\\s+");
                String id = fields[0];
                sequenceNumbers.put(id, i);
            } catch (ArrayIndexOutOfBoundsException | NullPointerException e) {
                Pantools.logger.error("Could not find sequence for genome {} and sequence number {}.", genome, i);
                throw new RuntimeException("Invalid sequence number");
            }
        }

        // print the number of each sequence with Pantools.logger
        Pantools.logger.trace("Sequence numbers for genome {}: ", genome);
        for (String id : sequenceNumbers.keySet()) {
            Pantools.logger.trace("\t{}: {}", id, sequenceNumbers.get(id));
        }

        return sequenceNumbers;
    }

    /**
     * Gets the name of the mrna node, if it has one
     * @param mrnaNode node to get name from
     * @return name of mrna node
     * NB: has to be run within a transaction
     * NB: name used to be stored as a string, nowadays as a string array (keeping old style for backwards compatibility)
     */
    public static String getMrnaNodeName(Node mrnaNode) {
        assert mrnaNode.hasLabel(MRNA_LABEL);

        String name = "-";
        if (mrnaNode.hasProperty("name")) {
            if (mrnaNode.getProperty("name") instanceof String) { // old style
                name = (String) mrnaNode.getProperty("name");
            } else if (mrnaNode.getProperty("name") instanceof String[]) { // new style
                String[] nameArray = (String[])(mrnaNode.getProperty("name"));
                name = Arrays.toString(nameArray).replace("[","").replace("]","").replace(",","/").replace(" ","");
            } else {
                throw new RuntimeException("Unexpected type for property 'name' of node " + mrnaNode.getId());
            }
        }

        return name;
    }

    /**
     * Gets the node identifier of a given gene name. It searches (in this order)
     * in the "id" property of all gene nodes, then in the "name" property of all
     * gene nodes and finally in the "locus_tag" property of all gene nodes.
     * NB: it has to be an exact match
     * @param geneName gene name to get node identifier for
     * @return node identifier of gene name
     */
    public static long getGeneNodeId(String geneName) {
        try (Transaction ignored = GRAPH_DB.beginTx();
             ResourceIterator<Node> geneNodes = GRAPH_DB.findNodes(GENE_LABEL)) {
            while (geneNodes.hasNext()) {
                Node geneNode = geneNodes.next();
                if (geneNode.hasProperty("id")) { //contains String value
                    if (geneNode.getProperty("id").equals(geneName)) {
                        return geneNode.getId();
                    }
                }
                if (geneNode.hasProperty("name")) { //contains String[] value
                    String[] nameArray = (String[])(geneNode.getProperty("name"));
                    for (String name : nameArray) {
                        if (name.equals(geneName)) {
                            return geneNode.getId();
                        }
                    }
                }
                if (geneNode.hasProperty("locus_tag")) { //contains String[] value
                    String[] locusTagArray = (String[])(geneNode.getProperty("locus_tag"));
                    for (String locusTag : locusTagArray) {
                        if (locusTag.equals(geneName)) {
                            return geneNode.getId();
                        }
                    }
                }
            }
        }

        Pantools.logger.error("Could not find gene node with name {}", geneName);
        throw new RuntimeException("Could not find gene node");
    }

    /**
     * Gets all accession identifiers connected to a genome node, if it has any
     * @param genomeNr genome number to get accessions from
     * @return set of accession identifiers
     */
    public static Set<String> getVcfAccessions(int genomeNr) {
        Set<String> accessions = new HashSet<>();
        try (Transaction ignored = GRAPH_DB.beginTx();
             ResourceIterator<Node> genomeNodes = GRAPH_DB.findNodes(GENOME_LABEL, "number", genomeNr)) {
            if (genomeNodes.hasNext()) {
                Node genomeNode = genomeNodes.next();
                for (Relationship accessionRelationship : genomeNode.getRelationships(RelTypes.has)) {
                    Node accessionNode = accessionRelationship.getEndNode();
                    if (accessionNode.hasLabel(ACCESSION_LABEL) && accessionNode.hasProperty("VCF")) {
                        String sample = (String) accessionNode.getProperty("sample");
                        if ((boolean) accessionNode.getProperty("VCF")) {
                            accessions.add(String.format("%s|%s", genomeNr, sample));
                        } else {
                            Pantools.logger.trace("Skipping non-VCF accession {}", sample);
                        }
                    }
                }
            }
        }
        return accessions;
    }

    /**
     * Gets the genome range for a given gene
     * @param gene gene name to get genome range for
     * @param context number of flanking genes to include
     * @return genome range for gene
     */
    public static GenomeRange getGenomeRangeForGene(String gene, final int context) {
        // get the gene address
        final int[] geneAddress;
        try (Transaction ignored = GRAPH_DB.beginTx()) {
            Node geneNode = GRAPH_DB.getNodeById(getGeneNodeId(gene));
            geneAddress = (int[]) geneNode.getProperty("address");
        } catch (NotFoundException nfe) {
            Pantools.logger.error("Could not find gene node with name {}", gene);
            throw new RuntimeException("Could not find gene node");
        }

        // return the gene address if no context is required
        if (context == 0) {
            return new GenomeRange(geneAddress, true); //check whether the genome range is valid
        }

        // extend by context on same genome, same sequence
        final int genomeNr = geneAddress[0];
        final int sequenceNr = geneAddress[1];
        final Set<Integer[]> genes = new TreeSet<>(Comparator.comparing(o -> o[0])); // will contain start and stop locations of genes, ordered by start position
        try (Transaction ignored = GRAPH_DB.beginTx();
             ResourceIterator<Node> geneNodes = GRAPH_DB.findNodes(GENE_LABEL, "genome", genomeNr, "sequence", sequenceNr)) {
            if (geneNodes.hasNext()) {
                while (geneNodes.hasNext()) {
                    final Node geneNode = geneNodes.next();
                    final int[] address = (int[]) geneNode.getProperty("address");
                    genes.add(new Integer[]{address[2], address[3]});
                }
            }
        }
        Pantools.logger.debug("Found {} genes in genome {} sequence {}.", genes.size(), genomeNr, sequenceNr);
        Pantools.logger.trace("Genes: {}", genes);

        // we now have the gene address and the ordered gene locations, so we can find the context
        final int originalGeneIndex = genes.stream().map(o -> o[0]).collect(Collectors.toList()).indexOf(geneAddress[2]);
        final int firstGeneIndex = Math.max(0, originalGeneIndex - context);
        final int lastGeneIndex = Math.min(genes.size() - 1, originalGeneIndex + context);
        final int adjustedStart = genes.stream().skip(firstGeneIndex).findFirst().orElse(new Integer[]{0, 0})[0];
        final int adjustedEnd = genes.stream().skip(lastGeneIndex).findFirst().orElse(new Integer[]{0, 0})[1];
        Pantools.logger.debug("Adjusted original range {} to {}",
                new GenomeRange(geneAddress, false),
                new GenomeRange(genomeNr, sequenceNr, adjustedStart, adjustedEnd, false)); //don't validate genome range for logging

        // return the genome range
        return new GenomeRange(genomeNr, sequenceNr, adjustedStart, adjustedEnd, true); //validate genome range because it is used for further processing
    }

    /**
     * Get all genome ranges for a function
     * @param function the function to search for
     * @return a set of the genome ranges of all mRNAs that contain the function
     */
    public static Set<GenomeRange> getGenomeRangeForFunction(final String function) {
        Set<GenomeRange> genomeRanges = new HashSet<>();

        SELECTED_NAME = function; //to let find_functional_annotation_node work
        try (Transaction ignored = GRAPH_DB.beginTx()) {
            int functionCounter = 0;
            for (Node functionNode : Classification.find_functional_annotation_node()) {
                RelationshipType relType = null;
                if (functionNode.hasLabel(Globals.GO_LABEL)) {
                    relType = RelTypes.has_go;
                } else if (functionNode.hasLabel(Globals.PFAM_LABEL)) {
                    relType = RelTypes.has_pfam;
                } else if (functionNode.hasLabel(Globals.INTERPRO_LABEL)) {
                    relType = RelTypes.has_interpro;
                } else if (functionNode.hasLabel(Globals.TIGRFAM_LABEL)) {
                    relType = RelTypes.has_tigrfam;
                } else {
                    Pantools.logger.error("Did not recognise function type: {}", relType);
                    throw new IllegalArgumentException("Unknown function type");
                }

                for (Relationship rel : functionNode.getRelationships(relType)) {
                    Node mrnaNode = rel.getStartNode();
                    genomeRanges.add(new GenomeRange((int[]) mrnaNode.getProperty("address"), false));
                }
                functionCounter++;
            }
            if (functionCounter == 0) {
                Pantools.logger.error("Could not find function {}", function);
                throw new RuntimeException("Invalid graph structure");
            }
        } catch (NotFoundException nfe) {
            Pantools.logger.error("Could not find the function nodes in the graph database.");
            throw new RuntimeException("Invalid graph structure", nfe);
        }

        return genomeRanges;
    }

    /**
     * Gets all mRNAs that overlap with (part of) a given genome range
     * @param genomeRange genome range to get mRNAs for
     * @return set of mRNA nodes
     * NB: assumes within transaction!
     */
    public static HashSet<Node> getMrnasForGenomeRange(GenomeRange genomeRange) {
        Pantools.logger.debug("Searching for mRNAs for genome range {}.", genomeRange);
        HashSet<Node> mrnas = new HashSet<>();

        ResourceIterator<Node> mrnaNodesOnOrigin = GRAPH_DB.findNodes(MRNA_LABEL,
                "genome", genomeRange.getGenomeNr(),
                "sequence", genomeRange.getSequenceNr());
        while (mrnaNodesOnOrigin.hasNext()) {
            Node mrnaNode = mrnaNodesOnOrigin.next();
            int[] mrnaAddress = (int[]) mrnaNode.getProperty("address");
            if (mrnaAddress[3] >= genomeRange.getStart() && mrnaAddress[2] <= genomeRange.getEnd()) {
                mrnas.add(mrnaNode);
            }
        }

        Pantools.logger.debug("Found {} mRNAs ({}) for genome range {}.", mrnas.size(), mrnas, genomeRange);
        return mrnas;
    }

    /**
     * Gets all homology groups to which all mRNAs in a given genome range belong
     * @param genomeRange genome range to get homology groups for
     * @return set of homology group nodes
     * NB: assumes within transaction!
     */
    public static HashSet<Node> getHomologyGroupsForGenomeRange(GenomeRange genomeRange) {
        Pantools.logger.debug("Searching for homology groups for genome range {}.", genomeRange);
        HashSet<Node> homologyGroups = new HashSet<>();

        for (Node mrnaNode : getMrnasForGenomeRange(genomeRange)) {
            for (Relationship mrnaHomologyGroupRelationship : mrnaNode.getRelationships(RelTypes.has_homolog)) { // homology_group -[has_homolog]-> mrna
                Node homologyGroupNode = mrnaHomologyGroupRelationship.getStartNode();
                if (homologyGroupNode.hasLabel(HOMOLOGY_GROUP_LABEL)) {
                    homologyGroups.add(homologyGroupNode);
                }
            }
        }

        Pantools.logger.debug("Found {} homology groups ({}) for genome range {}.", homologyGroups.size(), homologyGroups, genomeRange);
        return homologyGroups;
    }


    /**
     * Get a list of node IDs for a given graph range
     * @param genomeRange genome range to get node IDs for
     * @return list of node IDs
     */
    public static List<Long> getNodeIdsForGenomeRange(GenomeRange genomeRange) {
        Pantools.logger.debug("Locating nodes for genome range {}", genomeRange);
        final List<Long> nodeIds = new ArrayList<>();
        try (ProgressBar pb = Utils.createProgressBar("Locating nodes", genomeRange.getEnd() - genomeRange.getStart() + 1)) {
            for (int position = genomeRange.getStart() - 1; position < genomeRange.getEnd(); position++) {
                final IndexPointer nodePointer = locate(GRAPH_DB,
                        GENOME_SC,
                        INDEX_SC,
                        genomeRange.getGenomeNr(),
                        genomeRange.getSequenceNr(),
                        position);
                nodeIds.add(nodePointer.node_id);
                pb.step();
            }
        }
        Pantools.logger.debug("Found {} nodes for genome range {}.", nodeIds.size(), genomeRange);
        return nodeIds;
    }

    /**
     * Obtain a set of positions for a given set of homology groups in a given genome below a certain occurrence frequency
     * @param homologyGroups set of homology groups to search for occurrences in
     * @param currentGenomeNr genome number to search for occurrences in
     * @param maxFrequency maximum frequency of the mRNAs in the target genome
     * @return set of positions (sequenceNr, start, end) for the mRNAs in the homology groups
     * NB: assumes within transaction!
     */
    public static Set<Integer[]> getHomologyGroupsBelowFrequency(final Set<Node> homologyGroups, final int currentGenomeNr, final int maxFrequency) {
        final Set<Integer[]> positions = new ArrayHashSet<>();
        for (Node homologyGroup : homologyGroups) {
            Pantools.logger.trace("Searching for positions for homology group {}.", homologyGroup.getId());

            // Check if the homology group has a copy number of 1 for the current genome
            final int[] copyNumber = (int[]) homologyGroup.getProperty("copy_number");
            if (copyNumber[currentGenomeNr] > maxFrequency) {
                Pantools.logger.trace("Homology group {} has a frequency of {} for genome {}.", homologyGroup.getId(), copyNumber[currentGenomeNr], currentGenomeNr);
                continue;
            }

            // Obtain the position for the mRNA in the homology group
            for (Relationship homologyGroupMrnaRelationship : homologyGroup.getRelationships(RelTypes.has_homolog)) {
                final Node mrnaNode = homologyGroupMrnaRelationship.getEndNode();
                final int[] address = (int[]) mrnaNode.getProperty("address"); // [genome, sequence, start, end]
                if (address[0] == currentGenomeNr) {
                    Pantools.logger.trace("Found mRNA {} to be in genome {}.", mrnaNode.getProperty("id"), currentGenomeNr);
                    final int sequenceNr = address[1];
                    final int start = address[2];
                    final int end = address[3];
                    Pantools.logger.trace("Adding position (sequence: {}, start: {}, end: {}).", sequenceNr, start, end);
                    positions.add(new Integer[]{sequenceNr, start, end});
                }
            }
        }
        Pantools.logger.debug("Found {} homology groups for genome {}", positions.size(), currentGenomeNr);
        return positions;
    }

    /**
     * Obtain a set of positions for a given set of kmers in a given genome
     * @param kmers map of kmers to their position in an original sequence
     * @param genomeNr genome number to search for occurrences in
     * @param maxFrequency maximum frequency of the kmer in the pangenome
     * @return set of positions (sequenceNr, position)
     * NB: assumes within transaction!
     */
    public static Set<Integer[]> getKmersBelowFrequency(final Map<kmer, Integer> kmers, final int genomeNr, final int maxFrequency) {
        final HashSet<Node> seenNodes = new HashSet<>();
        final Set<Integer[]> positions = new ArrayHashSet<>();
        for (kmer kmer : kmers.keySet()) {
            final int originalPosition = kmers.get(kmer);
            Pantools.logger.trace("Searching for positions for kmer {} at genome {}, position {}", kmer, genomeNr, originalPosition);

            final IndexPointer pointer = new IndexPointer();
            INDEX_SC.get_pointer(pointer, INDEX_SC.find(kmer));
            final Node node = GRAPH_DB.getNodeById(pointer.node_id);
            if (seenNodes.contains(node)) {
                Pantools.logger.debug("Skipping node {} (kmer: {}) as it has already been seen.", node.getId(), kmer);
                continue;
            }
            Pantools.logger.trace("Pointer ({}) for kmer {} points to node {}", pointer, kmer, node.getId());
            if (node.hasProperty("frequencies")) {
                final long[] frequencies = (long[]) node.getProperty("frequencies");
                if (frequencies[genomeNr] <= maxFrequency) {
                    Pantools.logger.trace("Found node {} to have a frequency of {} for genome {}", node.getId(), frequencies[genomeNr], genomeNr);
                    for (Relationship incomingEdge : getIncomingEdges(node, genomeNr)) {
                        for (String origin : incomingEdge.getPropertyKeys()) {
                            if (origin.startsWith("G" + genomeNr + "S")) {
                                final int sequenceNr = Utils.getSequenceNrFromOrigin(origin);
                                for (int position : getKmerPositions(pointer, origin, node, incomingEdge, kmer.get_canonical())) {
                                    Pantools.logger.trace("Adding position (sequence: {}, position: {}).", sequenceNr, position + 1);
                                    positions.add(new Integer[]{sequenceNr, position + 1});
                                }
                            }
                        }
                    }
                }
            }
            seenNodes.add(node);
        }
        Pantools.logger.debug("Found {} kmers for genome {}", positions.size(), genomeNr);
        return positions;
    }

    /**
     * An inner class to create a HashSet that checks the contents of arrays for equality instead of the array reference
     * @param <T> type of the HashSet
     */
    private static class ArrayHashSet<T> extends HashSet<T> {
        @Override
        public boolean contains(Object o) {
            if (o instanceof Object[]) {
                for (T t : this) {
                    if (Arrays.equals((Object[]) t, (Object[]) o)) {
                        return true;
                    }
                }
            }
            return false;
        }

        @Override
        public boolean remove(Object o) {
            if (o instanceof Object[]) {
                for (T t : this) {
                    if (Arrays.equals((Object[]) t, (Object[]) o)) {
                        return super.remove(t);
                    }
                }
            }
            return false;
        }

        @Override
        public boolean add(T t) {
            if (t instanceof Object[]) {
                for (T t1 : this) {
                    if (Arrays.equals((Object[]) t1, (Object[]) t)) {
                        return false;
                    }
                }
            }
            return super.add(t);
        }

        @Override
        public boolean equals(Object o) {
            if (o instanceof ArrayHashSet) {
                return super.equals(o);
            }
            return false;
        }
    }

    /**
     * Get the position of a kmer given a pointer, origin and outgoing edge
     * (Partially copied from MapReadsParallel.java)
     * @param pointer pointer to the node
     * @param origin the origin of the range (format: G<genome>S<sequence>)
     * @param incomingEdge the incoming edge of the node
     * @return an array of kmer positions
     * NB: assumes within transaction!
     */
    private static int[] getKmerPositions(final IndexPointer pointer, final String origin, final Node node, final Relationship incomingEdge, final boolean isCanonical) {
        Pantools.logger.trace("\tPointer({}); origin({}); node({}); incomingEdge({}); isCanonical({})", pointer, origin, node.getId(), incomingEdge.getId(), isCanonical);

        final int nodeLength = (int) node.getProperty("length");
        final char side = incomingEdge.getType().name().charAt(1);
        final int[] positions = (int[]) incomingEdge.getProperty(origin);
        final int[] kmerPositions = new int[positions.length];

        // TODO: this idea is copied from MapReadsParallel.java, no guarantee it always works!!!
        for (int i = 0; i < positions.length; i++) {
            final int position = positions[i];
            if (side == 'F') {
                if (pointer.canonical ^ isCanonical) {
                    Pantools.logger.trace("\tReturning1: {}", position + pointer.offset + K_SIZE);
                    kmerPositions[i] = position + pointer.offset + K_SIZE;
                } else {
                    Pantools.logger.trace("\tReturning2: {}", position + pointer.offset);
                    kmerPositions[i] = position + pointer.offset;
                }
            } else {
                if (pointer.canonical ^ isCanonical) {
                    Pantools.logger.trace("\tReturning3: {}", position - pointer.offset + nodeLength - K_SIZE);
                    kmerPositions[i] = position - pointer.offset + nodeLength - K_SIZE;
                } else {
                    Pantools.logger.trace("\tReturning4: {}", position - pointer.offset + nodeLength);
                    kmerPositions[i] = position - pointer.offset + nodeLength;
                }
            }
        }

        return kmerPositions;
    }

    /**
     * Write the protein sequence of mRNA nodes to a file
     * @param mrnaNodes set of mRNA nodes to write protein sequences for
     * @param outputFasta output file to write protein sequences to
     * @throws IOException when the output file cannot be written
     */
    public static void writeProteinToFasta(final Set<Node> mrnaNodes, final Path outputFasta) throws IOException {
        try (BufferedWriter writer = Files.newBufferedWriter(outputFasta)) {
            for (Node mrnaNode : mrnaNodes) {
                final String protein = mrnaNode.hasProperty("protein") ?
                        (String) mrnaNode.getProperty("protein", "") :
                        (String) mrnaNode.getProperty("protein_sequence", "");
                if (mrnaNode.hasProperty("protein_ID")) {
                    final String name = (String) mrnaNode.getProperty("protein_ID");
                    writer.write(String.format(">%s\n%s\n", name, protein));
                } else {
                    Pantools.logger.debug("Skipping mRNA node {} as it does not have a protein ID.", mrnaNode.getId());
                }
            }
        }
    }
}
