package nl.wur.bif.pantools.utils.graph;

import htsjdk.tribble.annotation.Strand;
import htsjdk.tribble.gff.Gff3Feature;
import htsjdk.tribble.gff.Gff3FeatureImpl;
import nl.wur.bif.pantools.Pantools;
import nl.wur.bif.pantools.construction.index.kmer;
import nl.wur.bif.pantools.utils.Globals;
import nl.wur.bif.pantools.utils.Utils;
import nl.wur.bif.pantools.utils.local_sequence_alignment.LocalSequenceAlignment;
import org.neo4j.graphdb.*;

import java.io.IOException;
import java.nio.file.Files;
import java.util.*;

import static nl.wur.bif.pantools.utils.GraphUtils.getSequenceNumbers;

public class GenomeRange {
    private final int genomeNr;
    private final int sequenceNr;
    private final int start;
    private final int end;

    private Set<Node> treeGeneNodeCache = null;

    /**
     * Constructor for a genomic range with optional validation.
     *
     * @param genomeNr   The genome number.
     * @param sequenceNr The sequence number.
     * @param start      The start position.
     * @param end        The end position.
     * @param validate   Whether to validate the genomic range.
     */
    public GenomeRange(int genomeNr, int sequenceNr, int start, int end, boolean validate) {
        this.genomeNr = genomeNr;
        this.sequenceNr = sequenceNr;
        this.start = start;
        this.end = end;
        if (validate) {
            validate();
        }
    }

    /**
     * Constructor for a genomic range.
     *
     * @param address The address of the genomic range.
     * @param validate Whether to validate the genomic range.
     */
    public GenomeRange(int[] address, boolean validate) {
        this.genomeNr = address[0];
        this.sequenceNr = address[1];
        this.start = address[2];
        this.end = address[3];
        if (validate) {
            validate();
        }
    }

    /**
     * Method for merging two genomic ranges that have the same genome and sequence number.
     * @param other The other genomic range.
     * @return The merged genomic range.
     * @throws IllegalArgumentException If the genomic ranges cannot be merged.
     */
    public GenomeRange merge(GenomeRange other) throws IllegalArgumentException {
        if (this.genomeNr != other.genomeNr || this.sequenceNr != other.sequenceNr) {
            Pantools.logger.error("Cannot merge genomic ranges from different genomes or sequences: {} and {}.", this, other);
            throw new IllegalArgumentException("Cannot merge unrelated genomic ranges");
        }

        int newStart = Math.min(this.start, other.start);
        int newEnd = Math.max(this.end, other.end);
        return new GenomeRange(this.genomeNr, this.sequenceNr, newStart, newEnd, true);
    }

    /**
     * Validate the genomic range.
     * @throws IllegalArgumentException If the genomic range is invalid.
     */
    private void validate() throws IllegalArgumentException {
        // basic validation
        if (genomeNr <= 0 || sequenceNr < 0 || start <= 0 || end <= 0 || start > end) {
            Pantools.logger.error("Impossible genomic range: {}.", this);
            throw new IllegalArgumentException("Impossible genomic range: " + this);
        }

        // check if genome number exists
        if (Globals.GENOME_DB.num_genomes < genomeNr) {
            Pantools.logger.error("Genome number does not exist: {}.", this);
            throw new IllegalArgumentException("Invalid genomic range (genome number does not exist): " + this);
        }

        // check if sequence exist
        if (!getSequenceNumbers(genomeNr).containsKey(Utils.getSequenceName(genomeNr, sequenceNr))) {
            Pantools.logger.error("Sequence number does not exist: {}.", this);
            throw new IllegalArgumentException("Invalid genomic range (sequence does not exist): " + this);
        }

        // check if start and end are within the sequence
        if (end > Globals.GENOME_SC.get_sequence_length(genomeNr, sequenceNr)) { //only need to check end, since start is always less than end by here
            Pantools.logger.error("Coordinates are beyond the sequence length: {}.", this);
            throw new IllegalArgumentException("Invalid genomic range (coordinates are beyond the sequence length): " + this);
        }
    }

    @Override
    public boolean equals(Object o) {
        if (this == o) return true;
        if (o == null || getClass() != o.getClass()) return false;

        GenomeRange that = (GenomeRange) o;

        if (genomeNr != that.genomeNr) return false;
        if (sequenceNr != that.sequenceNr) return false;
        if (start != that.start) return false;
        return end == that.end;
    }

    @Override
    public int hashCode() {
        int result = genomeNr;
        result = 31 * result + sequenceNr;
        result = 31 * result + start;
        result = 31 * result + end;
        return result;
    }

    /**
     * toString method.
     *
     * @return The string representation of the genomic range.
     */
    @Override
    public String toString() {
        return String.format("%d#%d:%d-%d", genomeNr, sequenceNr, start, end);
    }

    /**
     * Parser for a genomic range. Should be formatted as: genome#sequence:start-end.
     *
     * @param genomeRange The genomic range string.
     */
    public static GenomeRange parseGenomeRange(String genomeRange) {
        try {
            String[] parts = genomeRange.split(":");

            String[] genomeParts = parts[0].split("#");
            int genomeNr = Integer.parseInt(genomeParts[0]);
            int sequenceNr = getSequenceNr(genomeNr, genomeParts[1]);

            String[] rangeParts = parts[1].split("-");
            int start = Integer.parseInt(rangeParts[0]);
            int end = Integer.parseInt(rangeParts[1]);
            return new GenomeRange(genomeNr, sequenceNr, start, end, true);
        } catch (Exception e) {
            throw new IllegalArgumentException("Invalid genome range: " + genomeRange, e);
        }
    }

    /**
     * Get the sequence number from a sequence name.
     *
     * @param sequenceName The sequence name.
     * @return The sequence number.
     */
    private static int getSequenceNr(int genomeNr, String sequenceName) {
        final HashMap<String, Integer> sequenceNumbers = getSequenceNumbers(genomeNr);
        try {
            return sequenceNumbers.get(sequenceName);
        } catch (NullPointerException e) {
            Pantools.logger.error("Could not find sequence {} for genome {}.", sequenceName, genomeNr);
            throw new RuntimeException("Invalid sequence");
        }
    }

    /**
     * Get the genome number.
     *
     * @return The genome number.
     */
    public int getGenomeNr() {
        return genomeNr;
    }

    /**
     * Get the sequence number.
     *
     * @return The sequence number.
     */
    public int getSequenceNr() {
        return sequenceNr;
    }

    /**
     * Get the origin.
     *
     * @return The origin.
     */
    public String getOrigin() {
        return String.format("G%dS%d", genomeNr, sequenceNr);
    }

    /**
     * Get the start position.
     *
     * @return The start position.
     */
    public int getStart() {
        return start;
    }

    /**
     * Get the end position.
     *
     * @return The end position.
     */
    public int getEnd() {
        return end;
    }

    /**
     * Get the length of the genomic range.
     * @return The length.
     */
    public int getLength() {
        return end - start + 1;
    }

    /**
     * Whether the genomic range fully contains another genomic range.
     * @param other The other genomic range.
     * @return Whether the genomic range contains the other genomic range.
     */
    public boolean fullyContains(GenomeRange other) {
        return this.genomeNr == other.genomeNr && this.sequenceNr == other.sequenceNr && this.start <= other.start && this.end >= other.end;
    }

    /**
     * Whether the genomic range overlaps (partially or completely) with another genomic range.
     * @param other The other genomic range.
     * @return Whether the genomic ranges overlap.
     */
    public boolean overlaps(GenomeRange other) {
        return this.genomeNr == other.genomeNr && this.sequenceNr == other.sequenceNr && this.start <= other.end && this.end >= other.start;
    }

    /**
     * Get the fasta header for the genomic range.
     * @return The fasta header.
     */
    public String getFastaHeader() {
        return String.format(">%s original:%s", getSequenceHeader(), getSequenceName());
    }

    /**
     * Get the first string from the fasta header as sequence ID
     * @return The sequence ID.
     */
    public String getSequenceHeader() {
        return String.format("%d#%d:%d-%d", genomeNr, sequenceNr, start, end);
    }

    /**
     * Get a cleaned version of the fasta header for the genomic range.
     * @return The cleaned fasta header.
     */
    public String getCleanedFastaHeader() {
        return String.format(">%s original:%s", getCleanedSequenceHeader(), getSequenceName());
    }

    /**
     * Get a cleaned version of the first string from the fasta header as sequence ID
     * @return The cleaned sequence ID.
     */
    public String getCleanedSequenceHeader() {
        return String.format("%d_%d_%d_%d", genomeNr, sequenceNr, start, end);
    }

    /**
     * Get the original sequence name.
     * @return The sequence name.
     */
    public String getSequenceName() {
        return Utils.getSequenceName(genomeNr, sequenceNr);
    }

    /**
     * Get the nucleotide sequence of the genomic range.
     * @return The nucleotide sequence.
     */
    public String getNucleotideSequence() {
        final StringBuilder seq = new StringBuilder();
        Globals.GENOME_SC.get_sub_sequence(seq, genomeNr, sequenceNr, start - 1, end - start + 1, true);
        return seq.toString();
    }

    /**
     * Get the phasing information of the sequence on which the genomic range is located.
     * @return The phasing information.
     */
    public String getPhasingInformation() {
        // obtain the sequence node
        try (Transaction ignored = Globals.GRAPH_DB.beginTx()) {
            Node sequenceNode = Globals.GRAPH_DB.findNode(Globals.SEQUENCE_LABEL, "identifier", String.format("%d_%d", genomeNr, sequenceNr));

            // get the phasing information if present
            if (sequenceNode.hasProperty("phasing_ID")) {
                return (String) sequenceNode.getProperty("phasing_ID");
            } else {
                return "NA";
            }
        } catch (NotFoundException nfe) {
            Pantools.logger.error("Could not find sequence node in the graph database.");
            throw new RuntimeException(nfe);
        }
    }

    /**
     * Performs a global alignment of another genomic range to this genomic range.
     * (Code has been partially copied from Grouping.java)
     * @param other The other genomic range.
     * @return The similarity score of the alignment.
     */
    public double alignToGenomeRange(GenomeRange other) {
        Pantools.logger.debug("Aligning {} to {}", this, other);
        final LocalSequenceAlignment aligner = new LocalSequenceAlignment(Globals.GAP_OPEN, Globals.GAP_EXT, Globals.MAX_ALIGNMENT_LENGTH, Globals.CLIPPING_STRINGENCY, "NUC.4.4");

        // Obtain sequences for this and the other genome range
        final String seq1 = this.getNucleotideSequence();
        final String seq2 = other.getNucleotideSequence();

        // Perform the alignment
        long score = 0, p_score = 0;
        StringBuilder cigar = new StringBuilder();
        final int m = seq1.length();
        final int n = seq2.length();
        if (n >= Globals.MAX_ALIGNMENT_LENGTH) {
            String query, subject;
            int parts_num = (n / Globals.MAX_ALIGNMENT_LENGTH) + (n % Globals.MAX_ALIGNMENT_LENGTH == 0 ? 0 : 1);
            if (n % Globals.MAX_ALIGNMENT_LENGTH == 0) { // sequences of exactly the maximum_alignment length (1000, 2000, etc) must be split in an extra part
                parts_num++;
            }
            final int part_len1 = m / parts_num;
            final int part_len2 = n / parts_num;
            Pantools.logger.trace("{} {} parts num {} {} {}.", m, n, parts_num, part_len1, part_len2);
            for (int i = 0; i < parts_num; ++i) {
                query = seq1.substring(i * part_len1, Math.min(m, (i + 1) * part_len1));
                subject = seq2.substring(i * part_len2, Math.min(n, (i + 1) * part_len2));
                aligner.align(query, subject);
                score += aligner.get_similarity();
                p_score += aligner.get_match_score(query, query);//5 * query.length();
                cigar.append(';').append(aligner.get_cigar());
            }
        } else {
            aligner.align(seq1, seq2);
            score = aligner.get_similarity();
            p_score = aligner.get_match_score(seq1, seq1);//5 * query.length();
        }

        // Log the CIGAR
        Pantools.logger.trace("CIGAR: {}", cigar);

        return score * 100.0 / p_score;
    }

    /**
     * Get evenly spread kmers from the genomic range.
     * (The code is partially copied from MapReadsParallel.java)
     * @param numKmerSamples The number of kmer samples to take.
     * @return A map of kmers and their location in the genomic range.
     */
    public Map<kmer, Integer> getKmers(final int numKmerSamples) {
        Pantools.logger.debug("Getting {} kmers for {}.", numKmerSamples, this);

        // Define the number of steps to take
        final int step = getKmerStep(numKmerSamples);
        Pantools.logger.trace("Step: {}", step);

        // Set up the kmerization
        final String sequence = this.getNucleotideSequence();
        kmer currentKmer = initializeKmer(sequence);
        final Map<kmer, Integer> kmers = new HashMap<>();

        for (int pos = Globals.K_SIZE; pos < this.getLength(); ) { //pos holds the end position of the kmer!
            // If current kmer exists, add to the set
            long currentIndex = Globals.INDEX_SC.find(currentKmer);
            if (currentIndex != -1L) {
                Pantools.logger.trace("Adding kmer {} at position {}.", currentKmer, pos - Globals.K_SIZE);
                kmers.put(new kmer(currentKmer), pos - Globals.K_SIZE);
            }

            // Exit the loop if the end of the sequence is reached
            if (pos + step >= this.getLength()) {
                break;
            }

            // Move the kmer
            for (int i = 0; i < step; ++i, ++pos) {
                currentKmer.next_fwd_kmer(Globals.binary[sequence.charAt(pos)] & 3);
            }
        }

        return kmers;
    }

    /**
     * Get the step size between kmers.
     * @param numKmerSamples The number of kmer samples to take.
     * @return The step size.
     */
    public int getKmerStep(final int numKmerSamples) {
        final int length = this.getLength();
        int step = length / numKmerSamples;
        return (step == 0 ? 1 : step);
    }

    /**
     * Initialize the kmer.
     * (The code is partially copied from MapReadsParallel.java)
     * @param sequence The sequence to be kmerized
     * @return The initialized kmer.
     */
    private kmer initializeKmer(String sequence) {
        int i;
        kmer currentKmer = new kmer(Globals.INDEX_SC.get_K(), Globals.INDEX_SC.get_pre_len());
        currentKmer.reset();
        for (i = 0; i < Globals.K_SIZE; ++i) {
            currentKmer.next_fwd_kmer(Globals.binary[sequence.charAt(i)] & 3);
        }
        return currentKmer;
    }

    /**
     * Get all BLAST results in the genomic range.
     * @param blastResultsPath The path to the BLAST results.
     * @return A set of GFF3 results.
     */
    public Set<Gff3Feature> getBlastGff3Features(java.nio.file.Path blastResultsPath) {
        final String sequenceName = getCleanedSequenceHeader();

        // loop over all BLAST results in the given region
        //
        // Example BLAST results file:
        //#TBLASTN
        //#Strictness for accepting hits was >= 0% sequence identity with >= 0% alignment length of input sequence length
        //#query	subject	identity	length	mismatches	gaps	alignment start query	alignment end query	alignment start subject	alignment end subject	evalue	bitscore	PASS minimum identity	PASS minimum alignment length	genome	sequence (number)	original sequence name	start coordinate	end coordinate	mRNA length (bp)	protein length (aa)	mRNA identifier	mRNA name	mRNA node identifier	homology group node identifier	gene identifier	gene name	gene node identifier
        //
        //##sequence 1,>cds-YP_009724397.2, 419 length
        //cds-YP_009724397.2	4_1	100.000	419	0	0	1	419	28274	29530	0.0	657	PASS	PASS	4	1	NC_045512.2	28274	29533	1260	419	cds-YP_009724397.2	N	291	158	gene-GU280_gp10	N	290
        //cds-YP_009724397.2	3_1	100.000	419	0	0	1	419	28274	29530	0.0	657	PASS	PASS	3	1	MN975262.1	28274	29533	1260	419	cds-YP_009724397.2	N	251	158	gene-GU280_gp10	N	250
        //cds-YP_009724397.2	2_1	100.000	419	0	0	1	419	28242	29498	0.0	657	PASS	PASS	2	1	MN938384.1	28242	29501	1260	419	cds-YP_009724397.2	N	215	158	gene-GU280_gp10	N	214
        //cds-YP_009724397.2	1_1	100.000	419	0	0	1	419	28274	29530	0.0	657	PASS	PASS	1	1	MN908947.3	28274	29533	1260	419	cds-YP_009724397.2	N	179	158	gene-GU280_gp10	N	178
        //
        //##sequence 2,>cds-YP_009725255.1, 38 length
        //cds-YP_009725255.1	4_1	100.000	38	0	0	1	38	29558	29671	4.99e-22	78.6	PASS	PASS	4	1	NC_045512.2	29558	29674	117	38	cds-YP_009725255.1	ORF10	294	1	gene-GU280_gp11	ORF10	293
        //cds-YP_009725255.1	4_1	53.333	15	4	1	9	20	17130	17174	1.6	18.1	PASS	PASS	4	1	NC_045512.2	266	21555	21290	11502	cds-YP_009724389.1	ORF1ab	260	159	gene-GU280_gp01	ORF1ab	259
        //cds-YP_009725255.1	1_1	100.000	38	0	0	1	38	29558	29671	4.99e-22	78.6	PASS	PASS	1	1	MN908947.3	29558	29674	117	38	cds-YP_009725255.1	ORF10	182	1	gene-GU280_gp11	ORF10	181
        //cds-YP_009725255.1	1_1	53.333	15	4	1	9	20	17130	17174	1.6	18.1	PASS	PASS	1	1	MN908947.3	266	21555	21290	14196	cds-YP_009724389.1	ORF1ab	10	149	gene-GU280_gp01	ORF1ab	9
        //cds-YP_009725255.1	3_1	100.000	38	0	0	1	38	29558	29671	4.99e-22	78.6	PASS	PASS	3	1	MN975262.1	29558	29674	117	38	cds-YP_009725255.1	ORF10	254	1	gene-GU280_gp11	ORF10	253
        //cds-YP_009725255.1	3_1	53.333	15	4	1	9	20	17130	17174	1.6	18.1	PASS	PASS	3	1	MN975262.1	266	21555	21290	14196	cds-YP_009724389.1	ORF1ab	222	149	gene-GU280_gp01	ORF1ab	221
        //cds-YP_009725255.1	2_1	100.000	38	0	0	1	38	29526	29639	4.99e-22	78.6	PASS	PASS	2	1	MN938384.1	29526	29642	117	38	cds-YP_009725255.1	ORF10	218	1	gene-GU280_gp11	ORF10	217
        //cds-YP_009725255.1	2_1	53.333	15	4	1	9	20	17098	17142	1.6	18.1	PASS	PASS	2	1	MN938384.1	234	21523	21290	14196	cds-YP_009724389.1	ORF1ab	186	149	gene-GU280_gp01	ORF1ab	185
        try {
            Set<Gff3Feature> gff3Features = new HashSet<>();
            int hitCounter = 0;
            for (String line : Files.readAllLines(blastResultsPath)) {
                // parse the line
                if (line.startsWith("#")) {
                    continue;
                }
                String[] parts = line.split("\t");
                if (parts.length < 12) {
                    continue;
                }

                // obtain the relevant information
                final String[] blastOrigin = parts[1].split("_");
                final int blastGenomeNr = Integer.parseInt(blastOrigin[0]);
                final int blastSequenceNr = Integer.parseInt(blastOrigin[1]);
                final int blastStart = Math.min(Integer.parseInt(parts[8]), Integer.parseInt(parts[9]));
                final int blastEnd = Math.max(Integer.parseInt(parts[8]), Integer.parseInt(parts[9]));
                final GenomeRange blastRange = new GenomeRange(blastGenomeNr, blastSequenceNr, blastStart, blastEnd, false);
                final double blastIdentity = Double.parseDouble(parts[2]);
                final double blastEval = Double.parseDouble(parts[10]);
                final double blastBitscore = Double.parseDouble(parts[11]);

                // check if the hit is in the region
                if (!overlaps(blastRange)) {
                    Pantools.logger.debug("Blast hit ({}) does not overlap with region ({}).", blastRange, this);
                    continue;
                }

                // define the attributes
                Map<String, List<String>> attributes = new LinkedHashMap<>();
                attributes.put("ID", Collections.singletonList(String.format("%s_hit%s_genome%d", parts[0], ++hitCounter, blastGenomeNr)));
                attributes.put("Name", Collections.singletonList(parts[0]));
                attributes.put("percent_identity", Collections.singletonList(Double.toString(blastIdentity)));
                attributes.put("evalue", Collections.singletonList(Double.toString(blastEval)));
                attributes.put("bitscore", Collections.singletonList(Double.toString(blastBitscore)));

                // add the BLAST hit to the set (also: adjust for the start of the region)
                gff3Features.add(new Gff3FeatureImpl(
                        sequenceName,
                        "TBLASTN",
                        "blast_hit",
                        blastStart - this.start + 1,
                        blastEnd - this.start + 1,
                        -1d,
                        Strand.NONE,
                        -1,
                        attributes
                ));
            }
            return gff3Features;
        } catch (IOException e) {
            Pantools.logger.error("Could not read BLAST results from file: {}.", blastResultsPath);
            throw new RuntimeException(e);
        }
    }

    /**
     * Get all repeat features in the genomic range.
     * @return A set of GFF3 features.
     */
    public Set<Gff3Feature> getRepeatGff3Features() {
        final String sequenceName = getCleanedSequenceHeader();

        // loop over all repeat nodes in the given region
        try (Transaction ignored = Globals.GRAPH_DB.beginTx()) {
            Set<Gff3Feature> gff3Features = new HashSet<>();
            for (Node repeatNode : getRepeatNodes()) {
                int[] address = (int[]) repeatNode.getProperty("address");
                // add the repeat node to the set (also: adjust for the start of the region)
                Map<String, List<String>> attributes = new LinkedHashMap<>();
                attributes.put("ID", Collections.singletonList(Long.toString(repeatNode.getId())));
                attributes.put("source", Collections.singletonList((String) repeatNode.getProperty("source")));
                attributes.put("type", Collections.singletonList((String) repeatNode.getProperty("type")));
                attributes.put("Name", Collections.singletonList((String) repeatNode.getProperty("type")));
                gff3Features.add(new Gff3FeatureImpl(
                        sequenceName,
                        (String) repeatNode.getProperty("source"),
                        (String) repeatNode.getProperty("type"),
                        address[2] - start + 1,
                        address[3] - start + 1,
                        -1.0,
                        Strand.NONE,
                        -1,
                        attributes
                ));
            }
            return gff3Features;
        } catch (NotFoundException nfe) {
            Pantools.logger.error("Could not find repeat nodes in the graph database.");
            throw new RuntimeException(nfe);
        }
    }

    /**
     * Get all repeat nodes in the genomic range.
     * @return A set of repeat nodes.
     */
    public Set<Node> getRepeatNodes() {
        // loop over all repeats on the given origin to check if they are in the region
        try (ResourceIterator<Node> repeatNodes = Globals.GRAPH_DB.findNodes(Globals.REPEAT_LABEL)) {
            Set<Node> nodes = new HashSet<>();
            repeatNodes.stream().forEach(repeatNode -> {
                final int[] address = (int[]) repeatNode.getProperty("address");
                if (address[0] != genomeNr || address[1] != sequenceNr) {
                    return;
                }
                if (overlaps(new GenomeRange(address, false))) {
                    nodes.add(repeatNode);
                }
            });
            return nodes;
        } catch (NotFoundException nfe) {
            Pantools.logger.error("Could not find repeat nodes in the graph database.");
            throw new RuntimeException(nfe);
        }
    }

    /**
     * Get all gene features and their children in the genomic range.
     * @return A set of GFF3 features.
     */
    public Set<Gff3Feature> getGff3Features() {
        final String sequenceName = getCleanedSequenceHeader();

        // loop over all gene nodes in the given region
        try (Transaction ignored = Globals.GRAPH_DB.beginTx()) {
            Set<Gff3Feature> gff3Features = new HashSet<>();
            for (Node featureNode : getTreeGeneNodes()) {
                // add the feature node to the set (also: adjust for the start of the region)
                final Strand strand = featureNode.getProperty("strand").equals("+") ? Strand.POSITIVE : Strand.NEGATIVE;
                Map<String, List<String>> attributes = Utils.parseStringToMap((String) featureNode.getProperty("attribute"));
                gff3Features.add(new Gff3FeatureImpl(
                        sequenceName,
                        (String) featureNode.getProperty("source"),
                        (String) featureNode.getProperty("type"),
                        (int) featureNode.getProperty("begin") - start + 1,
                        (int) featureNode.getProperty("end") - start + 1,
                        (double) featureNode.getProperty("score"),
                        strand,
                        (int) featureNode.getProperty("phase"),
                        attributes
                ));
            }
            return gff3Features;
        } catch (NotFoundException nfe) {
            Pantools.logger.error("Could not find gene nodes in the graph database.");
            throw new RuntimeException(nfe);
        }
    }

    /**
     * Get all gene nodes in the genomic range including their children.
     * @return A set of gene nodes.
     * NB: Assumes within neo4j transaction!
     */
    public Set<Node> getTreeGeneNodes() {
        // if the treeGeneNodeCache is not empty, return it
        if (treeGeneNodeCache != null) {
            return treeGeneNodeCache;
        }

        // loop over all genes on the given origin to check if they are in the region
        try (ResourceIterator<Node> geneNodes = Globals.GRAPH_DB.findNodes(Globals.GENE_LABEL, "genome", genomeNr, "sequence", sequenceNr)) {
            Set<Node> nodes = new HashSet<>();
            geneNodes.stream().forEach(geneNode -> {
                final int[] address = (int[]) geneNode.getProperty("address");
                if (overlaps(new GenomeRange(address, false))) {
                    nodes.addAll(getTreeGeneNode(geneNode));
                }
            });

            // cache the result for future use
            treeGeneNodeCache = nodes;
            return nodes;
        } catch (NotFoundException nfe) {
            Pantools.logger.error("Could not find gene nodes in the graph database.");
            throw new RuntimeException(nfe);
        }
    }

    /**
     * Retrieve the children of a gene node.
     * @param geneNode The gene node.
     * @return The children of the gene node and gene node itself.
     * NB: Assumes within neo4j transaction!
     */
    private List<Node> getTreeGeneNode(Node geneNode) {
        List<Node> children = new ArrayList<>();
        children.add(geneNode);

        // do a breadth first search to get all children via the is_parent_of relationships
        Queue<Node> queue = new LinkedList<>();
        queue.add(geneNode);
        while (!queue.isEmpty()) {
            Node node = queue.poll();
            for (Relationship relationship : node.getRelationships(Direction.OUTGOING, Globals.RelTypes.is_parent_of)) {
                Node child = relationship.getEndNode();
                children.add(child);
                queue.add(child);
            }
        }

        return children;
    }
}
