package nl.wur.bif.pantools.utils.graph;

import nl.wur.bif.pantools.Pantools;
import nl.wur.bif.pantools.construction.build_pangenome.GenomeLayer;
import nl.wur.bif.pantools.construction.index.IndexPointer;
import nl.wur.bif.pantools.utils.Globals;
import nl.wur.bif.pantools.utils.graph.graph_objects.*;
import nl.wur.bif.pantools.utils.graph.graph_objects.GraphPath;
import org.neo4j.graphdb.*;

import java.util.*;

/**
 * Extract a graph from the index database.
 */
public class IndexGraphExtractor {

    /**
     * Extract an entire pangenome graph from the index database.
     * @return The extracted graph.
     */
    public Graph extractNucleotideGraph() {
        GraphRange entirePangenome = createEntirePangenomeRange();
        return extractNucleotideGraph(entirePangenome);
    }

    /**
     * Extract a graph range from the index database.
     * @param graphRange The range to extract.
     * @return The extracted graph.
     */
    public Graph extractNucleotideGraph(GraphRange graphRange) {
        Pantools.logger.info("Extracting nucleotide graph from the index database.");

        // Create maps and set to store the nodes, edges and paths
        final Map<Long, GraphNode> graphNodes = new HashMap<>();
        final Map<Long, GraphEdge> graphEdges = new HashMap<>();
        final Set<GraphPath> graphPaths = new HashSet<>();

        // Start up neo4j database for node properties and to let GenomeLayer.locate() use it for finding degenerate nodes
        try (Transaction ignored = Globals.GRAPH_DB.beginTx()) {

            // For each genome range in the graph range, extract the nodes from the index database and add to the graph
            Map<GraphNodePair, GraphEdge> edgeCache = new HashMap<>(); // pair of node IDs -> edge
            long edgeId = 1L;
            long maxEdgeId = 0L;
            for (GenomeRange genomeRange : graphRange.getGenomeRanges()) {
                Pantools.logger.debug("Extracting nucleotide graph for genome range: {}", genomeRange);
                final List<GraphNode> graphNodesList = new ArrayList<>();
                final List<GraphEdge> graphEdgesList = new ArrayList<>();

                long previousNodeId = -1;
                GraphNode previousGraphNode = null;
                boolean previousCanonical = false;

                // For each k-mer in the genome range, locate the IndexPointer and obtain node id from index database
                for (int i = genomeRange.getStart() - 1; i <= genomeRange.getEnd() - Globals.K_SIZE; i++) {
//                    Pantools.logger.trace("Extracting nucleotide graph for genome range: {} at position: {}", genomeRange, i);

                    final IndexPointer indexPointer = GenomeLayer.locate(Globals.GRAPH_DB,
                            Globals.GENOME_SC,
                            Globals.INDEX_SC,
                            genomeRange.getGenomeNr(),
                            genomeRange.getSequenceNr(),
                            i
                    );
                    final long nodeId = indexPointer.node_id;
                    final int offset = indexPointer.offset;
                    final boolean canonical = !indexPointer.canonical;
//                    Pantools.logger.trace("Found node id: {} at offset: {} with canonical: {}", nodeId, offset, canonical);

                    // skip if offset is not 0 because that means the k-mer is not the first k-mer in the node (unitig magic)
                    if (offset != 0) {
                        continue;
                    }

                    // add to graphNodesList
                    final Node nucleotideNode = Globals.GRAPH_DB.getNodeById(nodeId);
                    final GraphNode graphNode = new GraphNode(nodeId,
                            (String) nucleotideNode.getProperty("sequence"),
                            (long) nucleotideNode.getProperty("frequency")
                    );
                    graphNodesList.add(graphNode);

                    // add to graphEdgesList
                    if (previousNodeId != -1) {
                        final GraphNodePair edgeKey = new GraphNodePair(previousNodeId, nodeId);
                        GraphEdge graphEdge = edgeCache.get(edgeKey);
                        if (graphEdge == null) {
                            graphEdge = new GraphEdge(edgeId,
                                    previousGraphNode,
                                    graphNode,
                                    getRelationshipType(previousCanonical, canonical)
                            );
                            edgeId++;
                            edgeCache.put(edgeKey, graphEdge);
                        }
                        graphEdgesList.add(graphEdge);
                        maxEdgeId = Math.max(maxEdgeId, edgeId);
                    }

                    // update previousNodeId, previousGraphNode and indexPointer
                    previousNodeId = nodeId;
                    previousGraphNode = graphNode;
                    previousCanonical = canonical;
                }

                // add to graphNodes and graphEdges
                graphNodesList.forEach(graphNode -> graphNodes.putIfAbsent(graphNode.getId(), graphNode));
                graphEdgesList.forEach(graphEdge -> graphEdges.putIfAbsent(graphEdge.getId(), graphEdge));

                // extend graphEdgesList by one fake GraphNode to indicate the end of the path
                graphEdgesList.add(new GraphEdge(maxEdgeId++,
                        previousGraphNode,
                        new GraphNode(0),
                        getRelationshipType(previousCanonical, true)
                ));
                graphPaths.add(new GraphPath(genomeRange.getOrigin(),
                        graphNodesList,
                        graphEdgesList
                ));
            }
        } catch (NotFoundException nfe) {
            Pantools.logger.error("Error while extracting nucleotide graph from the index database.", nfe);
            throw new RuntimeException("Could not extract nucleotide graph", nfe);
        }

        return new Graph(graphNodes, graphEdges, graphPaths);
    }

    /**
     * Inner class for GraphNodePair. Used as key in the edgeCache and ensures that source and target nodes can be compared.
     */
    private static class GraphNodePair {
        private final long source;
        private final long target;

        public GraphNodePair(long source, long target) {
            this.source = source;
            this.target = target;
        }

        @Override
        public boolean equals(Object o) {
            if (this == o) return true;
            if (!(o instanceof GraphNodePair)) return false;
            GraphNodePair that = (GraphNodePair) o;
            return source == that.source && target == that.target;
        }

        @Override
        public int hashCode() {
            return Objects.hash(source, target);
        }
    }

    /**
     * Get the relationship type (FF, FR, RF, RR) from the canonicality of the source and target node.
     * @param sourceCanonical The canonicality of the source node.
     * @param targetCanonical The canonicality of the target node.
     * @return The relationship type.
     */
    private RelationshipType getRelationshipType(boolean sourceCanonical, boolean targetCanonical) {
        if (sourceCanonical) {
            return targetCanonical ? Globals.RelTypes.FF : Globals.RelTypes.FR;
        } else {
            return targetCanonical ? Globals.RelTypes.RF : Globals.RelTypes.RR;
        }
    }

    /**
     * Create a graph range that spans the entire pangenome.
     * @return The graph range of the entire pangenome.
     */
    private GraphRange createEntirePangenomeRange() {
        // Obtain the genome ranges from the sequence database
        final Set<GenomeRange> genomeRanges = new HashSet<>();
        for (int g = 1; g <= Globals.GENOME_DB.num_genomes; g++) {
            for (int s = 1; s <= Globals.GENOME_DB.num_sequences[g]; s++) {
                genomeRanges.add(new GenomeRange(g,
                        s,
                        1,
                        (int) Globals.GENOME_DB.sequence_length[g][s],
                        false
                ));
            }
        }

        // Obtain all nucleotide node IDs from the neo4j database
        final Set<Long> nucleotideNodeIds = new HashSet<>();
        try (Transaction ignored = Globals.GRAPH_DB.beginTx();
             ResourceIterator<Node> nucleotideNodes = Globals.GRAPH_DB.findNodes(Globals.NUCLEOTIDE_LABEL)) {
            while (nucleotideNodes.hasNext()) {
                Node node = nucleotideNodes.next();
                nucleotideNodeIds.add(node.getId());
            }
        } catch (NotFoundException nfe) {
            Pantools.logger.error("Error while extracting nucleotide node IDs from the index database.", nfe);
            throw new RuntimeException("Could not retrieve nucleotide nodes", nfe);
        }

        return new GraphRange(nucleotideNodeIds, genomeRanges);
    }
}
