package nl.wur.bif.pantools.utils.graph.graph_objects;

import nl.wur.bif.pantools.Pantools;
import nl.wur.bif.pantools.utils.Utils;

import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Set;

public class GraphPath {
    final private String origin;
    final private GraphNode[] graphNodes;
    final private GraphEdge[] graphEdges;

    /**
     * Full constructor
     *
     * @param origin         name of the path (e.g.: G2S3 (genome 2, sequence 3))
     * @param graphNodes       graphNodes in the path (first and last graphNode should be the sequence graphNode (containing
     *                       metadata), all other graphNodes should represent nucleotides)
     * @param graphEdges          edges in the path (first edge should connect the sequence graphNode to the first nucleotide
     *                       graphNode, last edge should connect the last nucleotide graphNode to the sequence graphNode)
     */
    public GraphPath(String origin, List<GraphNode> graphNodes, List<GraphEdge> graphEdges) {
        this.origin = origin;
        this.graphNodes = graphNodes.toArray(new GraphNode[0]);
        this.graphEdges = graphEdges.toArray(new GraphEdge[0]);
    }

    /**
     * Empty constructor
     *
     * @param origin name of the path (e.g.: G2S3 (genome 2, sequence 3))
     */
    public GraphPath(String origin) {
        this.origin = origin;
        this.graphNodes = new GraphNode[0];
        this.graphEdges = new GraphEdge[0];
    }

    /**
     * Checks if the path is valid.
     * @param graph graph to check the path against
     * @return true if the path is valid, false otherwise
     */
    public boolean isValid(Graph graph) {
        if (graphNodes.length == 0) {
            Pantools.logger.warn("GraphPath {} is invalid: no graphNodes", origin);
            return false;
        }
        if (graphEdges.length == 0) {
            Pantools.logger.warn("GraphPath {} is invalid: no edges", origin);
            return false;
        }
        if (graphNodes.length != graphEdges.length + 1) {
            Pantools.logger.warn("GraphPath {} is invalid: number of graphNodes ({}) does not match number of edges ({})", origin, graphNodes.length, graphEdges.length);
            return false;
        }
        if (!graphNodes[0].equals(graphNodes[graphNodes.length - 1])) {
            Pantools.logger.warn("GraphPath {} is invalid: first and last graphNode are not the same", origin);
            return false;
        }
        if (graphEdges[0].getSource().isNucleotide()) {
            Pantools.logger.warn("GraphPath {} is invalid: first edge ({}) does not connect sequence graphNode to nucleotide graphNode", origin, graphEdges[0]);
            Pantools.logger.warn("\tsource: {}; target: {}", graphEdges[0].getSource().getId(), graphEdges[0].getTarget().getId());
            return false;
        }
        if (graphEdges[graphEdges.length - 1].getTarget().isNucleotide()) {
            Pantools.logger.warn("GraphPath {} is invalid: last edge ({}) does not connect nucleotide graphNode to sequence graphNode", origin, graphEdges[graphEdges.length - 1]);
            Pantools.logger.warn("\tsource: {}; target: {}", graphEdges[graphEdges.length - 1].getSource().getId(), graphEdges[graphEdges.length - 1].getTarget().getId());
            return false;
        }
        for (int i = 0; i < graphEdges.length - 2; i++) {
            GraphNode target = graphEdges[i].getTarget();
            if (!graph.contains(target)) {
                Pantools.logger.warn("GraphPath {} is invalid: graphNode {} does not exist in graph", origin, target.getId());
                return false;
            }
            if (!target.isNucleotide()) {
                Pantools.logger.warn("GraphPath {} is invalid: graphNode {} is not a nucleotide", origin, target.getId());
                return false;
            }
            GraphNode nextSource = graphEdges[i + 1].getSource();
            if (target != nextSource) {
                Pantools.logger.warn("GraphPath {} is invalid: graphNode {} is not connected to graphNode {}", origin, target.getId(), nextSource.getId());
                return false;
            }
        }
        return true;
    }

    /**
     * Return all positions in the path where the given graphNode occurs
     * @param graphNode graphNode to search for
     * @return all positions in the path where the given graphNode occurs
     */
    public Set<Integer> getOccurrences(GraphNode graphNode) {
        Set<Integer> occurrences = new HashSet<>();
        for (int i = 0; i < graphNodes.length; i++) {
            if (graphNodes[i] == graphNode) {
                occurrences.add(i);
            }
        }
        return occurrences;
    }

    /**
     * Generate an iterator over the edges in the path from start to end
     * @return iterator over the edges in the path from start to end
     */
    public Iterator<GraphEdge> walk() {
        return new Iterator<GraphEdge>() {
            private int index = 0;

            @Override
            public boolean hasNext() {
                return index < graphEdges.length;
            }

            @Override
            public GraphEdge next() {
                return graphEdges[index++];
            }
        };
    }

    /**
     * Generate an iterator over the edges in the path from end to start
     * @return iterator over the graphEdges in the path from end to start
     */
    public Iterator<GraphEdge> walkReverse() {
        return new Iterator<GraphEdge>() {
            private int index = graphEdges.length - 1;

            @Override
            public boolean hasNext() {
                return index >= 0;
            }

            @Override
            public GraphEdge next() {
                return graphEdges[index--];
            }
        };
    }

    /**
     * Returns the name of the path (origin) in the format G2S3 (genome 2, sequence 3)
     * @return name of the path (origin)
     */
    public String getOrigin() {
        return origin;
    }

    /**
     * Returns the identifier of the path (origin) in the format 2_3 (genome 2, sequence 3)
     * @return identifier of the path (origin)
     */
    public String getIdentifier() {
        return Utils.originToSequenceIdentifier(origin);
    }

    public GraphNode[] getGraphNodes() {
        return graphNodes;
    }

    public GraphEdge[] getEdges() {
        return graphEdges;
    }

    public boolean equals(GraphPath graphPath) {
        return this.origin.equals(graphPath.getOrigin());
    }

    public String toString() {
        StringBuilder sb = new StringBuilder();
        sb.append("GraphPath ").append(origin).append(": ");
        for (int i = 0; i < graphNodes.length; i++) {
            sb.append(graphNodes[i].getId());
            if (i < graphEdges.length) {
                sb.append(" -(").append(graphEdges[i].getId()).append(")-> ");
            }
        }
        return sb.toString();
    }
}
