package nl.wur.bif.pantools.utils.graph.graph_objects;

import me.tongfei.progressbar.ProgressBar;
import me.tongfei.progressbar.ProgressBarBuilder;
import me.tongfei.progressbar.ProgressBarStyle;
import nl.wur.bif.pantools.Pantools;
import nl.wur.bif.pantools.construction.index.kmer;
import nl.wur.bif.pantools.utils.FileUtils;
import nl.wur.bif.pantools.utils.Globals;
import nl.wur.bif.pantools.utils.GraphUtils;
import nl.wur.bif.pantools.utils.Utils;
import nl.wur.bif.pantools.utils.graph.GenomeRange;
import org.apache.commons.lang.ArrayUtils;
import org.neo4j.graphdb.*;

import java.io.IOException;
import java.nio.file.Path;
import java.util.*;
import java.util.concurrent.ExecutionException;

import static nl.wur.bif.pantools.utils.Globals.*;

public class GraphRange {
    private final Set<Long> nucleotideNodes; // node IDs
    private final Set<GenomeRange> genomeRanges; // genome ranges

    public GraphRange() {
        this.nucleotideNodes = new HashSet<>();
        this.genomeRanges = new HashSet<>();
    }

    public GraphRange(Set<Long> nucleotideNodes, Set<GenomeRange> genomeRanges) {
        this.nucleotideNodes = nucleotideNodes;
        this.genomeRanges = genomeRanges;
    }

    /**
     * Add another graph range to this graph range.
     * NB: this method is NOT thread-safe!
     * @param other graph range to add
     */
    public void add(GraphRange other) {
        if (other == null) {
            return;
        }

        // nodes can just be added to the set
        if (this.nucleotideNodes != null && other.nucleotideNodes != null) {
            this.nucleotideNodes.addAll(other.nucleotideNodes);
        }

        // genome ranges need to be merged in case of overlap
        Set<GenomeRange> newGenomeRanges = new HashSet<>(other.genomeRanges);
        for (GenomeRange otherGenomeRange : other.genomeRanges) {
            Set<GenomeRange> overlappingRanges = new HashSet<>();
            for (GenomeRange thisGenomeRange : this.genomeRanges) {
                if (thisGenomeRange.overlaps(otherGenomeRange)) {
                    overlappingRanges.add(thisGenomeRange);
                }
            }
            Pantools.logger.debug("For other range {}, found overlapping ranges: {}", otherGenomeRange, overlappingRanges);
            if (!overlappingRanges.isEmpty()) {
                GenomeRange mergedGenomeRange = otherGenomeRange;
                for (GenomeRange overlappingRange : overlappingRanges) {
                    mergedGenomeRange = mergedGenomeRange.merge(overlappingRange);
                }
                newGenomeRanges.remove(otherGenomeRange);
                newGenomeRanges.add(mergedGenomeRange);
                this.genomeRanges.removeAll(overlappingRanges);
            } else {
                newGenomeRanges.add(otherGenomeRange);
            }
        }
        this.genomeRanges.addAll(newGenomeRanges);
        Pantools.logger.debug("New genome ranges: {}", this.genomeRanges);
    }

    /**
     * Obtain the range of nodes from BLAST results. The range will contain paths from the current genome. The method
     * will find all BLAST results using the genes of the original genome range in the current genome and use them to
     * find the same region in other genomes based on a synteny block like identification approach.
     * NB: assumes within neo4j transaction!
     * @param currentGenomeNr the number of the current genome to search in
     * @param blastHits set of genome ranges (the tblastn hits)
     * @param allowSinglePositions whether to allow blocks with only one blast hit matching
     * @param maxDistance maximum distance between two consecutive positions
     * @param maxMergingDistance maximum distance between two blocks to consider them the same region (and thus merge them)
     * @param fractPositions fraction of total positions that need to be found to create a hit
     * @param flanking number of flanking bases to include in the output
     * @return a GraphRange object containing the nucleotide nodes
     */
    private static GraphRange addRangeFromBlastBlockMethod(final int currentGenomeNr, final Set<GenomeRange> blastHits, final boolean allowSinglePositions,
                                                           final int maxDistance, final int maxMergingDistance, final double fractPositions, final int flanking,
                                                           final boolean fast)
            throws ExecutionException, InterruptedException {
        Pantools.logger.debug("Obtaining range from BLAST results for genome {}", currentGenomeNr);

        // Determine the set of blast hits of the range on this other genome
        final Set<Integer[]> blastHitPositions = new HashSet<>(); // [sequence, start, end]
        for (GenomeRange blastHit : blastHits) {
            blastHitPositions.add(new Integer[]{blastHit.getSequenceNr(), blastHit.getStart(), blastHit.getEnd()});
        }

        // Run the finding block algorithm
        Pantools.logger.debug("Each block must contain at least {}% of positions within {} bp distance of each other.", fractPositions * 100, maxDistance);
        final List<Set<Integer[]>> blocks = findBlocks(blastHitPositions, maxDistance, fractPositions, allowSinglePositions);
        Pantools.logger.debug("Found {} blocks for genome {}", blocks.size(), currentGenomeNr);

        // Return the graph range
        final GraphRange graphRange = getRangeFromBlocks(currentGenomeNr, blocks, maxMergingDistance, flanking, flanking, fast);
        Pantools.logger.debug("Graph range for genome {}: {}", currentGenomeNr, graphRange);
        return graphRange;
    }

    /**
     * Obtain the range of nodes from homology groups. The range will contain paths from the current genome. The method
     * will find all homology groups in the genome range of the original genome range in the current genome and use them
     * to find the same region in other genomes based on a synteny block like identification approach.
     * NB: assumes within neo4j transaction!
     * @param currentGenomeNr the number of the current genome to search in
     * @param homologyGroups set of homology groups
     * @param allowSinglePositions whether to allow blocks with only one transcript matching
     * @param maxDistance maximum distance between two consecutive positions
     * @param maxMergingDistance maximum distance between two blocks to consider them the same region (and thus merge them)
     * @param fractPositions fraction of total positions that need to be found to create a hit
     * @param maxFrequency maximum frequency of a transcript in a target genome
     * @param flanking number of flanking bases to include in the output
     * @return a GraphRange object containing the nucleotide nodes
     */
    private static GraphRange addRangeFromHomologyBlockMethod(final int currentGenomeNr, final Set<Node> homologyGroups, final boolean allowSinglePositions,
                                                              final int maxDistance, final int maxMergingDistance, final double fractPositions, final int maxFrequency, final int flanking,
                                                              final boolean fast)
            throws ExecutionException, InterruptedException {
        Pantools.logger.debug("Obtaining range from homology groups for genome {}", currentGenomeNr);

        // Determine the set of homology groups of the range on this other genome
        final Set<Integer[]> filteredTranscripts = GraphUtils.getHomologyGroupsBelowFrequency(homologyGroups, currentGenomeNr, maxFrequency); // [sequence, start, end]
        if (filteredTranscripts.isEmpty()) {
            Pantools.logger.warn("Could not find mRNAs in genome {} with frequency below {}", currentGenomeNr, maxFrequency);
            return null;
        }

        // Run the finding block algorithm
        Pantools.logger.debug("Each block must contain at least {}% of positions within {} bp distance of each other.", fractPositions * 100, maxDistance);
        final List<Set<Integer[]>> blocks = findBlocks(filteredTranscripts, maxDistance, fractPositions, allowSinglePositions);
        Pantools.logger.debug("Found {} blocks for genome {}", blocks.size(), currentGenomeNr);

        // Return the graph range
        final GraphRange graphRange = getRangeFromBlocks(currentGenomeNr, blocks, maxMergingDistance, flanking, flanking, fast);
        Pantools.logger.debug("Graph range for genome {}: {}", currentGenomeNr, graphRange);
        return graphRange;
    }

    /**
     * Obtain the range of nodes from a genome range using a block method. The range will contain paths from the current
     * genome. The method will find all k-mers in the genome range of the original genome range in the current
     * genome and use them to find the same region in other genomes based on a synteny block like identification approach.
     * NB: assumes within neo4j transaction!
     * @param currentGenomeNr the number of the current genome to search in
     * @param kmers map of kmers (key: kmer, value: position of kmer in original genome range)
     * @param allowSinglePositions whether to allow blocks with only one kmer matching
     * @param extension extension to the end positions of each block (kmers are sampled as |------|------ instead of ---|------|---)
     * @param maxDistance maximum distance between two consecutive positions
     * @param maxMergingDistance maximum distance between two blocks to consider them the same region (and thus merge them)
     * @param fractPositions the fraction of total k-mers that need to be found to create a hit
     * @param maxFrequency maximum frequency of a k-mer in a target genome
     * @param flanking number of flanking bases to include in the output
     * @return a GraphRange object containing the nucleotide nodes
     */
    private static GraphRange getRangeFromKmerBlockMethod(final int currentGenomeNr, final Map<kmer, Integer> kmers, final boolean allowSinglePositions,
                                                          final int extension, final int maxDistance, final int maxMergingDistance, final double fractPositions, final int maxFrequency, final int flanking,
                                                          final boolean fast)
            throws ExecutionException, InterruptedException {
        Pantools.logger.debug("Extension for genome {}: {} bp.", currentGenomeNr, extension);

        // Determine the set of kmers of the range on this other genome
        final Set<Integer[]> filteredKmers = GraphUtils.getKmersBelowFrequency(kmers, currentGenomeNr, maxFrequency); // [sequence, position]
        if (filteredKmers.isEmpty()) {
            Pantools.logger.warn("Could not find k-mers for genome {}; try increasing the number of sampled k-mers or maximum frequency.", currentGenomeNr);
            return null;
        }

        // Run the finding block algorithm
        Pantools.logger.debug("Each block must contain at least {}% of positions within {} bp distance of each other.", fractPositions * 100, maxDistance);
        final List<Set<Integer[]>> blocks = findBlocks(filteredKmers, maxDistance, fractPositions, allowSinglePositions);
        Pantools.logger.debug("Found {} blocks for genome {}", blocks.size(), currentGenomeNr);

        // Return the graph range
        final GraphRange graphRange = getRangeFromBlocks(currentGenomeNr, blocks, maxMergingDistance, flanking, extension + flanking, fast);
        Pantools.logger.debug("Graph range for genome {}: {}", currentGenomeNr, graphRange);
        return graphRange;
    }

    /**
     * Obtain the graph range from a list of blocks. The range will contain paths from the current genome.
     * @param currentGenomeNr current genome number
     * @param blocks list of sets of sorted positions
     * @param maxMergingDistance maximum distance between two blocks to consider them the same region (and thus merge them)
     * @param startExtension number of flanking bases to include at the start of each block
     * @param endExtension number of flanking bases to include at the end of each block
     * @return a GraphRange object containing the nucleotide nodes
     */
    private static GraphRange getRangeFromBlocks(final int currentGenomeNr, final List<Set<Integer[]>> blocks, final int maxMergingDistance, final int startExtension, final int endExtension, final boolean fast) {
        // Define the blocks and extend them
        final Set<Integer[]> extendedBlocks = getExtendedBlocks(currentGenomeNr, blocks, startExtension, endExtension);
        Pantools.logger.trace("#extended blocks: {}", extendedBlocks.size());

        // Merge overlapping blocks
        final Set<Integer[]> mergedBlocks = getMergedBlocks(extendedBlocks, maxMergingDistance);
        Pantools.logger.trace("#merged blocks: {}", mergedBlocks.size());

        // Create a graph range for each block and add to the final graph range
        final GraphRange graphRange = new GraphRange(new HashSet<>(), new HashSet<>());
        for (Integer[] mergedBlock : mergedBlocks) {
            final GenomeRange currentGenomeRange = new GenomeRange(currentGenomeNr, mergedBlock[0], mergedBlock[1], mergedBlock[2], false); //don't validate genome range
            final HashSet<GenomeRange> currentGenomeRanges = new HashSet<>(Collections.singletonList(currentGenomeRange));
            if (fast) {
                graphRange.add(new GraphRange(null, currentGenomeRanges));
            } else {
                final List<Long> currentGenomeNucleotideNodes = GraphUtils.getNodeIdsForGenomeRange(currentGenomeRange);
                graphRange.add(new GraphRange(new HashSet<>(currentGenomeNucleotideNodes), currentGenomeRanges));
            }
        }

        // Log what was done
        Pantools.logger.info("Found {} blocks for genome {} after extending and merging.", mergedBlocks.size(), currentGenomeNr);

        // Return the graph range
        return graphRange;
    }

    /**
     * Extend the blocks with a certain number of bases at the start and end.
     * @param currentGenomeNr current genome number (needed for maximum sequence length)
     * @param blocks list of sets of sorted positions
     * @param startExtension number of flanking bases to include at the start of each block
     * @param endExtension number of flanking bases to include at the end of each block
     * @return a set of extended blocks
     */
    private static Set<Integer[]> getExtendedBlocks(int currentGenomeNr, List<Set<Integer[]>> blocks, int startExtension, int endExtension) {
        final Set<Integer[]> extendedBlocks = new TreeSet<>(comparePositions());
        for (Set<Integer[]> block : blocks) {
            // Get the first and last position within the block
            final Integer[] first = block.stream().min(Comparator.comparingInt(o -> o[1])).orElse(null);
            final Integer[] last = block.stream().max(Comparator.comparingInt(o -> o[1])).orElse(null);
            // Add the range to the graph range
            if (first != null) {
                final int sequenceNr = first[0];
                final int start = Math.max(1, first[1] - startExtension);
                final int end = (int) Math.min(Globals.GENOME_DB.sequence_length[currentGenomeNr][sequenceNr], last[1] + endExtension);
                extendedBlocks.add(new Integer[]{sequenceNr, start, end});
            }
        }
        return extendedBlocks;
    }

    /**
     * Merge overlapping blocks.
     * @param extendedBlocks set of extended blocks (sequence, start, end) that may overlap
     * @param maxMergingDistance maximum distance between two blocks to consider them the same region (and thus merge them)
     * @return a set of merged blocks
     */
    private static Set<Integer[]> getMergedBlocks(final Set<Integer[]> extendedBlocks, final int maxMergingDistance) {
        final Set<Integer[]> mergedBlocks = new TreeSet<>(comparePositions());
        Integer[] previousBlock = null;
        for (Integer[] currentBlock : extendedBlocks) {
            if (previousBlock == null) {
                previousBlock = currentBlock;
            } else {
                if (previousBlock[0].equals(currentBlock[0]) && previousBlock[2] >= currentBlock[1] - maxMergingDistance) {
                    previousBlock = new Integer[]{previousBlock[0], previousBlock[1], Math.max(previousBlock[2], currentBlock[2])};
                } else {
                    mergedBlocks.add(previousBlock);
                    previousBlock = currentBlock;
                }
            }
        }
        if (previousBlock != null) {
            mergedBlocks.add(previousBlock);
        }
        return mergedBlocks;
    }

    /**
     * Find blocks of positions.
     * <br>
     * A block is a set of positions with the following characteristics:
     * <ul>
     *     <li>They are sorted by sequence and start position</li>
     *     <li>They are no more than a certain distance apart</li>
     *     <li>They contain at least a certain number of positions</li>
     * </ul>
     * @param positions set of positions (sequence, start[, end])
     * @param maxDistance maximum distance between two consecutive positions
     * @param fractPositions fraction of total positions that need to be found to create a hit
     * @param allowSinglePosition whether to allow blocks with only one position
     * @return a list of sets of sorted positions
     */
    static List<Set<Integer[]>> findBlocks(final Set<Integer[]> positions, final int maxDistance, final double fractPositions, final boolean allowSinglePosition) {
        // Order the position first by sequence and then by start position
        Set<Integer[]> sortedPositions = new TreeSet<>(comparePositions());
        sortedPositions.addAll(positions);

        // Calculate minPositions from fractPositions by rounding up (but at least 2 or 1 of there is only 1 position)
        final int minPositions = Math.max(Math.min(allowSinglePosition ? 1 : 2, positions.size()), (int) Math.ceil(fractPositions * positions.size()));

        // Find blocks of positions: a block is a set of positions that are no more than a certain distance apart
        List<Set<Integer[]>> blocks = new ArrayList<>();
        Set<Integer[]> block = new TreeSet<>(comparePositions());
        Integer[] previousPosition = null;
        for (Integer[] currentPosition : sortedPositions) {
            if (previousPosition == null) { // first position only
                block.add(currentPosition.length == 3 ? new Integer[]{currentPosition[0], currentPosition[1]} : currentPosition); //to deal with optional end position
            } else {
                if (currentPosition[0].equals(previousPosition[0]) && currentPosition[1] - previousPosition[1] <= maxDistance) {
                    block.add(currentPosition.length == 3 ? new Integer[]{currentPosition[0], currentPosition[2]} : currentPosition); //to deal with optional end position
                } else {
                    blocks.add(block);
                    block = new TreeSet<>(comparePositions());
                    block.add(currentPosition.length == 3 ? new Integer[]{currentPosition[0], currentPosition[1]} : currentPosition); //to deal with optional end position
                }
            }
            previousPosition = currentPosition.length == 3 ? new Integer[]{currentPosition[0], currentPosition[2]} : currentPosition; //to deal with optional end position
        }
        blocks.add(block);

        // Log the blocks
        if (Pantools.logger != null) {
            Pantools.logger.debug("Found {} blocks before removing blocks with less than {} positions:", blocks.size(), minPositions);
            for (int i = 0; i < blocks.size(); i++) {
                Pantools.logger.debug("\tBlock {}; positions: {}", i, blocks.get(i).size());
            }
        }

        // Remove blocks that are too small
        blocks.removeIf(b -> b.size() < minPositions);

        // Return the blocks
        return blocks;
    }

    /**
     * Comparator for positions (first by sequence, then by start position).
     * @return a comparator for positions
     */
    static Comparator<Integer[]> comparePositions() {
        return (o1, o2) -> {
            int sequenceComparison = o1[0].compareTo(o2[0]);
            if (sequenceComparison == 0) {
                return o1[1].compareTo(o2[1]);
            } else {
                return sequenceComparison;
            }
        };
    }

    /**
     * Main method for searching a GenomeRange in a pangenome by obtaining a GraphRange. This GraphRange will contain the
     * relevant nucleotide nodes as well as the found genome ranges in the pangenome.
     * @param originalGenomeRange genome range in the format <genomeNr>#<sequence>:<start>-<end>
     * @param useKmers whether to use k-mers for determining the range
     * @param useHomology whether to use homology groups for determining the range
     * @param useBlast whether to use BLAST results for determining the range
     * @param allowSinglePositions whether to allow blocks with only one transcript/k-mer matching
     * @param searchSelf whether to search in the input genome
     * @param kmerDensity k-mer density for k-mer sampling (distance between k-mers)
     * @param maxEvalue maximum E-value for BLAST hits
     * @param maxDistance maximum distance between two transcripts/k-mers; this determines the maximum gap that can be spanned within one block
     * @param fractPositions the fraction of total k-mers/transcripts that need to be found to create a hit
     * @param maxFrequency maximum frequency of a transcript or k-mer in a target genome
     * @param flanking number of flanking bases to include in the output
     * @return a GraphRange object containing the nucleotide nodes
     */
    public static GraphRange getRangeFromGenomeRange(final GenomeRange originalGenomeRange, final boolean useKmers, final boolean useHomology, final boolean useBlast, final boolean allowSinglePositions,
                                                     final boolean searchSelf, final int kmerDensity, final double maxEvalue, final int maxDistance, final double fractPositions, final int maxFrequency, final int flanking,
                                                     final boolean fast) {
        Pantools.logger.debug("Obtaining range from genome range {}", originalGenomeRange);
        GraphRange graphRange;

        // make sure that at least one of useKmers, useHomology, or useBlast is true
        if (!useKmers && !useHomology && !useBlast) {
            Pantools.logger.error("At least one of useKmers, useHomology or useBlast must be true.");
            throw new IllegalArgumentException("Impossible method combination");
        }

        // warn the user if the maxDistance or maxMergingDistance is larger than the genome range
        if (maxDistance > originalGenomeRange.getLength()) {
            Pantools.logger.warn("The maximum distance between two positions ({}) is larger than the genome range ({}); this could lead to unwanted trailing sequences.", maxDistance, originalGenomeRange.getLength());
        }

        // set maxMergingDistance to the length of the original genome range
        final int maxMergingDistance = originalGenomeRange.getLength();

        // obtain genome number from the original genome range
        final int genomeNr = originalGenomeRange.getGenomeNr();
        final int sequenceNr = originalGenomeRange.getSequenceNr();
        final int start = originalGenomeRange.getStart();
        final int end = originalGenomeRange.getEnd();

        // create a set of genome ranges and fill with the original genome range including flanking
        final Set<GenomeRange> genomeRanges = new HashSet<>();
        final int adjustedStart = Math.max(1, start - flanking);
        final int adjustedEnd = (int) Math.min(Globals.GENOME_DB.sequence_length[genomeNr][sequenceNr], end + flanking);
        final GenomeRange genomeRangeIncludingFlanking = new GenomeRange(genomeNr,
                sequenceNr,
                adjustedStart,
                adjustedEnd,
                false); //don't validate genome range since we just checked it
        genomeRanges.add(genomeRangeIncludingFlanking);

        try (Transaction ignored = Globals.GRAPH_DB.beginTx();
             ResourceIterator<Node> genomeNodeIterator = Globals.GRAPH_DB.findNodes(Globals.GENOME_LABEL)) {
            final Set<Node> genomeNodes = new HashSet<>();
            genomeNodeIterator.forEachRemaining(genomeNodes::add);

            // use the skip array to select the genomes to use
            skip.create_skip_arrays(false, false); // create skip array if -skip/-ref is provided by user
            skip.retrieveSelectedGenomes(); // sets selectedGenomes
            Pantools.logger.debug("Selected genomes: {}", Globals.selectedGenomes);

            // check if the genome is in the selected genomes
            if (Globals.selectedGenomes.contains(genomeNr)) {
                // find all nodes that are part of the genome range
                final Set<Long> nucleotideNodeIds = fast ? null : new HashSet<>(GraphUtils.getNodeIdsForGenomeRange(genomeRangeIncludingFlanking));

                // create initial graph range
                graphRange = new GraphRange(nucleotideNodeIds, genomeRanges);
            } else {
                Pantools.logger.warn("Genome {} is not in the selected genomes.", genomeNr);
                graphRange = new GraphRange();
            }

            // add ranges based on k-mer search
            if (useKmers) {
                Pantools.logger.info("Obtaining range based on k-mers.");

                // Determine the number of k-mers based on kmerDensity, which is the distance between the k-mers.
                // Therefore, the number of k-mers is the length of the genome range divided by the kmerDensity.
                final int numKmers = (int) Math.ceil((double) originalGenomeRange.getLength() / kmerDensity);
                final Map<kmer, Integer> kmers = originalGenomeRange.getKmers(numKmers);
                Pantools.logger.info("Sampled {} k-mers from input genome range.", numKmers);
                Pantools.logger.debug("K-mers (key=k-mer, value=position): {}", kmers);

                for (Node currentGenomeNode : genomeNodes) {
                    final int currentGenomeNr = (int) currentGenomeNode.getProperty("number");
                    if (currentGenomeNr == genomeNr && !searchSelf) {
                        continue;
                    }
                    if (Globals.selectedGenomes.contains((int) currentGenomeNode.getProperty("number"))) {
                        graphRange.add(getRangeFromKmerBlockMethod(currentGenomeNr, kmers, allowSinglePositions,
                                originalGenomeRange.getKmerStep(numKmers), maxDistance, maxMergingDistance, fractPositions, maxFrequency, flanking, fast));
                    }
                }
            }

            // add ranges based on homology search
            if (useHomology) {
                Pantools.logger.info("Obtaining range based on homology groups.");

                // Find all homology groups that are part of the genome range in all genomes
                final HashSet<Node> hmGroups = GraphUtils.getHomologyGroupsForGenomeRange(originalGenomeRange);
                for (Node currentGenomeNode : genomeNodes) {
                    final int currentGenomeNr = (int) currentGenomeNode.getProperty("number");
                    if (currentGenomeNr == genomeNr && !searchSelf) {
                        continue;
                    }
                    if (Globals.selectedGenomes.contains((int) currentGenomeNode.getProperty("number"))) {
                        graphRange.add(addRangeFromHomologyBlockMethod(currentGenomeNr, hmGroups, allowSinglePositions, maxDistance, maxMergingDistance, fractPositions, maxFrequency, flanking, fast));
                    }
                }
            }

            // add ranges based on BLAST search
            if (useBlast) {
                Pantools.logger.info("Obtaining range based on BLAST results.");
                Pantools.logger.warn("VERY EXPERIMENTAL: This method is still under development and may not work as expected.");

                // Run TBLASTN with all proteins in the genome range
                final HashSet<Node> mrnaNodes = GraphUtils.getMrnasForGenomeRange(originalGenomeRange);
                Path blastInputPath = FileUtils.createScratchDirectory(null).resolve("blast_input.fasta");
                GraphUtils.writeProteinToFasta(mrnaNodes, blastInputPath);
                Pantools.logger.debug("Wrote {} proteins to temporary file ({}) for TBLASTN.", mrnaNodes.size(), blastInputPath);
                HashMap<Integer, HashSet<GenomeRange>> blastHitsPerGenome = Utils.performTblastn(blastInputPath, 0, 0, maxEvalue, maxFrequency);
                if (blastHitsPerGenome.isEmpty()) {
                    Pantools.logger.warn("No BLAST hits found; perhaps the input is too small? (see here: {})", blastInputPath);
                } else {
                    FileUtils.deleteDirectoryRecursively(blastInputPath.getParent());
                }

                // Find all BLAST protein results that are part of the genome range in all genomes (tblastn)
                for (Node currentGenomeNode : genomeNodes) {
                    final int currentGenomeNr = (int) currentGenomeNode.getProperty("number");
                    if (currentGenomeNr == genomeNr && !searchSelf) {
                        continue;
                    }
                    if (Globals.selectedGenomes.contains((int) currentGenomeNode.getProperty("number"))) {
                        graphRange.add(addRangeFromBlastBlockMethod(currentGenomeNr, blastHitsPerGenome.getOrDefault(currentGenomeNr, new HashSet<>()), allowSinglePositions, maxDistance, maxMergingDistance, fractPositions, flanking, fast));
                    }
                }
            }
        } catch (NotFoundException nfe) {
            Pantools.logger.error("Could not find the genome nodes in the graph database.");
            throw new RuntimeException("Invalid graph structure", nfe);
        } catch (ExecutionException | InterruptedException e) {
            Pantools.logger.error("Multithreading error.");
            throw new RuntimeException("Multithreading error", e);
        } catch (IOException e) {
            Pantools.logger.error("Could not create temporary file for BLAST input.");
            throw new RuntimeException("File system error", e);
        }

        // return the range
        return graphRange;
    }

    /**
     * Obtain the graph range for the homologues of a gene.
     * @param genomeRange genome range of the gene
     * @param flanking number of flanking bases to include in the output
     * @return a GraphRange object containing the nucleotide nodes
     */
    private static GraphRange getRangeForGene(final GenomeRange genomeRange, final int flanking, final boolean fast) {
        Pantools.logger.debug("Obtaining range for gene at {}", genomeRange);
        GraphRange graphRange = null;

        // make sure that genomeRange has been set
        if (genomeRange == null) {
            throw new IllegalStateException("No genome range has been set.");
        }

        // for each genome, find the members of the homology group and add them to the graph range
        try (Transaction ignored = Globals.GRAPH_DB.beginTx()) {
            // use the skip array to select the genomes to use
            skip.create_skip_arrays(false, false); // create skip array if -skip/-ref is provided by user
            skip.retrieveSelectedGenomes(); // sets selectedGenomes
            Pantools.logger.debug("Selected genomes: {}", Globals.selectedGenomes);

            final Set<Node> homologyGroups = GraphUtils.getHomologyGroupsForGenomeRange(genomeRange);
            for (Node homologyGroup : homologyGroups) { // homology_group -[has_homolog]-> mrna
                for (Relationship rel : homologyGroup.getRelationships(Globals.RelTypes.has_homolog, Direction.OUTGOING)) {
                    final Node mrnaNode = rel.getEndNode();
                    final int[] mrnaAddress = (int[]) mrnaNode.getProperty("address");

                    // skip if the genome is not selected
                    if (!Globals.selectedGenomes.contains(mrnaAddress[0])) {
                        continue;
                    }

                    // get the genome range for the mRNA
                    final GenomeRange mrnaGenomeRange = new GenomeRange(
                            mrnaAddress[0],
                            mrnaAddress[1],
                            Math.max(1, mrnaAddress[2] - flanking),
                            (int) Math.min(Globals.GENOME_DB.sequence_length[mrnaAddress[0]][mrnaAddress[1]], mrnaAddress[3] + flanking),
                            false); //don't validate genome range since we make sure it is valid
                    final Set<Long> mrnaNucleotideNodes = fast ? null : new HashSet<>(GraphUtils.getNodeIdsForGenomeRange(mrnaGenomeRange));

                    // add the range to the graph range
                    if (graphRange == null) {
                        graphRange = new GraphRange(mrnaNucleotideNodes, new HashSet<>(Collections.singletonList(mrnaGenomeRange)));
                    } else {
                        graphRange.add(new GraphRange(mrnaNucleotideNodes, new HashSet<>(Collections.singletonList(mrnaGenomeRange))));
                    }
                }
            }
        } catch (NotFoundException nfe) {
            Pantools.logger.error("Could not find the genome nodes in the graph database.");
            throw new RuntimeException("Invalid graph structure", nfe);
        }

        // return the range
        return graphRange;
    }

    /**
     * Obtain the graph range for the homologues of multiple genes.
     * @param genes list of gene names
     * @param context number of context genes to include
     * @param flanking number of flanking bases to include in the output
     * @return a GraphRange object containing the nucleotide nodes
     */
    public static GraphRange getRangeForGenes(final String[] genes, final int context, final int flanking, final boolean fast) {
        Pantools.logger.debug("Obtaining range for genes: {}", Arrays.toString(genes));
        GraphRange graphRange = null;

        // make sure that genes has been set
        if (genes == null) {
            throw new IllegalStateException("No genes have been set.");
        }

        // for each gene, find the genome range and add the result of getRangeForGene to the graph range
        for (String gene : genes) {
            final GenomeRange genomeRange = GraphUtils.getGenomeRangeForGene(gene, context);

            final GraphRange geneGraphRange = getRangeForGene(genomeRange, flanking, fast);
            if (graphRange == null) {
                graphRange = geneGraphRange;
            } else {
                graphRange.add(geneGraphRange);
            }
        }

        // return the range
        return graphRange;
    }

    /**
     * Obtain the graph range for homologues.
     * @param hmGroups list of homology group node identifiers
     * @param flanking number of flanking bases to include in the output
     * @return a GraphRange object containing the nucleotide nodes
     */
    public static GraphRange getRangeForHmGroups(final List<Long> hmGroups, final int flanking, final boolean fast) {
        Pantools.logger.debug("Obtaining range for homology groups: {}", hmGroups);
        GraphRange graphRange = null;

        // make sure that homology groups are defined
        if (hmGroups == null) {
            throw new IllegalStateException("No genes have been set.");
        }

        // for each homology group, find the genome range and add the result of getRangeForGene to the graph range
        try (Transaction ignored = Globals.GRAPH_DB.beginTx()) {
            // use the skip array to select the genomes to use
            skip.create_skip_arrays(false, false); // create skip array if -skip/-ref is provided by user
            skip.retrieveSelectedGenomes(); // sets selectedGenomes
            Pantools.logger.debug("Selected genomes: {}", Globals.selectedGenomes);

            final List<Node> homologyGroups = Utils.findHmNodes(hmGroups, 1);
            for (Node homologyGroup : homologyGroups) { // homology_group -[has_homolog]-> mrna
                for (Relationship rel : homologyGroup.getRelationships(Globals.RelTypes.has_homolog, Direction.OUTGOING)) {
                    final Node mrnaNode = rel.getEndNode();
                    final int[] mrnaAddress = (int[]) mrnaNode.getProperty("address");

                    // skip if the genome is not selected
                    if (!Globals.selectedGenomes.contains(mrnaAddress[0])) {
                        continue;
                    }

                    // get the genome range for the mRNA
                    final GenomeRange mrnaGenomeRange = new GenomeRange(
                            mrnaAddress[0],
                            mrnaAddress[1],
                            Math.max(1, mrnaAddress[2] - flanking),
                            (int) Math.min(Globals.GENOME_DB.sequence_length[mrnaAddress[0]][mrnaAddress[1]], mrnaAddress[3] + flanking),
                            false); //don't validate genome range since we make sure it is valid
                    final Set<Long> mrnaNucleotideNodes = fast ? null : new HashSet<>(GraphUtils.getNodeIdsForGenomeRange(mrnaGenomeRange));

                    // add the range to the graph range
                    if (graphRange == null) {
                        graphRange = new GraphRange(mrnaNucleotideNodes, new HashSet<>(Collections.singletonList(mrnaGenomeRange)));
                    } else {
                        graphRange.add(new GraphRange(mrnaNucleotideNodes, new HashSet<>(Collections.singletonList(mrnaGenomeRange))));
                    }
                }
            }
        } catch (NotFoundException nfe) {
            Pantools.logger.error("Could not find the genome nodes in the graph database.");
            throw new RuntimeException("Invalid graph structure", nfe);
        }

        // return the range
        return graphRange;
    }

    /**
     * Obtain the graph range based on functional annotations.
     * @param functions list of functional annotations
     * @param flanking number of flanking bases to include in the output
     * @return a GraphRange object containing the nucleotide nodes
     */
    public static GraphRange getRangeForFunctions(final String[] functions, final int flanking, final boolean fast) {
        Pantools.logger.debug("Obtaining range for functional annotations: {}", Arrays.toString(functions));
        GraphRange graphRange = null;

        // make sure that functions are defined
        if (functions == null) {
            throw new IllegalStateException("No functional annotations have been set.");
        }

        try (Transaction ignored = Globals.GRAPH_DB.beginTx()) {
            // use the skip array to select the genomes to use
            skip.create_skip_arrays(false, false); // create skip array if -skip/-ref is provided by user
            skip.retrieveSelectedGenomes(); // sets selectedGenomes
            Pantools.logger.debug("Selected genomes: {}", Globals.selectedGenomes);

            for (String function : functions) {
                // for each function, find the genes that contain it
                final Set<GenomeRange> genomeRanges = GraphUtils.getGenomeRangeForFunction(function);
                Pantools.logger.info("Found {} transcripts for function {}", genomeRanges.size(), function);

                // extend each genome range with flanking bases and add to the graph range
                for (GenomeRange genomeRange : genomeRanges) {
                    // skip if the genome is not selected
                    if (!Globals.selectedGenomes.contains(genomeRange.getGenomeNr())) {
                        continue;
                    }

                    // get the genome range for the mRNA
                    final GenomeRange genomeRangeIncludingFlanking = new GenomeRange(genomeRange.getGenomeNr(),
                            genomeRange.getSequenceNr(),
                            Math.max(1, genomeRange.getStart() - flanking),
                            (int) Math.min(Globals.GENOME_DB.sequence_length[genomeRange.getGenomeNr()][genomeRange.getSequenceNr()], genomeRange.getEnd() + flanking),
                            false); //don't validate genome range since we make sure it is valid
                    final Set<Long> nucleotideNodes = fast ? null : new HashSet<>(GraphUtils.getNodeIdsForGenomeRange(genomeRangeIncludingFlanking));

                    // add the range to the graph range
                    if (graphRange == null) {
                        graphRange = new GraphRange(nucleotideNodes, new HashSet<>(Collections.singletonList(genomeRangeIncludingFlanking)));
                    } else {
                        graphRange.add(new GraphRange(nucleotideNodes, new HashSet<>(Collections.singletonList(genomeRangeIncludingFlanking))));
                    }
                }
            }
        } catch (NotFoundException nfe) {
            Pantools.logger.error("Could not find the genome nodes in the graph database.");
            throw new RuntimeException("Invalid graph structure", nfe);
        }

        // return the range
        return graphRange;
    }

    /**
     * Obtain k-mers in the graph range in 2bit format using neo4j.
     * @return sorted array of k-mers in 2bit format (long)
     */
    public long[] getKmersAs2bit() {
        if (this.nucleotideNodes == null) {
            throw new IllegalStateException("Nucleotide nodes are not set.");
        }

        final List<Long> kmers = new ArrayList<>();
        try (Transaction ignored = GRAPH_DB.beginTx()) {
            for (Long nucleotideNodeId : this.nucleotideNodes) {
                final Node nucleotideNode = GRAPH_DB.getNodeById(nucleotideNodeId);
                for (long kmer : Utils.kmerizeAs2bit((String) nucleotideNode.getProperty("sequence"), Globals.K_SIZE)) {
                    kmers.add(Math.min(kmer, Utils.reverseComplement(kmer))); // add minimum of k-mer and reverse complement to list
                }
            }
        }

        Pantools.logger.debug("Found {} distinct k-mers for graph range.", kmers.size());
        Collections.sort(kmers);

        // return the array
        return ArrayUtils.toPrimitive(kmers.toArray(new Long[0]));
    }

    /**
     * Obtain k-mers in the graph range in 2bit format using neo4j.
     * @param minFrequency The minimum frequency of a k-mer to be included
     * @return sorted array of k-mers in 2bit format (long)
     */
    public long[] getKmersAs2bit(final int minFrequency) {
        if (this.nucleotideNodes == null) {
            throw new IllegalStateException("Nucleotide nodes are not set.");
        }

        final List<Long> kmers = new ArrayList<>();
        try (Transaction ignored = GRAPH_DB.beginTx()) {
            for (Long nucleotideNodeId : this.nucleotideNodes) {
                final Node nucleotideNode = GRAPH_DB.getNodeById(nucleotideNodeId);
                if ((long) nucleotideNode.getProperty("frequency") >= minFrequency) {
                    for (long kmer : Utils.kmerizeAs2bit((String) nucleotideNode.getProperty("sequence"), Globals.K_SIZE)) {
                        kmers.add(Math.min(kmer, Utils.reverseComplement(kmer))); // add minimum of k-mer and reverse complement to list
                    }
                }
            }
        }

        Pantools.logger.debug("Found {} distinct k-mers for graph range (min frequency = {}).", kmers.size(), minFrequency);
        Collections.sort(kmers);

        // return the array
        return ArrayUtils.toPrimitive(kmers.toArray(new Long[0]));
    }

    /**
     * Obtain the genome ranges of the range.
     * @return a set of genome ranges
     */
    public Set<GenomeRange> getGenomeRanges() {
        return genomeRanges;
    }

    /**
     * Obtain the nucleotide nodes of the range.
     * @return a set of node IDs
     * @throws IllegalStateException if the nucleotide nodes are not set
     */
    public Set<Long> getNucleotideNodes() throws IllegalStateException {
        if (nucleotideNodes == null) {
            throw new IllegalStateException("Nucleotide nodes are not set.");
        }
        return nucleotideNodes;
    }

    /**
     * Whether the range contains a specific node identifier.
     * @param nodeId node identifier to check
     * @return true if the range contains the node identifier, false otherwise
     */
    public boolean contains(long nodeId) {
        return nucleotideNodes != null && nucleotideNodes.contains(nodeId);
    }

    /**
     * Whether the range contains an exact specific genome range
     * @param genomeRange genome range to check
     * @return true if the range contains the genome range, false otherwise
     */
    public boolean contains(GenomeRange genomeRange) {
        for (GenomeRange range : genomeRanges) {
            if (range.overlaps(genomeRange)) {
                return true;
            }
        }
        return false;
    }

    /**
     * Whether the range overlaps (partially or completely) with a specific genome range
     * @param genomeRange genome range to check
     * @return true if the graph range overlaps with the genome range, false otherwise
     */
    public boolean overlaps(GenomeRange genomeRange) {
        for (GenomeRange range : genomeRanges) {
            if (range.overlaps(genomeRange)) {
                return true;
            }
        }
        return false;
    }

    /**
     * toString method.
     * @return A string representation of this object.
     */
    @Override
    public String toString() {
        return String.format("GraphRange{nucleotideNodes=%s, genomeRanges=%s}", this.nucleotideNodes, this.genomeRanges);
    }
}
