package nl.wur.bif.pantools.utils.reader;

import htsjdk.samtools.fastq.FastqRecord;
import nl.wur.bif.pantools.Pantools;

import java.io.Closeable;
import java.io.IOException;
import java.nio.channels.Channels;
import java.nio.file.*;
import java.nio.ByteBuffer;
import java.nio.channels.FileChannel;
import java.util.Iterator;
import java.util.NoSuchElementException;
import java.util.concurrent.BlockingQueue;
import java.util.concurrent.LinkedBlockingQueue;
import java.util.zip.GZIPInputStream;

/**
 * FastqReader is a class that reads Fastq files in the fastest way possible to my knowledge.
 * It uses NIO to read the file and a BlockingQueue to store the records.
 * The records are read in a separate thread to allow for parallel processing of the records.
 * NB: most of this class was written with the help of claude.ai and two days of prompting.
 */
public class FastqReader implements Iterator<FastqRecord>, Iterable<FastqRecord>, Closeable {
    private static final int BUFFER_SIZE = 4 * 1024 * 1024; // 4MB buffer
    private static final int QUEUE_SIZE = 1_000_000;
    private static final FastqRecord POISON_PILL = new FastqRecord("", "", "", "");

    private final Path filePath;
    private FastqRecord nextRecord;
    private final BlockingQueue<FastqRecord> queue;
    private final Thread readerThread;
    private volatile boolean closed = false;

    public FastqReader(Path filePath) {
        this.filePath = filePath;
        this.queue = new LinkedBlockingQueue<>(QUEUE_SIZE);

        // Create and start the reader thread
        this.readerThread = new Thread(this::processFile);
        this.readerThread.setDaemon(true);
        this.readerThread.start();

        // Get the first record
        this.nextRecord = this.readNextRecord();
    }

    /**
     * Process the file and add the records to the queue.
     */
    private void processFile() {
        try {
            final boolean isGzipped = filePath.toString().endsWith(".gz");
            if (isGzipped) {
                this.readFileGzipNIO();
            } else {
                this.readFileNIO();
            }
        } catch (IOException | InterruptedException e) {
            Pantools.logger.error("Error processing the file: {}", e.getMessage());
        } finally {
            try {
                queue.put(POISON_PILL);
            } catch (InterruptedException e) {
                Pantools.logger.error("Error adding end marker to queue: {}", e.getMessage());
            }
        }
    }

    @Override
    public boolean hasNext() {
        return this.nextRecord != null && this.nextRecord != POISON_PILL;
    }

    @Override
    public FastqRecord next() {
        if (!this.hasNext()) {
            throw new NoSuchElementException("next() called when !hasNext()");
        }
        FastqRecord record = this.nextRecord;
        this.nextRecord = this.readNextRecord();
        return record;
    }

    @Override
    public Iterator<FastqRecord> iterator() {
        return this;
    }

    @Override
    public void close() {
        if (!closed) {
            closed = true;
            readerThread.interrupt();
            queue.clear();
        }
    }

    /**
     * Read the next record from the queue.
     * @return The next FastqRecord or an empty FastqRecord if the queue is empty.
     */
    private FastqRecord readNextRecord() {
        try {
            return queue.take();
        } catch (InterruptedException e) {
            Pantools.logger.error("Error reading the next record: {}", e.getMessage());
            return POISON_PILL;
        }
    }

    /**
     * Read a Fastq file using NIO.
     * @throws IOException If an I/O error occurs
     * @throws InterruptedException If the thread is interrupted
     */
    private void readFileGzipNIO() throws IOException, InterruptedException {
        try (FileChannel fileChannel = FileChannel.open(this.filePath, StandardOpenOption.READ);
             GZIPInputStream gzipInputStream = new GZIPInputStream(Channels.newInputStream(fileChannel))) {

            StringBuilder currentLine = new StringBuilder();
            String[] readComponents = new String[4];

            int lineCount = 0;
            byte lastByte = 0;

            ByteBuffer buffer = ByteBuffer.allocate(BUFFER_SIZE);
            int bytesRead;
            byte[] byteArray = new byte[BUFFER_SIZE];

            while ((bytesRead = gzipInputStream.read(byteArray)) != -1) {
                buffer.put(byteArray, 0, bytesRead);
                buffer.flip();
                while (buffer.hasRemaining()) {
                    byte currentByte = buffer.get();
                    if (currentByte == '\n' && lastByte != '\r') {
                        // Process completed line
                        readComponents[lineCount & 3] = currentLine.toString();
                        currentLine.setLength(0);

                        // When we've collected all 4 lines, create a FastqRecord object
                        if (lineCount % 4 == 3) {
                            addNewFastqRecord(readComponents);
                        }

                        lineCount++;
                    } else if (currentByte != '\r') {
                        currentLine.append((char) currentByte);
                    }
                    lastByte = currentByte;
                }
                buffer.clear();
            }

            // handle last bit
            if (currentLine.length() > 0) {
                readComponents[lineCount & 3] = currentLine.toString();
                if (lineCount % 4 == 3) {
                    addNewFastqRecord(readComponents);
                }
            }
        }
    }

    /**
     * Read a Fastq file using NIO.
     * @throws IOException If an I/O error occurs
     * @throws InterruptedException If the thread is interrupted
     */
    private void readFileNIO() throws IOException, InterruptedException {
        try (FileChannel fileChannel = FileChannel.open(this.filePath, StandardOpenOption.READ)) {
            ByteBuffer buffer = ByteBuffer.allocateDirect(BUFFER_SIZE);

            StringBuilder currentLine = new StringBuilder();
            String[] readComponents = new String[4];

            int lineCount = 0;
            byte lastByte = 0;

            while (fileChannel.read(buffer) > 0) {
                buffer.flip();
                while (buffer.hasRemaining()) {
                    byte currentByte = buffer.get();
                    if (currentByte == '\n' && lastByte != '\r') {
                        // Process completed line
                        readComponents[lineCount & 3] = currentLine.toString();
                        currentLine.setLength(0);

                        // When we've collected all 4 lines, create a FastqRecord object
                        if (lineCount % 4 == 3) {
                            addNewFastqRecord(readComponents);
                        }

                        lineCount++;
                    } else if (currentByte != '\r') {
                        currentLine.append((char) currentByte);
                    }
                    lastByte = currentByte;
                }
                buffer.clear();
            }

            // handle last bit
            if (currentLine.length() > 0) {
                readComponents[lineCount & 3] = currentLine.toString();
                if (lineCount % 4 == 3) {
                    addNewFastqRecord(readComponents);
                }
            }
        }
    }

    private void addNewFastqRecord(final String[] readComponents) throws InterruptedException {
        queue.put(new FastqRecord(
                readComponents[0].substring(1),
                readComponents[1],
                readComponents[2].substring(1),
                readComponents[3]
        ));
    }
}