package nl.wur.bif.pantools.utils.writer;

import nl.wur.bif.pantools.Pantools;
import nl.wur.bif.pantools.utils.Globals;
import nl.wur.bif.pantools.utils.GraphUtils;
import nl.wur.bif.pantools.utils.graph.GenomeRange;
import nl.wur.bif.pantools.utils.graph.graph_objects.GraphRange;
import org.neo4j.graphdb.Node;
import org.neo4j.graphdb.NotFoundException;
import org.neo4j.graphdb.Transaction;

import java.io.BufferedWriter;
import java.io.IOException;
import java.nio.file.Files;
import java.nio.file.Path;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

public class AnnotationCsvWriter extends AnnotationWriter {
    public AnnotationCsvWriter() {
        super();
    }

    /**
     * Implements the abstract method to write the annotation to a CSV file.
     * @param graphRange The graph range to write.
     * @param outputFile The output CSV file.
     * @param writeRepeats If true, write the repeat annotations. NB: NOT IMPLEMENTED
     * @param writeBlast If true, write the BLAST annotations. NB: NOT IMPLEMENTED
     * @throws IOException If an error occurs while writing the CSV file.
     */
    @Override
    public void write(GraphRange graphRange, Path outputFile, boolean writeRepeats, boolean writeBlast) throws IOException {
        Pantools.logger.info("Writing annotations to GFA graph to {}", outputFile);

        // log that repeat annotations are not implemented
        if (GraphUtils.containsRepeats() && writeRepeats) {
            Pantools.logger.warn("Writing repeat annotations is not implemented in CSV format.");
        }

        // log that blast annotations are not implemented
        if (writeBlast) {
            Pantools.logger.warn("Writing BLAST annotations is not implemented in CSV format.");
        }

        // create the directory of the file if it does not exist
        if (!Files.exists(outputFile.getParent())) {
            Files.createDirectories(outputFile.getParent());
        }

        // Write the CSV file
        try (BufferedWriter writer = Files.newBufferedWriter(outputFile)) {
            // write the header
            writeCsvHeader(writer);

            // write the annotations
            writeCsvAnnotations(writer, graphRange);
        }
    }

    /**
     * Write the header of the CSV file.
     * @param writer The writer to write the header to.
     * @throws IOException If an error occurs while writing the header.
     */
    private void writeCsvHeader(BufferedWriter writer) throws IOException {
        writer.write("NODE_ID,ANNOTATION,COLOR");
        writer.newLine();
    }

    /**
     * Write the all gene locations of a given range to the CSV file.
     * @param writer The writer to write the annotations to.
     * @param graphRange The graph range to search for annotations.
     * @throws IOException If an error occurs while writing the annotations.
     */
    private void writeCsvAnnotations(BufferedWriter writer, GraphRange graphRange) throws IOException {
        // get the nodes for all gene locations in the graph range
        Map<Long, List<String>> geneNucleotideNodes = getNodeIdsForGenes(graphRange); // nodeId -> unique gene ID (gene ID + "#" + genome number)

        // loop over all nucleotide nodes in the range and annotate those belonging to a gene
        for (Long nodeId : graphRange.getNucleotideNodes()) {
            if (geneNucleotideNodes.containsKey(nodeId)) {
                final List<String> geneIds = geneNucleotideNodes.get(nodeId);

                // combine the gene IDs into a single annotation
                final String annotation = geneIds.stream()
                        .sorted()
                        .distinct()
                        .reduce((s1, s2) -> s1 + ";" + s2)
                        .orElse("");

                // get the unique color for the annotation based on the first gene ID
                final String color = String.format("#%s", Integer.toHexString(geneIds.get(0).hashCode()));

                // write the annotation to the CSV file
                writer.write(String.format("%d,%s,%s", nodeId, annotation, color));
            } else {
                writer.write(String.format("%d,,", nodeId));
            }
            writer.newLine();
        }
    }

    /**
     * Get the node IDs for all genes in the given graph range. The gene IDs are unique and are formatted as gene ID + "#" + genome number.
     * @param graphRange The graph range to search for gene locations.
     * @return A map of node IDs to a list of unique gene IDs.
     */
    private Map<Long, List<String>> getNodeIdsForGenes(GraphRange graphRange) {
        Pantools.logger.debug("Getting nucleotide nodes for all genes in {}", graphRange);
        try (Transaction ignored = Globals.GRAPH_DB.beginTx()) {
            Map<Long, List<String>> geneNucleotideNodes = new HashMap<>();
            for (GenomeRange genomeRange : graphRange.getGenomeRanges()) {
                for (Node featureNode : genomeRange.getTreeGeneNodes()) {

                    // get the nucleotide nodes for the gene location and add the gene ID to the map
                    if (featureNode.hasLabel(Globals.GENE_LABEL)) {
                        final int[] address = (int[]) featureNode.getProperty("address");
                        for (long nodeId : GraphUtils.getNodeIdsForGenomeRange(new GenomeRange(address, false))) {
                            if (!geneNucleotideNodes.containsKey(nodeId)) {
                                geneNucleotideNodes.put(nodeId, new ArrayList<>());
                            }
                            Pantools.logger.trace("Found nucleotide node {} for gene {} (genome {})", nodeId, featureNode.getProperty("id"), address[0]);
                            geneNucleotideNodes.get(nodeId).add(String.format("%s#%d", featureNode.getProperty("id"), address[0]));
                        }
                    }
                }
            }
            return geneNucleotideNodes;
        } catch (NotFoundException nfe) {
            Pantools.logger.error("Error while getting gene nodes from the graph database", nfe);
            throw new RuntimeException(nfe);
        }
    }
}
