package nl.wur.bif.pantools.utils.writer;

import htsjdk.tribble.gff.Gff3Feature;
import htsjdk.tribble.gff.Gff3Writer;
import nl.wur.bif.pantools.Pantools;
import nl.wur.bif.pantools.utils.GraphUtils;
import nl.wur.bif.pantools.utils.graph.GenomeRange;
import nl.wur.bif.pantools.utils.graph.graph_objects.GraphRange;

import java.io.IOException;
import java.nio.file.Files;
import java.nio.file.Path;
import java.nio.file.Paths;

import static nl.wur.bif.pantools.utils.Globals.WORKING_DIRECTORY;

public class AnnotationGffWriter extends AnnotationWriter {
    public AnnotationGffWriter() {
        super();
    }

    /**
     * Implements the abstract method to write the annotation to GFF files.
     * @param graphRange The graph range to write.
     * @param outputDir The output directory.
     * @param writeRepeats Whether to write the repeats to a separate GFF file.
     * @param writeBlast Whether to write the BLAST results to a separate GFF file.
     * @throws IOException If an error occurs while creating the directory of the GFF files.
     */
    @Override
    public void write(GraphRange graphRange, Path outputDir, boolean writeRepeats, boolean writeBlast) throws IOException {
        Pantools.logger.info("Writing the graph as gff3 files to {}/", outputDir);

        // create the directory if it does not exist
        if (!Files.exists(outputDir)) {
            Files.createDirectories(outputDir);
        }

        // Go over each of the genomes in the genome ranges of the graph range and create a GFF file per genome
        for (int genomeNr : graphRange.getGenomeRanges().stream().mapToInt(GenomeRange::getGenomeNr).distinct().toArray()) {
            // write the genes to a GFF file
            final Path gffFile = outputDir.resolve(genomeNr + ".gff");
            Pantools.logger.debug("Writing GFF for genome '{}' to '{}'.", genomeNr, gffFile);

            // Write the GFF to the file
            try (Gff3Writer gff3Writer = new ModifiedGff3Writer(gffFile)) {
                for (GenomeRange genomeRange : graphRange.getGenomeRanges()) {
                    if (genomeRange.getGenomeNr() == genomeNr) {
                        gff3Writer.addComment(String.format("PanTools GFF3 file for genome range %s", genomeRange));
                        for (Gff3Feature gff3Feature : genomeRange.getGff3Features()) {
                            gff3Writer.addFeature(gff3Feature);
                        }
                    }
                }
            } catch (IOException e) {
                throw new RuntimeException(e);
            }

            // now write the repeats to a GFF file if available
            if (GraphUtils.containsRepeats() && writeRepeats) {
                final Path repeatGffFile = outputDir.resolve(genomeNr + "_repeats.gff");
                Pantools.logger.debug("Writing repeats for genome '{}' to '{}'.", genomeNr, repeatGffFile);

                // Write the GFF to the file
                try (Gff3Writer gff3Writer = new ModifiedGff3Writer(repeatGffFile)) {
                    for (GenomeRange genomeRange : graphRange.getGenomeRanges()) {
                        if (genomeRange.getGenomeNr() == genomeNr) {
                            gff3Writer.addComment(String.format("PanTools GFF3 file for repeats in genome range %s", genomeRange));
                            for (Gff3Feature gff3Feature : genomeRange.getRepeatGff3Features()) {
                                gff3Writer.addFeature(gff3Feature);
                            }
                        }
                    }
                } catch (IOException e) {
                    throw new RuntimeException(e);
                }
            }

            // now write the BLAST results to a GFF file if available
            final Path blastResultsPath = Paths.get(WORKING_DIRECTORY).resolve("BLAST").resolve("blast_filtered.tsv");
            if (Files.exists(blastResultsPath) && writeBlast) {
                final Path blastGffFile = outputDir.resolve(genomeNr + "_blast.gff");
                Pantools.logger.debug("Writing BLAST results for genome '{}' to '{}'.", genomeNr, blastGffFile);

                // Write the GFF to the file
                try (Gff3Writer gff3Writer = new ModifiedGff3Writer(blastGffFile)) {
                    for (GenomeRange genomeRange : graphRange.getGenomeRanges()) {
                        if (genomeRange.getGenomeNr() == genomeNr) {
                            gff3Writer.addComment(String.format("PanTools GFF3 file for TBLASTN hits in genome range %s", genomeRange));
                            for (Gff3Feature gff3Feature : genomeRange.getBlastGff3Features(blastResultsPath)) {
                                gff3Writer.addFeature(gff3Feature);
                            }
                        }
                    }
                } catch (IOException e) {
                    throw new RuntimeException(e);
                }
            }
        }
    }

    /**
     * Inner class to override the default encode method for Gff3Writer sequence names.
     */
    private static class ModifiedGff3Writer extends Gff3Writer {
        public ModifiedGff3Writer(Path path) throws IOException {
            super(path);
        }

        @Override
        protected String escapeString(String sequenceName) {
            return sequenceName;
        }
    }
}
