package nl.wur.bif.pantools.utils.writer;

import htsjdk.tribble.gff.Gff3Feature;
import nl.wur.bif.pantools.Pantools;
import nl.wur.bif.pantools.utils.Globals;
import nl.wur.bif.pantools.utils.GraphUtils;
import nl.wur.bif.pantools.utils.Utils;
import nl.wur.bif.pantools.utils.graph.GenomeRange;
import nl.wur.bif.pantools.utils.graph.graph_objects.GraphRange;
import org.json.JSONArray;
import org.json.JSONObject;
import org.neo4j.graphdb.*;

import java.nio.file.Paths;
import java.util.Random;

import java.io.BufferedWriter;
import java.io.IOException;
import java.nio.file.Files;
import java.nio.file.Path;
import java.util.*;
import java.util.stream.Collectors;

import static nl.wur.bif.pantools.utils.Globals.GENOME_DB;
import static nl.wur.bif.pantools.utils.Globals.WORKING_DIRECTORY;

public class CollinearityWriter {
    public CollinearityWriter() {}

    /**
     * Method to write a collinearity file to a path
     * @param graphRange the graph range to write
     * @param outputDir path to write the directory in which to write the collinearity file
     * @param writeRepeats whether to write repeats to a GFF file
     * @param writeBlast whether to write BLAST results to a GFF file
     * @throws IOException if an error occurs while writing the collinearity file
     */
    public void write(GraphRange graphRange, Path outputDir, boolean writeRepeats, boolean writeBlast) throws IOException {
        Pantools.logger.info("Writing the collinearity files to {}/", outputDir);

        // create the directory if it does not exist
        if (!Files.exists(outputDir)) {
            Files.createDirectories(outputDir);
        }

        // get a map of mRNA ID to genome range for the graph range
        Map<Long, GenomeRange> mrnaIdToGenomeRange = getMrnaIdToGenomeRange(graphRange); // mRNA ID -> genome range

        // write a single fai and gff3 file for the relevant graph range (which has mRNAs)
        writeFai(mrnaIdToGenomeRange, outputDir.resolve("all.fai"));
        writeGff3(mrnaIdToGenomeRange, outputDir.resolve("all.gff3"));

        // if available, write repeats to a GFF file
        if (GraphUtils.containsRepeats() && writeRepeats) {
            writeRepeatsGff3(graphRange, outputDir.resolve("all_repeats.gff3"));
        }

        // if available, write BLAST results to a GFF file
        final Path blastResultsPath = Paths.get(WORKING_DIRECTORY).resolve("BLAST").resolve("blast_filtered.tsv");
        if (Files.exists(blastResultsPath) && writeBlast) {
            writeBlastGff3(graphRange, blastResultsPath, outputDir.resolve("all_blast.gff3"));
        }

        // obtain all homology groups that have mRNAs in the given graphRange including location
        Map<Long, Set<Long>> hmGroups = getHmGroups(mrnaIdToGenomeRange); // homology group -> set of mRNA IDs in the graph range

        // write collinearity files for the graph range (m8, bedpe and clustermap.js)
        writeM8(hmGroups, mrnaIdToGenomeRange, outputDir.resolve("all.collinearity.m8"));
        writeBedpe(hmGroups, mrnaIdToGenomeRange, outputDir.resolve("all.collinearity.bedpe"));
        writeClustermapJs(hmGroups, mrnaIdToGenomeRange, outputDir.resolve("all.collinearity.clustermap.json"));
        writeGenesetExplorerJs(graphRange, outputDir.resolve("all.collinearity.genesetexplorer.json"), writeRepeats);

        // write an Rscript to plot the collinearity file
        //TODO: writeRscript(hmGroups, outputDir.resolve("collinearity.R"));
    }

    /**
     * Method to write a fai file for a graph range
     * @param mrnaIdToGenomeRange a map of mRNA ID to genome range
     * @param faiPath the path to write the fai file (format: sequence, length, offset, lineBases, lineLength)
     * @throws IOException if an error occurs while writing the fai file
     */
    private void writeFai(Map<Long, GenomeRange> mrnaIdToGenomeRange, Path faiPath) throws IOException {
        Pantools.logger.debug("Writing the fai file to {}", faiPath);

        try (BufferedWriter writer = Files.newBufferedWriter(faiPath)) {
            int offset = 0;
            for (GenomeRange genomeRange : new HashSet<>(mrnaIdToGenomeRange.values())) { //TODO: make unique
                final int length = genomeRange.getLength();
                final String sequenceHeader = genomeRange.getCleanedSequenceHeader();
                offset += sequenceHeader.length() + 1; // add the length of the fasta header + 1 for the newline character
                writer.write(String.format("%s\t%d\t%d\t%d\t%d\n", sequenceHeader, length, offset, 80, 81));
                offset += length + (length / 80) + 1; // add the length of the sequence + the number of newlines
            }
        }
    }

    /**
     * Method to write a SIMPLIFIED gff3 file for a graph range (only genes are written, not mRNAs)
     * @param mrnaIdToGenomeRange a map of mRNA ID to genome range
     * @param gff3Path the path to write the gff3 file (format: sequence, source, feature, start, end, score, strand, frame, attributes)
     * @throws IOException if an error occurs while writing the gff3 file
     */
    private void writeGff3(Map<Long, GenomeRange> mrnaIdToGenomeRange, Path gff3Path) throws IOException {
        Pantools.logger.debug("Writing the gff3 file to {}", gff3Path);

        try (BufferedWriter writer = Files.newBufferedWriter(gff3Path)) {
            for (GenomeRange genomeRange : new HashSet<>(mrnaIdToGenomeRange.values())) {
                final Set<Gff3Feature> gff3Features = genomeRange.getGff3Features();
                for (Gff3Feature gff3Feature : gff3Features) {
                    if (gff3Feature.getEnd() < 0) {
                        continue;
                    }
                    writer.write(String.format("%s\t%s\t%s\t%s\t%s\t%s\t%s\t%s\tID=%s;Name=%s\n",
                            genomeRange.getCleanedSequenceHeader(),
                            gff3Feature.getSource(),
                            gff3Feature.getType(),
                            Math.max(gff3Feature.getStart(), 0),
                            gff3Feature.getEnd(),
                            '.',
                            gff3Feature.getStrand(),
                            '.',
                            String.format("%d_%s", genomeRange.getGenomeNr(), gff3Feature.getID()), // add genome number to ID (to make it unique)
                            gff3Feature.getName()));
                }
            }
        }
    }

    /**
     * Method to write a SIMPLIFIED gff3 file for a graph range (only repeats are written, not mRNAs or genes)
     * @param graphRange the graph range to write the repeats from
     * @param gff3Path the path to write the gff3 file (format: sequence, source, feature, start, end, score, strand, frame, attributes)
     * @throws IOException if an error occurs while writing the gff3 file
     */
    private void writeRepeatsGff3(GraphRange graphRange, Path gff3Path) throws IOException {
        Pantools.logger.debug("Writing the repeats GFF3 file to {}", gff3Path);

        try (BufferedWriter writer = Files.newBufferedWriter(gff3Path)) {
            for (GenomeRange genomeRange : graphRange.getGenomeRanges()) {
                final Set<Gff3Feature> repeatGff3Features = genomeRange.getRepeatGff3Features();
                for (Gff3Feature gff3Feature : repeatGff3Features) {
                    writer.write(String.format("%s\t%s\t%s\t%s\t%s\t%s\t%s\t%s\tID=%s;Name=%s\n",
                            genomeRange.getCleanedSequenceHeader(),
                            gff3Feature.getSource(),
                            gff3Feature.getType(),
                            Math.max(gff3Feature.getStart(), 0),
                            gff3Feature.getEnd(),
                            '.',
                            gff3Feature.getStrand(),
                            '.',
                            gff3Feature.getID(),
                            gff3Feature.getName()));
                }
            }
        }
    }

    /**
     * Method to write a SIMPLIFIED gff3 file for a graph range (only BLAST results are written)
     * @param graphRange the graph range to write the BLAST results from
     * @param blastResultsPath the path to the BLAST results file
     * @param gff3Path the path to write the gff3 file (format: sequence, source, feature, start, end, score, strand, frame, attributes)
     * @throws IOException if an error occurs while writing the gff3 file
     */
    private void writeBlastGff3(GraphRange graphRange, Path blastResultsPath, Path gff3Path) throws IOException {
        Pantools.logger.debug("Writing the BLAST results GFF3 file to {}", gff3Path);

        try (BufferedWriter writer = Files.newBufferedWriter(gff3Path)) {
            for (GenomeRange genomeRange : graphRange.getGenomeRanges()) {
                final Set<Gff3Feature> blastGff3Features = genomeRange.getBlastGff3Features(blastResultsPath);
                for (Gff3Feature gff3Feature : blastGff3Features) {
                    writer.write(String.format("%s\t%s\t%s\t%s\t%s\t%s\t%s\t%s\tID=%s;Name=%s\n",
                            genomeRange.getCleanedSequenceHeader(),
                            gff3Feature.getSource(),
                            gff3Feature.getType(),
                            Math.max(gff3Feature.getStart(), 0),
                            gff3Feature.getEnd(),
                            '.',
                            gff3Feature.getStrand(),
                            '.',
                            gff3Feature.getID(),
                            gff3Feature.getName()));
                }
            }
        }
    }

    /**
     * Method to write an m8 file for a graph range
     * @param hmGroups a map of homology group ID to a set of mRNA IDs in the graph range
     * @param mrnaIdToGenomeRange a map of mRNA ID to genome range
     * @param m8Path the path to write the m8 file (format: chrom1, chrom2, percentIdentity, alignmentLength, #gaps, #mismatches, start1, end1, start2, end2, eValue, bitScore)
     * @throws IOException if an error occurs while writing the m8 file
     */
    private void writeM8(Map<Long, Set<Long>> hmGroups, Map<Long, GenomeRange> mrnaIdToGenomeRange, Path m8Path) throws IOException {
        Pantools.logger.debug("Writing the m8 file to {}", m8Path);

        try (BufferedWriter writer = Files.newBufferedWriter(m8Path)) {
            try (Transaction ignored = Globals.GRAPH_DB.beginTx()) {
                for (Map.Entry<Long, Set<Long>> entry : hmGroups.entrySet()) {
                    final List<Long> mrnaIds = new ArrayList<>(entry.getValue());

                    // write pairwise comparisons of mRNAs in the same homology group
                    for (int i = 0; i < mrnaIds.size(); i++) {
                        final long mrnaId1 = mrnaIds.get(i);
                        final Node mrnaNode1 = Globals.GRAPH_DB.getNodeById(mrnaId1);
                        final int[] address1 = (int[]) mrnaNode1.getProperty("address");
                        final String strand1 = (String) mrnaNode1.getProperty("strand");
                        final GenomeRange genomeRange1 = mrnaIdToGenomeRange.get(mrnaId1);
                        int adjustedStart1 = Math.max(address1[2] - genomeRange1.getStart() + 1, 0);
                        int adjustedEnd1 = address1[3] - genomeRange1.getStart() + 1;

                        // flip start and end if the strand is negative since m8 doesn't have a strand column
                        if (strand1.equals("-")) {
                            int temp = adjustedStart1;
                            adjustedStart1 = adjustedEnd1;
                            adjustedEnd1 = temp;
                        }

                        for (int j = i + 1; j < mrnaIds.size(); j++) {
                            final long mrnaId2 = mrnaIds.get(j);
                            final Node mrnaNode2 = Globals.GRAPH_DB.getNodeById(mrnaId2);
                            final int[] address2 = (int[]) mrnaNode2.getProperty("address");
                            final String strand2 = (String) mrnaNode2.getProperty("strand");
                            final GenomeRange genomeRange2 = mrnaIdToGenomeRange.get(mrnaId2);
                            int adjustedStart2 = Math.max(address2[2] - genomeRange2.getStart() + 1, 0);
                            int adjustedEnd2 = address2[3] - genomeRange2.getStart() + 1;

                            // flip start and end if the strand is negative since m8 doesn't have a strand column
                            if (strand2.equals("-")) {
                                int temp = adjustedStart2;
                                adjustedStart2 = adjustedEnd2;
                                adjustedEnd2 = temp;
                            }

                            //get the percent identity by checking the relationship between the two mRNA nodes (if it exists)
                            Relationship rel = null;
                            for (Relationship relationship : mrnaNode1.getRelationships(Globals.RelTypes.is_similar_to)) {
                                if (relationship.getEndNode().equals(mrnaNode2) || relationship.getStartNode().equals(mrnaNode2)) {
                                    rel = relationship;
                                    break;
                                }
                            }

                            final int alignmentLength = 0; //can be filled in later
                            final int numGaps = 0; //can be filled in later
                            final int numMismatches = 0; //can be filled in later
                            final double percentIdentity = rel != null ? (double) rel.getProperty("similarity") : -1.0;
                            final double eValue = 0.0; //can be filled in later
                            final double bitScore = 0.0; //can be filled in later

                            writer.write(String.format("%s\t%s\t%.2f\t%d\t%d\t%d\t%d\t%d\t%d\t%d\t%.2f\t%.2f\n",
                                    genomeRange1.getCleanedSequenceHeader(), genomeRange2.getCleanedSequenceHeader(), percentIdentity, alignmentLength, numGaps, numMismatches, adjustedStart1, adjustedEnd1, adjustedStart2, adjustedEnd2, eValue, bitScore));
                            writer.write(String.format("%s\t%s\t%.2f\t%d\t%d\t%d\t%d\t%d\t%d\t%d\t%.2f\t%.2f\n",
                                    genomeRange2.getCleanedSequenceHeader(), genomeRange1.getCleanedSequenceHeader(), percentIdentity, alignmentLength, numGaps, numMismatches, adjustedStart2, adjustedEnd2, adjustedStart1, adjustedEnd1, eValue, bitScore));
                        }
                    }
                }
            } catch (NotFoundException nfe) {
                Pantools.logger.error("Could not find nodes in the graph database.");
                throw new RuntimeException(nfe);
            }
        }
    }

    /**
     * Method to write a bedpe file for a graph range
     * @param hmGroups a map of homology group ID to a set of mRNA IDs in the graph range
     * @param mrnaIdToGenomeRange a map of mRNA ID to genome range
     * @param bedpePath the path to write the bedpe file (format: chrom1, start1, end1, chrom2, start2, end2, name, score, strand1, strand2)
     * @throws IOException if an error occurs while writing the bedpe file
     */
    private void writeBedpe(Map<Long, Set<Long>> hmGroups, Map<Long, GenomeRange> mrnaIdToGenomeRange, Path bedpePath) throws IOException {
        Pantools.logger.debug("Writing the bedpe file to {}", bedpePath);

        try (BufferedWriter writer = Files.newBufferedWriter(bedpePath)) {
            try (Transaction ignored = Globals.GRAPH_DB.beginTx()) {
                for (Map.Entry<Long, Set<Long>> entry : hmGroups.entrySet()) {
                    final List<Long> mrnaIds = new ArrayList<>(entry.getValue());

                    // write pairwise comparisons of mRNAs in the same homology group
                    for (int i = 0; i < mrnaIds.size(); i++) {
                        final Node mrnaNode1 = Globals.GRAPH_DB.getNodeById(mrnaIds.get(i));
                        final int[] address1 = (int[]) mrnaNode1.getProperty("address");
                        final String strand1 = (String) mrnaNode1.getProperty("strand");
                        final GenomeRange genomeRange1 = mrnaIdToGenomeRange.get(mrnaIds.get(i));
                        final int adjustedStart1 = Math.max(address1[2] - genomeRange1.getStart() + 1, 0);
                        final int adjustedEnd1 = address1[3] - genomeRange1.getStart() + 1;

                        for (int j = i + 1; j < mrnaIds.size(); j++) {
                            final Node mrnaNode2 = Globals.GRAPH_DB.getNodeById(mrnaIds.get(j));
                            final int[] address2 = (int[]) mrnaNode2.getProperty("address");
                            final String strand2 = (String) mrnaNode2.getProperty("strand");
                            final GenomeRange genomeRange2 = mrnaIdToGenomeRange.get(mrnaIds.get(j));
                            final int adjustedStart2 = Math.max(address2[2] - genomeRange2.getStart() + 1, 0);
                            final int adjustedEnd2 = address2[3] - genomeRange2.getStart() + 1;

                            final String name = entry.getKey().toString();
                            final int score = 0; //can be filled in later

                            writer.write(String.format("%s\t%d\t%d\t%s\t%d\t%d\t%s\t%d\t%s\t%s\n",
                                    genomeRange1.getCleanedSequenceHeader(), adjustedStart1, adjustedEnd1, genomeRange2.getCleanedSequenceHeader(), adjustedStart2, adjustedEnd2, name, score, strand1, strand2));
                        }
                    }
                }
            } catch (NotFoundException nfe) {
                Pantools.logger.error("Could not find nodes in the graph database.");
                throw new RuntimeException(nfe);
            }
        }
    }

    /**
     * Method to write a clustermap.js file for a graph range
     * @param hmGroups a map of homology group ID to a set of mRNA IDs in the graph range
     * @param mrnaIdToGenomeRange a map of mRNA ID to genome range
     * @param clustermapJsPath the path to write the clustermap.js file (format: JSON)
     * @throws IOException if an error occurs while writing the clustermap.js file
     */
    private void writeClustermapJs(Map<Long, Set<Long>> hmGroups, Map<Long, GenomeRange> mrnaIdToGenomeRange, Path clustermapJsPath) throws IOException {
        Pantools.logger.debug("Writing the clustermap.js file to {}", clustermapJsPath);

        /*
        format for JSON:
        {
          "clusters": [
            {
              "uid":  <str: Unique ID>,
              "name": <str: Cluster name>,
              "loci": [
                {
                  "uid":   <str: Unique ID>,
                  "name":  <str: Locus name>,
                  "start": <int: Locus start position>,
                  "end":   <int: Locus end position>,
                  "genes": [
                    {
                      "uid":    <str: Unique ID>,
                      "names": [<str: Gene name>],
                      "start":  <int: Gene start position>,
                      "end":    <int: Gene end position>,
                      "strand": <int: Gene strand (0 or 1)>,
                    }
                  ]}
              ]}
          ],
          "links": [
            {
              "query": {
                  "uid":  <str: Unique ID of query gene>,
                  "name": <str: Name of query gene>
              },
              "target": {
                  "uid":  <str: Unique ID of target gene>,
                  "name": <str: Name of target gene>
              },
              "identity": <float: Percent identity query-target alignment>
            }
          ],
          "groups": [
            {
              "uid":     <str: Unique ID of group>,
              "label":   <str: Group label>,
              "genes":  [<str: Gene UID>],
              "colour":  <str: Colour code for gene fill>,
              "hidden":  <bool: Hide group in the plot>
            }
          ]
        }
         */

        JSONObject clustermap = new JSONObject();
        try (Transaction ignored = Globals.GRAPH_DB.beginTx()) {

            // get the clusters (AKA the genome ranges)
            List<JSONObject> clusters = new ArrayList<>();
            for (GenomeRange genomeRange : new HashSet<>(mrnaIdToGenomeRange.values())) {
                JSONObject cluster = new JSONObject();
                cluster.put("uid", UUID.randomUUID());
                cluster.put("name", genomeRange.getSequenceHeader());

                JSONObject locus = new JSONObject();
                locus.put("uid", UUID.randomUUID());
                locus.put("name", genomeRange.getSequenceHeader());
                locus.put("start", 0);
                locus.put("end", genomeRange.getLength());

                List<JSONObject> genes = new ArrayList<>();
                for (Map.Entry<Long, GenomeRange> entry : mrnaIdToGenomeRange.entrySet()) {
                    if (entry.getValue().equals(genomeRange)) {
                        final long mrnaId = entry.getKey();
                        final Node mrnaNode = Globals.GRAPH_DB.getNodeById(mrnaId);
                        final int[] address = (int[]) mrnaNode.getProperty("address");
                        final int strand = mrnaNode.getProperty("strand").toString().equals("+") ? 1 : 0;
                        JSONObject gene = new JSONObject();
                        gene.put("uid", String.format("%s", mrnaId));
                        List<String> names = new ArrayList<>();
                        names.add(mrnaNode.getProperty("id").toString());
                        gene.put("names", names); //can be extended to also include e.g. functional domains
                        gene.put("start", address[2] - genomeRange.getStart() + 1);
                        gene.put("end", address[3] - genomeRange.getStart() + 1);
                        gene.put("strand", strand);
                        genes.add(gene);
                    }
                }
                locus.put("genes", new JSONArray(genes));

                List<JSONObject> loci = new ArrayList<>();
                loci.add(locus);
                cluster.put("loci", new JSONArray(loci));
                clusters.add(cluster);
            }
            clustermap.put("clusters", new JSONArray(clusters));

            // get the links
            List<JSONObject> links = new ArrayList<>();
            for (Map.Entry<Long, Set<Long>> entry : hmGroups.entrySet()) {
                final List<Long> mrnaIds = new ArrayList<>(entry.getValue());
                for (int i = 0; i < mrnaIds.size(); i++) {
                    final Node mrnaNode1 = Globals.GRAPH_DB.getNodeById(mrnaIds.get(i));
                    final int[] address1 = (int[]) mrnaNode1.getProperty("address");
                    for (int j = i + 1; j < mrnaIds.size(); j++) {
                        final Node mrnaNode2 = Globals.GRAPH_DB.getNodeById(mrnaIds.get(j));
                        final int[] address2 = (int[]) mrnaNode2.getProperty("address");
                        final double percentIdentity = 100.0; //can be filled in later
                        JSONObject link = new JSONObject();
                        JSONObject query = new JSONObject();
                        query.put("uid", String.format("%s", mrnaIds.get(i)));
                        query.put("name", mrnaNode1.getProperty("id"));
                        JSONObject target = new JSONObject();
                        target.put("uid", String.format("%s", mrnaIds.get(j)));
                        target.put("name", mrnaNode2.getProperty("id"));
                        link.put("uid", String.format("%d#%s_%d#%s", address1[0], mrnaNode1.getProperty("id"), address2[0], mrnaNode2.getProperty("id")));
                        link.put("query", query);
                        link.put("target", target);
                        link.put("identity", percentIdentity);
                        links.add(link);
                    }
                }
            }
            clustermap.put("links", new JSONArray(links));

            // get the groups
            List<JSONObject> groups = new ArrayList<>();
            final Random random = new Random();
            for (Map.Entry<Long, Set<Long>> entry : hmGroups.entrySet()) {
                JSONObject group = new JSONObject();
                final String hmId = String.format("hg_%d", entry.getKey());
                group.put("uid", hmId);
                group.put("label", hmId);
                List<String> genes = new ArrayList<>();
                for (long mrnaId : entry.getValue()) {
                    genes.add(String.format("%s", mrnaId));
                }
                group.put("genes", new JSONArray(genes));
                group.put("colour", String.format("#%06X", random.nextInt(0x1000000)));
                group.put("hidden", false);
                groups.add(group);
            }
            clustermap.put("groups", new JSONArray(groups));
        } catch (NotFoundException nfe) {
            Pantools.logger.error("Could not find nodes in the graph database.");
            throw new RuntimeException(nfe);
        }

        // write the JSON to the file
        try (BufferedWriter writer = Files.newBufferedWriter(clustermapJsPath)) {
            writer.write(clustermap.toString());
        }
    }

    /**
     * Method to write a GeneSet Explorer JSON file for a graph range
     * @param graphRange the graph range to write
     * @param genesetExplorerJsPath the path to write the GeneSet Explorer JSON file (format: JSON)
     * @param writeRepeats whether to write repeats to the GeneSet Explorer JSON file
     * @throws IOException if an error occurs while writing the GeneSet Explorer JSON file
     */
    private void writeGenesetExplorerJs(final GraphRange graphRange, final Path genesetExplorerJsPath, final boolean writeRepeats) throws IOException {
        Pantools.logger.debug("Writing the GeneSet Explorer JSON file to {}", genesetExplorerJsPath);

        /*
        format for JSON:
        {
            "genomes": [
                {
                    "uid": <str: unique ID>,
                    "name": <str: genome name or number>,
                    "sequences": [
                        {
                            "uid": <str: unique ID>,
                            "name": <str: sequence name>,
                            "id": <str: sequence ID in format `{genome_nr}_{sequence_nr}`>,
                            "phasing": <str: unique ID> | undefined,
                            "sequence_length_nuc": <int: sequence length in nucleotides>,
                            "gene_count": <int: number of genes>,
                            "loci": [
                                <str: loci ID>
                            ]
                        }
                    ]
                }
            ],
            "loci": [
                        {
                            "uid": <str: unique ID>,
                            "name": <str: locus name>,
                            "start": <int: locus start position>,
                            "end": <int: locus end position>,
                            "loci_length_nuc": <int: sequence length in nucleotides>,
                            "genes": [
                                <str: gene ID>
                            ],
                            "repeats": [
                                <str: unique ID>
                            ]
                        }
            ],
            "genes": [
                {
                    "uid": <str: unique ID>,
                    "label": <str: PanTools gene id>,
                    "names": [
                        <str: gene name>
                    ],
                    "start": <int: gene start position>,
                    "end": <int: gene end position>,
                    "strand": <int: gene strand (0 or 1)> | undefined,
                    "sequence_id": <str: unique locus id>,
                    "gene_length_nuc": <int: gene length in nucleotides>,
                    "mrnas": [
                        <str: unique ID>
                    ],
                    "noncoding_rnas": {
                        trnas: [
                            <str: unique ID>
                        ],
                        rrnas: [
                            <str: unique ID>
                        ]
                    }
                }
            ],
            "mrnas": [
                {
                    "uid": <str: unique ID>,
                    "label": <str: PanTools mrna id>,
                    "start": <int: mrna start position>,
                    "end": <int: mrna end position>,
                    "mrna_length_nuc": <int: mrna length in nucleotides>,
                    "cds_length_nuc": <int: cds length in nucleotides>,
                    "protein_length_aa": <int: protein length in amino acids>,
                    "exons": [
                        <str: Unique ID>
                    ],
                    "cdss": [
                        <str: Unique ID>
                    ],
                    "functional_domains": [
                        {
                            "uid": <str: unique ID>,
                            "start": <int: domain start position on mrna>,
                            "end": <int: domain end position on mrna>
                        }
                    ]
                }
            ],
            "functional_domains": [
                        {
                            "uid": <str: unique ID>,
                            "domain_id": <str: domain ID database>,
                            "domain_type": <str: database type>,
                            "name": <str: domain name>,
                        }
            ],
            "exons": [
                {
                    "uid": <str: unique ID>,
                    "label": <str: PanTools exon label>,
                    "start": <int: exon start position>,
                    "end": <int: exon end position>
                }
            ],
            "cds": [
                {
                    "uid": <str: unique ID>,
                    "label": <str: PanTools cds label>,
                    "start": <int: cds start position>,
                    "end": <int: cds end position>
                }
            ],
            "repeats": [
                {
                    "uid": <str: unique ID>,
                    "label": <str: PanTools repeat label>,
                    "start": <int: start position>,
                    "end": <int: end position>
                }
            ],
            "noncoding_rnas": {
                "trnas": [
                    {
                        "uid": <str: unique ID>,
                        "label": <str: PanTools trna id>,
                        "start": <int: start position>,
                        "end": <int: end position>,
                        "exons": [
                            <str: unique ID>
                        ]
                    }
                ],
                "rrnas": [
                    {
                        "uid": <str: unique ID>,
                        "label": <str: PanTools rrna id>,
                        "start": <int: start position>,
                        "end": <int: end position>,
                        "exons": [
                            <str: unique ID>
                        ]
                    }
                ]
            },
            "links": [
                {
                    "uid": <str: unique ID of query-target link>,
                    "identity": <float: percent identity query-target alignment>,
                    "query": {
                        "uid": <str: unique ID of query mrna>,
                    },
                    "target": {
                        "uid": <str: unique ID of target mrna>,
                    }
                }
            ],
            "groups": [
                {
                    "uid": <str: unique ID of homology group>,
                    "label": <str: group label>,
                    "mrnas": [
                        <str: ID of mrnas>
                    ],
                    "hidden": <bool: hide group in the plot>
                }
            ],
            "version": <string: version of file>,
            "origin": <string: free field information on PanTools version and commands>
        }
         */

        // start building the JSON object
        JSONObject genesetExplorer = new JSONObject();
        genesetExplorer.put("version", "v0.2.2");

        // add PanTools version
        final String[] versionInfo = new Pantools.GitVersionProvider().getVersion();
        genesetExplorer.put("origin", String.format("Generated by PanTools %s",
                Arrays.stream(versionInfo)
                        .map(Objects::toString)
                        .collect(Collectors.joining(", "))));

        try (Transaction ignored = Globals.GRAPH_DB.beginTx();
             ResourceIterator<Node> geneNodesForCaching = Globals.GRAPH_DB.findNodes(Globals.GENE_LABEL);
             ResourceIterator<Node> sequenceNodesForCaching = Globals.GRAPH_DB.findNodes(Globals.SEQUENCE_LABEL)) {

            // create an array for gene counts per genome and sequence
            final int[][] geneCounts = new int[GENOME_DB.num_genomes + 1][];
            while (geneNodesForCaching.hasNext()) {
                final Node geneNode = geneNodesForCaching.next();
                final int[] address = (int[]) geneNode.getProperty("address");
                geneCounts[address[0]] = geneCounts[address[0]] == null ? new int[GENOME_DB.num_sequences[address[0]] + 1] : geneCounts[address[0]];
                geneCounts[address[0]][address[1]] = geneCounts[address[0]][address[1]] + 1;
            }

            // create a map for <genome nr>_<sequence nr> -> sequence node
            final Map<String, Node> genomeSequenceToSequenceNode = new HashMap<>();
            while (sequenceNodesForCaching.hasNext()) {
                final Node sequenceNode = sequenceNodesForCaching.next();
                final String genomeSequence = (String) sequenceNode.getProperty("identifier");
                genomeSequenceToSequenceNode.put(genomeSequence, sequenceNode);
            }

            // create a map for <genome nr>_<sequence nr> -> genome range
            final Map<String, List<GenomeRange>> genomeSequenceToGenomeRange = new HashMap<>();
            for (GenomeRange genomeRange : graphRange.getGenomeRanges()) {
                final String genomeSequence = String.format("%d_%d", genomeRange.getGenomeNr(), genomeRange.getSequenceNr());
                genomeSequenceToGenomeRange.computeIfAbsent(genomeSequence, k -> new ArrayList<>()).add(genomeRange);
            }

            // get the genomes and sequences of the entire pangenome and add loci when available
            List<JSONObject> genomes = new ArrayList<>();
            Map<UUID, GenomeRange> UidToGenomeRange = new HashMap<>();
            for (int genomeNr = 1; genomeNr <= GENOME_DB.num_genomes; ++genomeNr) {
                JSONObject genome = new JSONObject();
                genome.put("uid", UUID.randomUUID());
                genome.put("name", String.format("%d", genomeNr));

                List<JSONObject> sequences = new ArrayList<>();
                for (int sequenceNr = 1; sequenceNr <= GENOME_DB.num_sequences[genomeNr]; ++sequenceNr) {
                    final String genomeSequence = String.format("%d_%d", genomeNr, sequenceNr);
                    final Node sequenceNode = genomeSequenceToSequenceNode.get(genomeSequence);
                    final Object phasing = sequenceNode.hasProperty("phasing_ID") ? (String) sequenceNode.getProperty("phasing_ID") : JSONObject.NULL;

                    JSONObject sequence = new JSONObject();
                    sequence.put("uid", UUID.randomUUID());
                    sequence.put("name", GENOME_DB.sequence_titles[genomeNr][sequenceNr]);
                    sequence.put("id", genomeSequence);
                    sequence.put("phasing", phasing);
                    sequence.put("sequence_length_nuc", GENOME_DB.sequence_length[genomeNr][sequenceNr]);
                    try {
                        sequence.put("gene_count", geneCounts[genomeNr][sequenceNr]);
                    } catch (NullPointerException npe) {
                        sequence.put("gene_count", 0);
                    }

                    List<String> loci = new ArrayList<>();
                    if (genomeSequenceToGenomeRange.containsKey(genomeSequence)) {
                        for (GenomeRange genomeRange : genomeSequenceToGenomeRange.get(genomeSequence)) {
                            UUID locusUid = UUID.randomUUID();
                            loci.add(locusUid.toString());
                            UidToGenomeRange.put(locusUid, genomeRange);
                        }
                    }
                    sequence.put("loci", loci);
                    sequences.add(sequence);
                }
                genome.put("sequences", new JSONArray(sequences));
                genomes.add(genome);
            }
            genesetExplorer.put("genomes", new JSONArray(genomes));

            // get the loci
            List<JSONObject> loci = new ArrayList<>();
            Map<UUID, Node> geneNodes = new HashMap<>();
            Map<UUID, Node> repeatNodes = new HashMap<>();
            for (Map.Entry<UUID, GenomeRange> entry : UidToGenomeRange.entrySet()) {
                final GenomeRange genomeRange = entry.getValue();
                JSONObject locus = new JSONObject();
                locus.put("uid", entry.getKey().toString());
                locus.put("name", genomeRange.getSequenceHeader());
                locus.put("start", genomeRange.getStart());
                locus.put("end", genomeRange.getEnd());
                locus.put("loci_length_nuc", genomeRange.getLength());

                List<String> genes = new ArrayList<>();
                for (Node geneNode : genomeRange.getTreeGeneNodes()) {
                    if (geneNode.hasLabel(Globals.GENE_LABEL)) {
                        UUID geneUid = UUID.randomUUID();
                        genes.add(geneUid.toString());
                        geneNodes.put(geneUid, geneNode);
                    }
                }
                locus.put("genes", new JSONArray(genes));

                if (writeRepeats) {
                    List<String> repeats = new ArrayList<>();
                    for (Node repeatNode : genomeRange.getRepeatNodes()) {
                        UUID repeatUid = UUID.randomUUID();
                        repeats.add(repeatUid.toString());
                        repeatNodes.put(repeatUid, repeatNode);
                    }
                    locus.put("repeats", new JSONArray(repeats));
                }

                loci.add(locus);
            }
            genesetExplorer.put("loci", new JSONArray(loci));

            // get the genes
            List<JSONObject> genes = new ArrayList<>();
            Map<UUID, Node> mrnaNodes = new HashMap<>();
            Map<Node, UUID> mrnaNodeToUid = new HashMap<>();
            Map<UUID, Node> noncodingRnaNodes = new HashMap<>();
            for (Map.Entry<UUID, Node> entry : geneNodes.entrySet()) {
                final Node geneNode = entry.getValue();
                final int[] address = (int[]) geneNode.getProperty("address");
                final int strand = geneNode.getProperty("strand").toString().equals("+") ? 1 : 0;
                JSONObject gene = new JSONObject();
                gene.put("uid", entry.getKey().toString());
                gene.put("label", geneNode.getProperty("id"));
                gene.put("names", geneNode.getProperty("name"));
                gene.put("start", address[2]);
                gene.put("end", address[3]);
                gene.put("strand", strand);
                gene.put("sequence_id", Utils.getSequenceName(address[0], address[1]));
                gene.put("gene_length_nuc", address[3] - address[2] + 1);

                List<String> mrnas = new ArrayList<>();
                List<String> trnas = new ArrayList<>();
                List<String> rnas = new ArrayList<>();
                for (Relationship relationship : geneNode.getRelationships(Direction.OUTGOING, Globals.RelTypes.is_parent_of)) {
                    final Node transcriptNode = relationship.getEndNode();
                    if (transcriptNode.hasLabel(Globals.MRNA_LABEL)) {
                        UUID mrnaUid = UUID.randomUUID();
                        mrnas.add(mrnaUid.toString());
                        mrnaNodes.put(mrnaUid, transcriptNode);
                        mrnaNodeToUid.put(transcriptNode, mrnaUid);
                    } else if (transcriptNode.hasLabel(Globals.TRNA_LABEL)) {
                        UUID noncodingTrnaUid = UUID.randomUUID();
                        gene.put("noncoding_rnas", new JSONObject());
                        trnas.add(noncodingTrnaUid.toString());
                        noncodingRnaNodes.put(noncodingTrnaUid, transcriptNode);
                    } else if (transcriptNode.hasLabel(Globals.RRNA_LABEL)) {
                        UUID noncodingRrnaUid = UUID.randomUUID();
                        gene.put("noncoding_rnas", new JSONObject());
                        rnas.add(noncodingRrnaUid.toString());
                        noncodingRnaNodes.put(noncodingRrnaUid, transcriptNode);
                    }
                }
                gene.put("mrnas", new JSONArray(mrnas));
                gene.put("noncoding_rnas", new JSONObject().put("trnas", new JSONArray(trnas)).put("rrnas", new JSONArray(rnas)));
                genes.add(gene);
            }
            genesetExplorer.put("genes", new JSONArray(genes));

            // get the mRNAs
            List<JSONObject> mrnas = new ArrayList<>();
            Map<Node, UUID> exonNodes = new HashMap<>();
            Map<Node, UUID> cdsNodes = new HashMap<>();
            Map<Node, UUID> functionNodes = new HashMap<>();
            for (Map.Entry<UUID, Node> entry : mrnaNodes.entrySet()) {
                final Node mrnaNode = entry.getValue();
                final int[] address = (int[]) mrnaNode.getProperty("address");
                JSONObject mrna = new JSONObject();
                mrna.put("uid", entry.getKey().toString());
                mrna.put("label", mrnaNode.getProperty("id"));
                mrna.put("start", address[2]);
                mrna.put("end", address[3]);
                mrna.put("mrna_length_nuc", address[3] - address[2] + 1);

                List<String> exons = new ArrayList<>();
                List<String> cdss = new ArrayList<>();
                int cdsLengthNuc = 0;
                for (Relationship relationship : mrnaNode.getRelationships(Direction.OUTGOING, Globals.RelTypes.is_parent_of)) {
                    final Node exonCdsNode = relationship.getEndNode();
                    if (exonCdsNode.hasLabel(Globals.EXON_LABEL)) {
                        if (exonNodes.containsKey(exonCdsNode)) {
                            exons.add(exonNodes.get(exonCdsNode).toString());
                        } else {
                            UUID exonUid = UUID.randomUUID();
                            exons.add(exonUid.toString());
                            exonNodes.put(exonCdsNode, exonUid);
                        }
                    } else if (exonCdsNode.hasLabel(Globals.CDS_LABEL)) {
                        if (cdsNodes.containsKey(exonCdsNode)) {
                            cdss.add(cdsNodes.get(exonCdsNode).toString());
                        } else {
                            UUID cdsUid = UUID.randomUUID();
                            cdss.add(cdsUid.toString());
                            cdsNodes.put(exonCdsNode, cdsUid);
                        }
                        cdsLengthNuc += ((int) exonCdsNode.getProperty("length"));
                    }
                }
                mrna.put("exons", new JSONArray(exons));
                mrna.put("cdss", new JSONArray(cdss));
                mrna.put("cds_length_nuc", cdsLengthNuc);
                mrna.put("protein_length_aa", cdsLengthNuc / 3);

                List<JSONObject> domains = new ArrayList<>();
                for (Relationship relationship : mrnaNode.getRelationships(Direction.OUTGOING)) {
                    JSONObject domain = new JSONObject();
                    if (relationship.isType(Globals.RelTypes.has_go) || relationship.isType(Globals.RelTypes.has_interpro) || relationship.isType(Globals.RelTypes.has_pfam) || relationship.isType(Globals.RelTypes.has_tigrfam)) {
                        final Node functionNode = relationship.getEndNode();
                        if (functionNodes.containsKey(functionNode)) {
                            domain.put("uid", functionNodes.get(functionNode).toString());
                        } else {
                            UUID functionUid = UUID.randomUUID();
                            domain.put("uid", functionUid);
                            functionNodes.put(functionNode, functionUid);
                        }
                        domain.put("start", relationship.hasProperty("aa_start") ? (int) relationship.getProperty("aa_start") : -1);
                        domain.put("end", relationship.hasProperty("aa_end") ? (int) relationship.getProperty("aa_end") : -1);
                    } else {
                        continue;
                    }
                    domains.add(domain);
                }
                mrna.put("functional_domains", new JSONArray(domains));

                mrnas.add(mrna);
            }
            genesetExplorer.put("mrnas", new JSONArray(mrnas));

            // get the domains
            List<JSONObject> domains = new ArrayList<>();
            for (Map.Entry<Node, UUID> entry : functionNodes.entrySet()) {
                final Node functionNode = entry.getKey();
                JSONObject domain = new JSONObject();
                domain.put("uid", entry.getValue().toString());
                domain.put("domain_id", functionNode.getProperty("id"));
                if (functionNode.hasLabel(Globals.GO_LABEL)) {
                    domain.put("domain_type", "GO");
                } else if (functionNode.hasLabel(Globals.INTERPRO_LABEL)) {
                    domain.put("domain_type", "InterPro");
                } else if (functionNode.hasLabel(Globals.PFAM_LABEL)) {
                    domain.put("domain_type", "Pfam");
                } else if (functionNode.hasLabel(Globals.TIGRFAM_LABEL)) {
                    domain.put("domain_type", "TIGRFAM");
                }
                domain.put("name", functionNode.getProperty("name"));
                domains.add(domain);
            }
            genesetExplorer.put("functional_domains", new JSONArray(domains));

            // get the tRNAs
            List<JSONObject> trnas = new ArrayList<>();
            for (Map.Entry<UUID, Node> entry : noncodingRnaNodes.entrySet()) {
                final Node trnaNode = entry.getValue();
                final int[] address = (int[]) trnaNode.getProperty("address");
                JSONObject trna = new JSONObject();
                trna.put("uid", entry.getKey().toString());
                trna.put("label", trnaNode.getProperty("id"));
                trna.put("start", address[2]);
                trna.put("end", address[3]);

                List<String> exons = new ArrayList<>();
                for (Relationship relationship : trnaNode.getRelationships(Direction.OUTGOING, Globals.RelTypes.is_parent_of)) {
                    final Node exonNode = relationship.getEndNode();
                    if (exonNode.hasLabel(Globals.EXON_LABEL)) {
                        if (exonNodes.containsKey(exonNode)) {
                            exons.add(exonNodes.get(exonNode).toString());
                        } else {
                            UUID exonUid = UUID.randomUUID();
                            exons.add(exonUid.toString());
                            exonNodes.put(exonNode, exonUid);
                        }
                    }
                }
                trna.put("exons", new JSONArray(exons));

                trnas.add(trna);
            }

            // get the rRNAs
            List<JSONObject> rrnas = new ArrayList<>();
            for (Map.Entry<UUID, Node> entry : noncodingRnaNodes.entrySet()) {
                final Node rrnaNode = entry.getValue();
                final int[] address = (int[]) rrnaNode.getProperty("address");
                JSONObject rrna = new JSONObject();
                rrna.put("uid", entry.getKey().toString());
                rrna.put("label", rrnaNode.getProperty("id"));
                rrna.put("start", address[2]);
                rrna.put("end", address[3]);

                List<String> exons = new ArrayList<>();
                for (Relationship relationship : rrnaNode.getRelationships(Direction.OUTGOING, Globals.RelTypes.is_parent_of)) {
                    final Node exonNode = relationship.getEndNode();
                    if (exonNode.hasLabel(Globals.EXON_LABEL)) {
                        if (exonNodes.containsKey(exonNode)) {
                            exons.add(exonNodes.get(exonNode).toString());
                        } else {
                            UUID exonUid = UUID.randomUUID();
                            exons.add(exonUid.toString());
                            exonNodes.put(exonNode, exonUid);
                        }
                    }
                }
                rrna.put("exons", new JSONArray(exons));

                rrnas.add(rrna);
            }

            // add the noncoding RNAs to the geneset explorer object
            JSONObject noncodingRnas = new JSONObject();
            noncodingRnas.put("trnas", new JSONArray(trnas));
            noncodingRnas.put("rrnas", new JSONArray(rrnas));
            genesetExplorer.put("noncoding_rnas", noncodingRnas);

            // get the exons
            List<JSONObject> exons = new ArrayList<>();
            for (Map.Entry<Node, UUID> entry : exonNodes.entrySet()) {
                final Node exonNode = entry.getKey();
                final int[] address = (int[]) exonNode.getProperty("address");
                JSONObject exon = new JSONObject();
                exon.put("uid", entry.getValue().toString());
                exon.put("start", address[2]);
                exon.put("end", address[3]);
                exons.add(exon);
            }
            genesetExplorer.put("exons", new JSONArray(exons));

            // get the cds
            List<JSONObject> cdss = new ArrayList<>();
            for (Map.Entry<Node, UUID> entry : cdsNodes.entrySet()) {
                final Node cdsNode = entry.getKey();
                final int[] address = (int[]) cdsNode.getProperty("address");
                JSONObject cds = new JSONObject();
                cds.put("uid", entry.getValue().toString());
                cds.put("start", address[2]);
                cds.put("end", address[3]);
                cdss.add(cds);
            }
            genesetExplorer.put("cds", new JSONArray(cdss));

            // get the repeats
            if (writeRepeats) {
                List<JSONObject> repeats = new ArrayList<>();
                for (Map.Entry<UUID, Node> entry : repeatNodes.entrySet()) {
                    final Node repeatNode = entry.getValue();
                    final int[] address = (int[]) repeatNode.getProperty("address");
                    JSONObject repeat = new JSONObject();
                    repeat.put("uid", entry.getKey().toString());
                    repeat.put("label", repeatNode.getProperty("type"));
                    repeat.put("start", address[2]);
                    repeat.put("end", address[3]);
                    repeats.add(repeat);
                }
                genesetExplorer.put("repeats", new JSONArray(repeats));
            } else {
                genesetExplorer.put("repeats", new JSONArray());
            }

            // get all homology nodes belonging to the mRNA nodes
            Map<Node, UUID> homologyNodes = new HashMap<>();
            for (Map.Entry<UUID, Node> entry : mrnaNodes.entrySet()) {
                final Node mrnaNode = entry.getValue();
                for (Relationship relationship : mrnaNode.getRelationships(Direction.INCOMING, Globals.RelTypes.has_homolog)) {
                    final Node homologNode = relationship.getStartNode();
                    if (homologyNodes.containsKey(homologNode)) {
                        continue;
                    }
                    homologyNodes.put(homologNode, UUID.randomUUID());
                }
            }

            // get the groups
            List<JSONObject> groups = new ArrayList<>();
            Set<List<Node>> homologyGroups = new HashSet<>();
            for (Map.Entry<Node, UUID> entry : homologyNodes.entrySet()) { //ignore warning about empty map, it is not empty
                final Node homologyNode = entry.getKey();
                JSONObject group = new JSONObject();
                group.put("uid", entry.getValue().toString());
                group.put("label", homologyNode.getId()); //TODO: make proper label

                List<String> hmMrnas = new ArrayList<>();
                List<Node> homologyGroup = new ArrayList<>();
                for (Relationship relationship : homologyNode.getRelationships(Direction.OUTGOING, Globals.RelTypes.has_homolog)) {
                    final Node mrnaNode = relationship.getEndNode();
                    if (mrnaNodeToUid.containsKey(mrnaNode)) {
                        hmMrnas.add(mrnaNodeToUid.get(mrnaNode).toString());
                        homologyGroup.add(mrnaNode);
                    }
                }
                group.put("mrnas", new JSONArray(hmMrnas));
                homologyGroups.add(homologyGroup);

                group.put("hidden", false);
                groups.add(group);
            }
            genesetExplorer.put("groups", new JSONArray(groups));

            // get the links
            List<JSONObject> links = new ArrayList<>();
            for (List<Node> homologyGroup : homologyGroups) {
                for (int i = 0; i < homologyGroup.size(); i++) {
                    final Node mrnaNode1 = homologyGroup.get(i);
                    for (int j = i + 1; j < homologyGroup.size(); j++) {
                        final Node mrnaNode2 = homologyGroup.get(j);

                        //get the percent identity by checking the relationship between the two mRNA nodes (if it exists)
                        Relationship rel = null;
                        for (Relationship relationship : mrnaNode1.getRelationships(Globals.RelTypes.is_similar_to)) {
                            if (relationship.getEndNode().equals(mrnaNode2) || relationship.getStartNode().equals(mrnaNode2)) {
                                rel = relationship;
                                break;
                            }
                        }

                        JSONObject link = new JSONObject();
                        link.put("uid", UUID.randomUUID());
                        link.put("identity", rel != null ? rel.getProperty("similarity") : JSONObject.NULL);
                        JSONObject query = new JSONObject();
                        query.put("uid", mrnaNodeToUid.get(mrnaNode1).toString());
                        link.put("query", query);
                        JSONObject target = new JSONObject();
                        target.put("uid", mrnaNodeToUid.get(mrnaNode2).toString());
                        link.put("target", target);
                        links.add(link);
                    }
                }
            }
            genesetExplorer.put("links", new JSONArray(links));
        } catch (NotFoundException nfe) {
            Pantools.logger.error("Could not find nodes in the graph database.");
            throw new RuntimeException(nfe);
        }

        // write the JSON to the file
        try (BufferedWriter writer = Files.newBufferedWriter(genesetExplorerJsPath)) {
            writer.write(genesetExplorer.toString());
        }
    }

    /**
     * Method to obtain the total gene count for a genome and sequence
     * @param genomeNr the genome number
     * @param sequenceNr the sequence number
     * @return the total gene count for the genome and sequence
     */
    private int getTotalGeneCount(int genomeNr, int sequenceNr) {
        int geneCount = 0;
        try (Transaction ignored = Globals.GRAPH_DB.beginTx();
             ResourceIterator<Node> featureNodes = Globals.GRAPH_DB.findNodes(Globals.GENE_LABEL, "genome", genomeNr, "sequence", sequenceNr)) {
            while (featureNodes.hasNext()) {
                featureNodes.next();
                geneCount++;
            }
        } catch (NotFoundException nfe) {
            Pantools.logger.warn("Genome {} (sequence {}) don't have any genes.", genomeNr, sequenceNr);
        }

        return geneCount;
    }

    /**
     * Method to obtain a map of mRNA ID to genome range for a graph range
     * @param graphRange the graph range to write
     * @return a map of mRNA ID to genome range
     */
    private Map<Long, GenomeRange> getMrnaIdToGenomeRange(GraphRange graphRange) {
        Pantools.logger.debug("Obtaining a map of mRNA ID to genome range for a graph range.");

        Map<Long, GenomeRange> mrnaIdToGenomeRange = new HashMap<>(); // homology group -> set of mRNA IDs in the graph range
        graphRange.getGenomeRanges().forEach(genomeRange -> {
            try (Transaction ignored = Globals.GRAPH_DB.beginTx()) {
                for (Node mrnaNode : genomeRange.getTreeGeneNodes()) {
                    if (mrnaNode.hasLabel(Globals.MRNA_LABEL)) {
                        final int[] address = (int[]) mrnaNode.getProperty("address");
                        if (genomeRange.overlaps(new GenomeRange(address, false))) {
                            mrnaIdToGenomeRange.put(mrnaNode.getId(), genomeRange);
                        }
                    }
                }
            }
        });

        return mrnaIdToGenomeRange;
    }

    /**
     * Method to obtain all homology groups that have mRNAs in the given graphRange including location
     * @param mrnaIdToGenomeRange a map of mRNA ID to genome range
     * @return a map of homology group ID to a set of mRNA IDs in the graph range
     */
    private Map<Long, Set<Long>> getHmGroups(Map<Long, GenomeRange> mrnaIdToGenomeRange) {
        Pantools.logger.debug("Obtaining all homology groups that have mRNAs in the given graphRange including location.");

        Map<Long, Set<Long>> hmGroups = new HashMap<>(); // homology group -> set of mRNA IDs in the graph range
        try (Transaction ignored = Globals.GRAPH_DB.beginTx();
             ResourceIterator<Node> hmNodes = Globals.GRAPH_DB.findNodes(Globals.HOMOLOGY_GROUP_LABEL)) {
            while (hmNodes.hasNext()) {
                final Node hmNode = hmNodes.next();
                final long hmId = hmNode.getId();
                for (Relationship rel : hmNode.getRelationships(Globals.RelTypes.has_homolog, Direction.OUTGOING)) {
                    final Node mrnaNode = rel.getEndNode();
                    if (!mrnaNode.hasLabel(Globals.MRNA_LABEL)) {
                        continue;
                    }
                    if (mrnaIdToGenomeRange.containsKey(mrnaNode.getId())) {
                        hmGroups.computeIfAbsent(hmId, k -> new HashSet<>()).add(mrnaNode.getId());
                    }
                }
            }
        } catch (NotFoundException nfe) {
            Pantools.logger.error("Could not find nodes in the graph database.");
            throw new RuntimeException(nfe);
        }

        return hmGroups;
    }
}
