package nl.wur.bif.pantools.utils.writer;

import nl.wur.bif.pantools.Pantools;
import nl.wur.bif.pantools.utils.graph.graph_objects.GraphEdge;
import nl.wur.bif.pantools.utils.graph.graph_objects.Graph;
import nl.wur.bif.pantools.utils.graph.graph_objects.GraphPath;
import nl.wur.bif.pantools.utils.graph.graph_objects.GraphNode;

import java.io.BufferedWriter;
import java.io.IOException;
import java.nio.file.Files;
import java.nio.file.Path;
import java.util.Iterator;

public abstract class GfaWriter {
    public GfaWriter() {
    }

    /**
     * Abstract method to obtain the GFA version
     * @return GFA version
     */
    protected abstract String getGfaVersion();

    /**
     * Abstract method to write a single graphNode to the GFA file
     * @param graphNode graphNode object to write
     * @return String representation of the graphNode
     */
    protected abstract String getGraphNodeOutputString(GraphNode graphNode);

    /**
     * Abstract method to write a single edge to the GFA file
     * @param graphEdge Edge object to write
     * @return String representation of the edge
     */
    protected abstract String getEdgeOutputString(GraphEdge graphEdge);

    /**
     * Abstract method to write a single path to the GFA file
     * @param graphPath Path object to write
     * @return String representation of the path
     */
    protected abstract String getGraphPathOutputString(GraphPath graphPath);

    /**
     * Abstract method to obtain the path separator character
     * @return path separator character
     */
    protected abstract char getGraphPathSeparator();

    /**
     * Writes the graph to a GFA file
     * @param graph graph object to write
     * @param filename name of the file to write the graph to
     * @throws IOException if the file cannot be written
     */
    public void write(Graph graph, Path filename) throws IOException {
        Pantools.logger.info("Writing graph as GFA format ({}) to {}",
                getGfaVersion(), filename);

        // create the directory of the file if it does not exist
        if (!Files.exists(filename.toAbsolutePath().getParent())) {
            Files.createDirectories(filename.getParent());
        }

        // start up the writer
        try (BufferedWriter writer = Files.newBufferedWriter(filename)) {
            // write the header
            writeHeader(writer);

            // write the sequences
            writeSequences(writer, graph);

            // write the links
            writeLinks(writer, graph);

            // write the paths
            writeGraphPaths(writer, graph);
        }
    }

    /**
     * Writes the header to the GFA file
     * @param writer writer to write the header to
     * @throws IOException if the header cannot be written
     */
    private void writeHeader(BufferedWriter writer) throws IOException {
        writer.write(String.format("H\tVN:Z:%s\n", getGfaVersion()));
    }

    /**
     * Writes the sequences to the GFA file
     * @param writer writer to write the sequences to
     * @param graph graph object to write the sequences from
     */
    private void writeSequences(BufferedWriter writer, Graph graph) {
        graph.getGraphNodes().forEach((id, graphNode) -> {
            Pantools.logger.trace("Writing graphNode {}", id);
            try {
                if (graphNode.isNucleotide()) {
                    writer.write(getGraphNodeOutputString(graphNode));
                }
            } catch (IOException ioe) {
                throw new RuntimeException(String.format("Could not write graphNode %d", id), ioe);
            }
        });
    }

    /**
     * Writes the links to the GFA file
     * @param writer writer to write the links to
     * @param graph graph object to write the links from
     */
    private void writeLinks(BufferedWriter writer, Graph graph) {
        graph.getEdges().forEach((id, edge) -> {
            Pantools.logger.trace("Writing edge {}", id);
            if (edge.getSource() != null && edge.getTarget() != null) { // happens with subgraphs
                if (edge.getSource().isNucleotide() && edge.getTarget().isNucleotide()) { // happens with sequence nodes
                    try {
                        writer.write(getEdgeOutputString(edge));
                    } catch (IOException ioe) {
                        throw new RuntimeException(String.format("Could not write edge %d", id), ioe);
                    }
                }
            }
        });
    }

    /**
     * Writes the paths to the GFA file
     * @param writer writer to write the paths to
     * @param graph graph object to write the paths from
     */
    private void writeGraphPaths(BufferedWriter writer, Graph graph) {
        graph.getGraphPaths().forEach(path -> {
            Pantools.logger.trace("Writing path {}", path.getOrigin());
            try {
                writer.write(getGraphPathOutputString(path));
            } catch (IOException ioe) {
                throw new RuntimeException(String.format("Could not write path %s", path.getOrigin()), ioe);
            }
        });
    }

    /**
     * Convert a path to a GFA formatted path
     * @param graphPath path to convert
     * @return GFA path string
     */
    protected String convertGraphPathToGfa(GraphPath graphPath) {
        StringBuilder pathBuilder = new StringBuilder();
        Iterator<GraphEdge> walker = graphPath.walk();

        while (walker.hasNext()) {
            GraphEdge graphEdge = walker.next(); //TODO: somehow this can give a null pointer exception
            if (graphEdge.getSource().isNucleotide()) {
                pathBuilder
                        .append(graphEdge.getSource().getId())
                        .append(graphEdge.getSourceOrientation());
                if (walker.hasNext()) {
                    pathBuilder.append(getGraphPathSeparator());
                }
            }
        }

        return pathBuilder.toString();
    }
}
