package nl.wur.bif.pantools.utils.writer;

import nl.wur.bif.pantools.utils.Globals;
import nl.wur.bif.pantools.utils.graph.graph_objects.GraphRange;

import java.io.BufferedWriter;
import java.io.IOException;
import java.nio.file.Files;
import java.nio.file.Path;
import java.util.Map;

public abstract class PresenceAbsenceWriter {
    public PresenceAbsenceWriter() {}

    /**
     * Abstract method to write a presence/absence file to a path
     * @param graphRange the graph range to write
     * @param outputFilePath path to write the presence/absence file to
     * @throws IOException if an error occurs while writing the presence/absence file
     */
    public abstract void write(GraphRange graphRange, Path outputFilePath) throws IOException;

    /**
     * Writes a presence/absence file to a TSV file
     * @param presenceAbsence a map of unique IDs to a map of genome numbers to the presence/absence of the unique ID in the genome
     * @param metadata a map of unique IDs to a string of metadata to be added as last column
     * @param outputFilePath path to write the presence/absence file to
     * @throws IOException if an error occurs while writing the presence/absence file
     */
    protected void writeTsv(Map<Long, Map<Integer, Long>> presenceAbsence, Map<Long, String> metadata, Path outputFilePath) throws IOException {
        // create the directory of the file if it does not exist
        if (!Files.exists(outputFilePath.getParent())) {
            Files.createDirectories(outputFilePath.getParent());
        }

        try (BufferedWriter writer = Files.newBufferedWriter(outputFilePath)) {
            // write header
            writer.write("node");
            for (int genomeNr = 1; genomeNr <= Globals.total_genomes; genomeNr++) {
                if (Globals.selectedGenomes.contains(genomeNr)) {
                    writer.write(String.format("\tgenome_%d", genomeNr));
                }
            }
            writer.write("\tmetadata");
            writer.newLine();

            // write presence/absence
            for (Map.Entry<Long, Map<Integer, Long>> entry : presenceAbsence.entrySet()) {
                writer.write(entry.getKey().toString());
                for (int genomeNr = 1; genomeNr <= Globals.total_genomes; genomeNr++) {
                    if (Globals.selectedGenomes.contains(genomeNr)) {
                        writer.write(String.format("\t%d", entry.getValue().getOrDefault(genomeNr, 0L)));
                    }
                }
                writer.write(String.format("\t%s", metadata.get(entry.getKey())));
                writer.newLine();
            }
        }
    }
}
