package nl.wur.bif.pantools.utils.graph.graph_objects;

import org.junit.jupiter.api.Test;

import java.util.*;

import static nl.wur.bif.pantools.utils.graph.graph_objects.GraphRange.comparePositions;
import static org.junit.jupiter.api.Assertions.*;

class GraphRangeTest {
    private String UniquePositionsToString(Set<Integer[]> uniquePositions) {
        final StringBuilder sb = new StringBuilder();
        for (Integer[] position : uniquePositions) {
            sb.append(position[0]).append(":").append(position[1]);
            if (position.length == 3) {
                sb.append("-").append(position[2]);
            }
            sb.append("; ");
        }
        return sb.substring(0, sb.length() - 2);
    }

    /**
     * Simple test for findBlocks with two expected blocks of 3 and 2 unique positions, respectively.
     */
    @Test
    void findBlocksTest1() {
        final Set<Integer[]> uniquePositions = new HashSet<>(); //contains arrays of [sequence, start, end]
        uniquePositions.add(new Integer[]{1, 1, 10});
        uniquePositions.add(new Integer[]{1, 10, 19});
        uniquePositions.add(new Integer[]{1, 13, 22});
        uniquePositions.add(new Integer[]{1, 33, 42});
        uniquePositions.add(new Integer[]{1, 37, 46});
        final int distance = 10;

        final List<Set<Integer[]>> expectedBlocks = new ArrayList<>();
        final Set<Integer[]> block1 = new TreeSet<>(comparePositions());
        block1.add(new Integer[]{1, 1});
        block1.add(new Integer[]{1, 19});
        block1.add(new Integer[]{1, 22});
        expectedBlocks.add(block1);
        final Set<Integer[]> block2 = new TreeSet<>(comparePositions());
        block2.add(new Integer[]{1, 33});
        block2.add(new Integer[]{1, 46});
        expectedBlocks.add(block2);

        final List<Set<Integer[]>> actualBlocks = GraphRange.findBlocks(uniquePositions, distance, 0.4, false);

        assertEquals(expectedBlocks.size(), actualBlocks.size());
        for (int i = 0; i < expectedBlocks.size(); i++) {
            System.out.println("Expected: " + UniquePositionsToString(expectedBlocks.get(i)));
            System.out.println("Actual: " + UniquePositionsToString(actualBlocks.get(i)));
            assertEquals(expectedBlocks.get(i).size(), actualBlocks.get(i).size());
            assertEquals(UniquePositionsToString(expectedBlocks.get(i)), UniquePositionsToString(actualBlocks.get(i)));
        }
    }

    /**
     * A bit more elaborate test for findBlocks with three expected blocks of 3, 2 and 2 unique positions, respectively.
     */
    @Test
    void findBlocksTest2() {
        final Set<Integer[]> uniquePositions = new HashSet<>(); //contains arrays of [sequence, start, end]
        uniquePositions.add(new Integer[]{1, 1, 10});
        uniquePositions.add(new Integer[]{1, 10, 19});
        uniquePositions.add(new Integer[]{1, 13, 22});
        uniquePositions.add(new Integer[]{1, 33, 42});
        uniquePositions.add(new Integer[]{1, 37, 46});
        uniquePositions.add(new Integer[]{1, 67, 76});
        uniquePositions.add(new Integer[]{1, 68, 77});
        final int distance = 10;

        final List<Set<Integer[]>> expectedBlocks = new ArrayList<>();
        final Set<Integer[]> block1 = new TreeSet<>(comparePositions());
        block1.add(new Integer[]{1, 1});
        block1.add(new Integer[]{1, 19});
        block1.add(new Integer[]{1, 22});
        expectedBlocks.add(block1);
        final Set<Integer[]> block2 = new TreeSet<>(comparePositions());
        block2.add(new Integer[]{1, 33});
        block2.add(new Integer[]{1, 46});
        expectedBlocks.add(block2);
        final Set<Integer[]> block3 = new TreeSet<>(comparePositions());
        block3.add(new Integer[]{1, 67});
        block3.add(new Integer[]{1, 77});
        expectedBlocks.add(block3);

        final List<Set<Integer[]>> actualBlocks = GraphRange.findBlocks(uniquePositions, distance, 0, false);

        assertEquals(expectedBlocks.size(), actualBlocks.size());
        for (int i = 0; i < expectedBlocks.size(); i++) {
            System.out.println("Expected: " + UniquePositionsToString(expectedBlocks.get(i)));
            System.out.println("Actual: " + UniquePositionsToString(actualBlocks.get(i)));
            assertEquals(expectedBlocks.get(i).size(), actualBlocks.get(i).size());
            assertEquals(UniquePositionsToString(expectedBlocks.get(i)), UniquePositionsToString(actualBlocks.get(i)));
        }
    }

    /**
     * Test for findBlocks with one expected block of 6 unique positions.
     */
    @Test
    void findBlocksTest3() {
        final Set<Integer[]> uniquePositions = new HashSet<>(); //contains arrays of [sequence, start, end]
        uniquePositions.add(new Integer[]{1, 1, 10});
        uniquePositions.add(new Integer[]{1, 10, 19});
        uniquePositions.add(new Integer[]{1, 13, 22});
        uniquePositions.add(new Integer[]{1, 33, 42});
        uniquePositions.add(new Integer[]{1, 37, 46});
        uniquePositions.add(new Integer[]{1, 67, 76});
        final int distance = 100;

        final List<Set<Integer[]>> expectedBlocks = new ArrayList<>();
        final Set<Integer[]> block1 = new TreeSet<>(comparePositions());
        block1.add(new Integer[]{1, 1});
        block1.add(new Integer[]{1, 19});
        block1.add(new Integer[]{1, 22});
        block1.add(new Integer[]{1, 42});
        block1.add(new Integer[]{1, 46});
        block1.add(new Integer[]{1, 76});
        expectedBlocks.add(block1);

        final List<Set<Integer[]>> actualBlocks = GraphRange.findBlocks(uniquePositions, distance, 0.5, false);

        assertEquals(expectedBlocks.size(), actualBlocks.size());
        for (int i = 0; i < expectedBlocks.size(); i++) {
            System.out.println("Expected: " + UniquePositionsToString(expectedBlocks.get(i)));
            System.out.println("Actual: " + UniquePositionsToString(actualBlocks.get(i)));
            assertEquals(expectedBlocks.get(i).size(), actualBlocks.get(i).size());
            assertEquals(UniquePositionsToString(expectedBlocks.get(i)), UniquePositionsToString(actualBlocks.get(i)));
        }
    }

    /**
     * Hard test for findBlocks with two expected blocks of 3 and 2 unique positions, respectively, but the blocks are right on the border of the distance.
     * Also, this includes a test for the sequence number and for filtering out blocks of size 1.
     */
    @Test
    void findBlocksTest4() {
        final Set<Integer[]> uniquePositions = new HashSet<>(); //contains arrays of [sequence, start, end]
        uniquePositions.add(new Integer[]{1, 1, 10});
        uniquePositions.add(new Integer[]{1, 10, 19});
        uniquePositions.add(new Integer[]{2, 11, 20});
        uniquePositions.add(new Integer[]{1, 13, 22});
        uniquePositions.add(new Integer[]{1, 29, 38});
        uniquePositions.add(new Integer[]{1, 37, 46});
        uniquePositions.add(new Integer[]{1, 77, 86});
        uniquePositions.add(new Integer[]{1, 82, 91});
        uniquePositions.add(new Integer[]{1, 87, 96});
        final int distance = 8;

        final List<Set<Integer[]>> expectedBlocks = new ArrayList<>();
        final Set<Integer[]> block1 = new TreeSet<>(comparePositions());
        block1.add(new Integer[]{1, 1});
        block1.add(new Integer[]{1, 19});
        block1.add(new Integer[]{1, 22});
        block1.add(new Integer[]{1, 38});
        block1.add(new Integer[]{1, 46});
        expectedBlocks.add(block1);
        final Set<Integer[]> block3 = new TreeSet<>(comparePositions());
        block3.add(new Integer[]{1, 77});
        block3.add(new Integer[]{1, 91});
        block3.add(new Integer[]{1, 96});
        expectedBlocks.add(block3);

        final List<Set<Integer[]>> actualBlocks = GraphRange.findBlocks(uniquePositions, distance, 0, false);

        assertEquals(expectedBlocks.size(), actualBlocks.size());
        for (int i = 0; i < expectedBlocks.size(); i++) {
            System.out.println("Expected: " + UniquePositionsToString(expectedBlocks.get(i)));
            System.out.println("Actual: " + UniquePositionsToString(actualBlocks.get(i)));
            assertEquals(expectedBlocks.get(i).size(), actualBlocks.get(i).size());
            assertEquals(UniquePositionsToString(expectedBlocks.get(i)), UniquePositionsToString(actualBlocks.get(i)));
        }
    }
}